Package: sfcurve
Type: Package
Title: 2x2, 3x3 and Nxn Space-Filling Curves
Version: 1.0.1
Date: 2026-04-09
Authors@R: person("Zuguang", "Gu", email = "guzuguang@suat-sz.edu.cn", role = c("aut", "cre"),
                  comment = c('ORCID'="0000-0002-7395-8709"))
Depends: R (>= 4.1.0)
Imports: grid, Rcpp, methods, colorRamp2
Suggests: rmarkdown, knitr, rgl, testthat, ComplexHeatmap, igraph,
        digest
VignetteBuilder: knitr
Description: Implementation of all possible forms of 2x2 and 3x3 space-filling curves, 
    i.e., the generalized forms of the Hilbert curve <https://en.wikipedia.org/wiki/Hilbert_curve>, 
    the Peano curve <https://en.wikipedia.org/wiki/Peano_curve> and the Peano curve in the 
    meander type (Figure 5 in <https://eudml.org/doc/141086>). It can generates nxn curves expanded from
    any specific level-1 units. It also implements the H-curve and the three-dimensional Hilbert curve.
    See <doi:10.48550/arXiv.2412.16962> for more details.
URL: https://github.com/jokergoo/sfcurve,
        https://jokergoo.github.io/sfcurve/
License: MIT + file LICENSE
LinkingTo: Rcpp
Encoding: UTF-8
RoxygenNote: 7.3.1
NeedsCompilation: yes
Packaged: 2026-04-10 07:30:15 UTC; guz
Author: Zuguang Gu [aut, cre] (ORCID: <https://orcid.org/0000-0002-7395-8709>)
Maintainer: Zuguang Gu <guzuguang@suat-sz.edu.cn>
Repository: CRAN
Date/Publication: 2026-04-10 11:00:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-04-15 04:57:08 UTC; unix
Archs: sfcurve.so.dSYM

Package: sfcurve
Type: Package
Title: 2x2, 3x3 and Nxn Space-Filling Curves
Version: 1.0.1
Date: 2026-04-09
Authors@R: person("Zuguang", "Gu", email = "guzuguang@suat-sz.edu.cn", role = c("aut", "cre"),
                  comment = c('ORCID'="0000-0002-7395-8709"))
Depends: R (>= 4.1.0)
Imports: grid, Rcpp, methods, colorRamp2
Suggests: rmarkdown, knitr, rgl, testthat, ComplexHeatmap, igraph,
        digest
VignetteBuilder: knitr
Description: Implementation of all possible forms of 2x2 and 3x3 space-filling curves, 
    i.e., the generalized forms of the Hilbert curve <https://en.wikipedia.org/wiki/Hilbert_curve>, 
    the Peano curve <https://en.wikipedia.org/wiki/Peano_curve> and the Peano curve in the 
    meander type (Figure 5 in <https://eudml.org/doc/141086>). It can generates nxn curves expanded from
    any specific level-1 units. It also implements the H-curve and the three-dimensional Hilbert curve.
    See <doi:10.48550/arXiv.2412.16962> for more details.
URL: https://github.com/jokergoo/sfcurve,
        https://jokergoo.github.io/sfcurve/
License: MIT + file LICENSE
LinkingTo: Rcpp
Encoding: UTF-8
RoxygenNote: 7.3.1
NeedsCompilation: yes
Packaged: 2026-04-10 07:30:15 UTC; guz
Author: Zuguang Gu [aut, cre] (ORCID: <https://orcid.org/0000-0002-7395-8709>)
Maintainer: Zuguang Gu <guzuguang@suat-sz.edu.cn>
Repository: CRAN
Date/Publication: 2026-04-10 11:00:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-04-10 14:44:13 UTC; unix
Archs: sfcurve.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("[",sfc_nxn)
S3method("[",sfc_rules)
S3method("[",sfc_sequence)
S3method("[<-",sfc_sequence)
S3method("^",sfc_sequence)
S3method(c,sfc_sequence)
S3method(length,sfc_sequence)
S3method(makeContext,grob_sfc_sequence)
S3method(plot,sfc_base)
S3method(plot,sfc_nxn)
S3method(plot,sfc_sequence)
S3method(rev,sfc_sequence)
export(BASE_B)
export(BASE_C)
export(BASE_D)
export(BASE_I)
export(BASE_J)
export(BASE_L)
export(BASE_LIST)
export(BASE_P)
export(BASE_Q)
export(BASE_R)
export(BASE_U)
export(H0)
export(H1)
export(H2)
export(SFC_RULES_2x2)
export(SFC_RULES_3x3_COMBINED)
export(SFC_RULES_3x3_MEANDER)
export(SFC_RULES_3x3_PEANO)
export(SFC_RULES_4x4_MEANDER_1)
export(SFC_RULES_4x4_MEANDER_2)
export(add_base_structure)
export(all_2x2_shapes)
export(all_3x3_meander_shapes)
export(all_3x3_peano_shapes)
export(all_traverse_paths)
export(beta_omega_curve)
export(draw_multiple_curves)
export(draw_rules_2x2)
export(draw_rules_3x3_combined)
export(draw_rules_3x3_meander)
export(draw_rules_3x3_peano)
export(draw_rules_4x4_meander)
export(expand_h)
export(get_one_traverse_path)
export(h_curve)
export(hilbert_3d)
export(hilbert_curve)
export(meander_curve)
export(moore_curve)
export(peano_curve)
export(plot_segments)
export(plot_traverse_paths)
export(sfc_2x2)
export(sfc_3x3_combined)
export(sfc_3x3_meander)
export(sfc_3x3_peano)
export(sfc_4x4_meander)
export(sfc_base)
export(sfc_generator)
export(sfc_h)
export(sfc_rules)
export(sfc_seed)
export(sfc_sequence)
export(sfc_unit)
export(test_sfc_index)
export(unit_orientation)
exportMethods(change_level1_unit_orientation)
exportMethods(level1_unit_orientation)
exportMethods(sfc_apply)
exportMethods(sfc_dflip)
exportMethods(sfc_expand)
exportMethods(sfc_expand_by_rules)
exportMethods(sfc_flip_unit)
exportMethods(sfc_grob)
exportMethods(sfc_hflip)
exportMethods(sfc_index)
exportMethods(sfc_is_compatible)
exportMethods(sfc_level)
exportMethods(sfc_mode)
exportMethods(sfc_next_point)
exportMethods(sfc_previous_point)
exportMethods(sfc_reduce)
exportMethods(sfc_reverse)
exportMethods(sfc_rotate)
exportMethods(sfc_segments)
exportMethods(sfc_shape)
exportMethods(sfc_universe)
exportMethods(sfc_validate)
exportMethods(sfc_vflip)
exportMethods(show)
import(Rcpp)
import(colorRamp2)
import(graphics)
import(grid)
import(methods)
importFrom(stats,runif)
useDynLib(sfcurve, .registration = TRUE)

#' Validate the sequence
#' @aliases sfc_validate
#' @rdname sfc_validate
#' @param p An `sfc_sequence` object or a character string. If it is a character string, rotation of zero is assigned to the first letter.
#' @param by One of `sfc_2x2`, `sfc_3x3_peano` and `sfc_3x3_meander`.
#' 
#' @details
#' It is mainly used to validate a seed sequence whether they follow the forward-left-right rule.
#' 
#' @export
#' @return A logical scalar.
#' @examples
#' try(sfc_validate("LLLLL"))
#' try(sfc_validate(sfc_sequence("IIIII", rot = c(0, 90, 180, 270, 0), 
#'         universe = sfc_universe(SFC_RULES_2x2))))
setMethod("sfc_validate",
	signature = "sfc_sequence",
	definition = function(p, by = "sfc_2x2") {

	n = length(p@seq)

	if(by %in% c("sfc_2x2", "sfc_3x3_meander")) {

		if(by == "sfc_2x2") {
			if(!sfc_is_compatible(p, SFC_RULES_2x2)) {
				stop_wrap("The universe of `p` is not compatible with `SFC_RULES_2x2`")
			}
		} else if(by == "sfc_3x3_meander") {
			if(!sfc_is_compatible(p, SFC_RULES_3x3_MEANDER)) {
				stop_wrap("The universe of `p` is not compatible with `SFC_RULES_3x3_MEANDER`")
			}
		}

		ind = which(p@seq == "U")
		if(length(ind)) {
			if( !(n == 1 || (n == 2 && length(ind) == 2)) ) {
				stop_wrap("If U is in a sequence, the sequence should only be U (singlet) or UU (doublet).")
			}
		}

		ind = which(p@seq == "C")
		if(length(ind)) {
			if(n > 1) {
				stop_wrap("If C is in a sequence, the sequence should only be C (singlet).")
			}
		}

		ind = which(p@seq %in% c("B", "D"))
		if(length(ind)) {
			if(!identical(ind, 1L)) {
				stop_wrap("If B/D is in a sequence, B/D should only be the start of the sequence.")
			}
		}

		ind = which(p@seq %in% c("P", "Q"))
		if(length(ind)) {
			if(!identical(ind, n)) {
				stop_wrap("If P/Q is in a sequence, P/Q should only be the end of the sequence.")
			}
		}
	} else if(by %in% "sfc_3x3_peano") {
		if(!sfc_is_compatible(p, SFC_RULES_3x3_PEANO)) {
			stop_wrap("The universe of `p` is not compatible with `SFC_RULES_3x3_PEANO`")
		}
	}

	pos = sfc_segments(p)
	n = nrow(pos)

	if(n > 1) {
		if(any(duplicated(pos))) {
			stop_wrap("Crossing is not allowed in the sequence.")
		} 

		# # two points should be in the same row or the same column
		# l = pos[1:(n-1), 1] == pos[2:n, 1] | pos[1:(n-1), 2] == pos[2:n, 2]
		# if(any(!l)) {
		# 	stop_wrap("Found gaps in the sequences.")
		# }
	}

	return(TRUE)
})


#' @rdname sfc_validate
#' @export
setMethod("sfc_validate",
	signature = "character",
	definition = function(p, by = "sfc_2x2") {

	if(by == "sfc_2x2") {
		p = sfc_sequence(p, universe = sfc_universe(SFC_RULES_2x2))
	} else if(by == "sfc_3x3_peano") {
		p = sfc_sequence(p, universe = sfc_universe(SFC_RULES_3x3_PEANO))
	} else if(by == "sfc_3x3_meander") {
		p = sfc_sequence(p, universe = sfc_universe(SFC_RULES_3x3_MEANDER))
	}

	sfc_validate(p, by = by)
})

#' Various curves in their standard forms
#' 
#' @rdname standard_curve
#' @param level Level of the curve.
#' @param by Which implementation? Only for the testing purpose.
#' 
#' @details
#' These are just special forms of [`sfc_2x2()`], [`sfc_3x3_peano()`], [`sfc_3x3_meander()`] and [`sfc_h()`].
#' @return
#' A two-column matrix of coordinates of points on the curve.
#' @export
#' @examples
#' hilbert_curve(2)
#' draw_multiple_curves(
#'     hilbert_curve(3),
#'     hilbert_curve(4),
#'     nrow = 1
#' )
hilbert_curve = function(level = 2L, by = "Cpp") {
	if(by == "Cpp") {
		lt = hilbert_curve_cpp(level)
		cbind(lt[[1]], lt[[2]])
	} else {
		p = sfc_2x2("R", code = rep(1, level))
		sfc_segments(p)
	}
}

#' @rdname standard_curve
#' @export
#' @examples
#' draw_multiple_curves(
#'     moore_curve(3),
#'     moore_curve(4),
#'     nrow = 1
#' )
moore_curve = function(level = 2L) {
	
	code = rep(2, level)
	code[1] = 1
	p = sfc_2x2("U", code = code)
	sfc_segments(p)
	
}

#' @rdname standard_curve
#' @export
#' @examples
#' draw_multiple_curves(
#'     beta_omega_curve(3),
#'     beta_omega_curve(4),
#'     nrow = 1
#' )
beta_omega_curve = function(level = 2L) {
	
	if(level %% 2 == 1) {
		code = rep(c(1L, 2L), times = (level-1)/2L)
		code = c(2L, code)
	} else {
		code = rep(c(1L, 2L), times = level/2L - 1)
		code = c(2L, code, 1L)
	}

	p = sfc_2x2("C", code = code)
	sfc_segments(p)
}

#' @rdname standard_curve
#' @param pattern The orientation of units on level-2, i.e. the orientation of the 9 3x3 units. The 
#'        value should be a string with 9 letters of "v"/"h" (vertical or horizontal) for the Peano curve,
#'        and "f"/"b" (forward or backward) for the Meander curve. The length of the string should be maximal 9.
#'        If the length is smaller than 9, the stringis automatically recycled.
#' @export
#' @examples
#' draw_multiple_curves(
#'     peano_curve(2),
#'     peano_curve(3),
#'     nrow = 1
#' )
#' draw_multiple_curves(
#'     peano_curve(3, pattern = "vh"),
#'     peano_curve(3, pattern = "vvvhhhvvv"),
#'     nrow = 1
#' )
peano_curve = function(level = 2L, pattern = "vvvvvvvvv", by = "Cpp") {

	if(by == "Cpp" && (pattern == "vvvvvvvvv" || pattern == "v")) {
		lt = peano_curve_cpp(level)
		cbind(lt[[1]], lt[[2]])
	} else {

		if(length(pattern) != 1) {
			stop_wrap("Length of `pattern` can only be 1.")
		}
		pattern = strsplit(pattern, "")[[1]]
		if(length(pattern) == 1) {
			pattern = rep(pattern, 9)
		}
		if(length(pattern) > 9) {
			stop_wrap("`pattern` can only contain maximal 9 letters.")
		}

		if(!all(pattern %in% c("h", "v"))) {
			stop_wrap("`pattern` should contain v/h")
		}

		bp = "I"
		l_v = pattern == "v"
		l_h = pattern == "h"
		p = sfc_3x3_peano(bp, code = rep(1, level), rot = 0, flip = function(p) {
			if(sfc_level(p) > 0) {
				n = length(p)
				l = rep(FALSE, n)
				l1 = rep(l_v, times = ceiling(9^(sfc_level(p))/length(pattern)))
				l1 = l1[1:n]
				l[l1] = p@rot[l1] %in% c(90, 270)
				
				l2 = rep(l_h, times = ceiling(9^(sfc_level(p))/length(pattern)))
				l2 = l2[1:n]
				l[l2] = p@rot[l2] %in% c(0, 180)
				l
			} else {
				FALSE
			}
		})

		sfc_segments(p)
	}
}

#' @rdname standard_curve
#' @param code Internally used.
#' @export
#' @examples
#' draw_multiple_curves(
#'     meander_curve(2),
#'     meander_curve(3),
#'     nrow = 1
#' )
#' draw_multiple_curves(
#'     meander_curve(3, pattern = "fbfbfbfbf"),
#'     meander_curve(3, pattern = "bbbbbffff"),
#'     nrow = 1
#' )
meander_curve = function(level = 2L, pattern = "fffffffff", code = rep(1, level)) {
	
	if(length(pattern) != 1) {
		stop_wrap("Length of `pattern` can only be 1.")
	}
	pattern = strsplit(pattern, "")[[1]]
	if(length(pattern) == 1) {
		pattern = rep(pattern, 9)
	}
	if(length(pattern) > 9) {
		stop_wrap("`pattern` can only contain maximal 9 letters.")
	}

	if(!all(pattern %in% c("f", "b"))) {
		stop_wrap("`pattern` should contain f/b")
	}

	bp = "R"
	rot = 0

	l_f = pattern == "f"
	l_b = pattern == "b"
	p = sfc_3x3_meander(bp, code = code, rot = 0, flip = function(p) {
		if(sfc_level(p) > 0) {
			n = length(p)
			l = rep(FALSE, n)

			l2 = rep(l_b, times = ceiling(9^(sfc_level(p))/length(pattern)))
			l2 = l2[1:n]
			l[l2] = TRUE
			l
		} else {
			FALSE
		}
	})
	sfc_segments(p)
	
}


#' @rdname standard_curve
#' @param iteration Number of iterations.
#' @export
#' @examples
#' draw_multiple_curves(
#'     h_curve(1),
#'     h_curve(2),
#'     nrow = 1, closed = TRUE
#' )
h_curve = function(iteration = 2L) {
	
	sfc_h(H1, iteration = iteration, connect = "h", random = FALSE)
	
}


#' Plot segments
#' @param x A two-column matrix of coordinates of points.
#' @param grid Whether to add grid lines on the plot?
#' @param title The value should be `FALSE` or a string.
#' @param closed Whether the curve is closed?
#' @param ... Other arguments passed to [`sfc_grob()`].
#' @export
#' @details
#' This function is only for a quick demonstration of curves represented as two-column coordinate matrices.
#' @return No value is returned.
#' @examples
#' pos = cbind(c(0, 0, 1, 1, 2, 2, 3, 3, 2, 2, 1, 1),
#'             c(1, 2, 2, 3, 3, 2, 2, 1, 1, 0, 0, 1))
#' plot_segments(pos)
plot_segments = function(x, grid = FALSE, title = FALSE, closed = FALSE, ...) {
	gb = sfc_grob(x, title = title, closed = closed, ...)
	grid.newpage()
	grid.draw(gb)

	if(grid) {
		add_grid_lines()
	}
}

library(testthat)

test_that("test standard curves", {
	loc1 = hilbert_curve(2, by = "Cpp")
	loc2 = hilbert_curve(2, by = "R")
	expect_equal(loc1, loc2)

	loc1 = hilbert_curve(3, by = "Cpp")
	loc2 = hilbert_curve(3, by = "R")
	expect_equal(loc1, loc2)

	loc1 = hilbert_curve(4, by = "Cpp")
	loc2 = hilbert_curve(4, by = "R")
	expect_equal(loc1, loc2)

	loc1 = peano_curve(2, by = "Cpp")
	loc2 = peano_curve(2, by = "R")
	expect_equal(loc1, loc2)

	loc1 = peano_curve(3, by = "Cpp")
	loc2 = peano_curve(3, by = "R")
	expect_equal(loc1, loc2)

	loc1 = peano_curve(4, by = "Cpp")
	loc2 = peano_curve(4, by = "R")
	expect_equal(loc1, loc2)
})

#' Three dimensional Hilbert curve
#' 
#' @param level The level of the curve.
#' 
#' @details
#' There are many forms of 3D Hilbert curve. Here we only implement one specific form.
#' 
#' @export
#' @import Rcpp
#' @useDynLib sfcurve, .registration = TRUE
#' @seealso Michael Bader. Space-Filling Curves: An Introduction with Applications in Scientific Computing, Springer Science & Business Media, 2012. \doi{10.1007/978-3-642-31046-1}.
#' @return A three-column matrix of coordinates of points on the 3D Hilbert curve.
#' @examples
#' pos = hilbert_3d(2)
#' \dontrun{
#' if(require(rgl) && interactive()) {
#'     plot3d(pos, type = "l", lwd = 4, col = 2)
#' }
#' }
hilbert_3d = function(level = 2L) {
	if(level < 1) {
		stop_wrap("`level` should not smaller than 1.")
	}
	lt = hilbert_curve_3d_cpp(level)

	cbind(x = lt[[1]], y = lt[[2]], z = lt[[3]])
}

#' Constructor of the sfc_sequence class
#' 
#' @param seq A sequence of base patterns. The value can be a vector of letters or a single string.
#' @param rot The corresponding rotations of base patterns. If it has length one and the sequence contains R/L/I (right/left/forward),
#'          `rot` controls the rotation of the first base pattern and the rotations for remaining base patterns in the sequence are automatically
#'          calculated.
#' @param universe The universe of base patterns. A vector of letters.
#' 
#' @details
#' This funtion is very low-level. Normally, users do not need to directly use this constructor.
#' 
#' @export
#' @return An `sfc_sequence` object.
#' @examples
#' sfc_sequence("ABCD", rot = c(0, 90, 180, 270), universe = c("A", "B", "C", "D"))
sfc_sequence = function(seq, rot = 0L, universe = NULL) {

	if(is.factor(seq)) {
		universe = levels(seq)
	} else {
		if(length(seq) == 1) {
			seq = strsplit(seq, "")[[1]]
		}
		if(is.null(universe)) {
			seq = factor(seq)
			universe = levels(seq)
		} else {
			seq = factor(seq, levels = universe)
		}
	}

	n = length(seq)

	rot = as.integer(rot)
	if(length(rot) == 1) {
		rot2 = integer(n)
		rot2[1] = rot[1]
		if(n > 1) {
			for(i in 2:n) {
				rot2[i] = next_rotation(seq[i-1], rot2[i-1])
			}
		}
		rot = rot2
	}

	diff = setdiff(levels(seq), universe)
	if(length(diff)) {
		stop_wrap("all letters should be included in universe.")
	}
	
	p = new("sfc_sequence")
	p@seq = seq
	p@rot = rot %% 360L
	p@universe = universe


	p
}

#' @rdname sfc_sequence
#' @details
#' `sfc_seed` class is the same as the `sfc_sequence` class. It is used specifically as the "seed sequence"
#' when generating the curves. 
#' @export
sfc_seed = function(seq, rot = 0L, universe = NULL) {
	p = sfc_sequence(seq, rot, universe)
	p2 = new("sfc_seed")
	p2@seq = p@seq
	p2@rot = p@rot
	p2@universe = p@universe
	p2
}



#' @rdname sfc_sequence
#' @details
#' `sfc_unit` class also inherits the `sfc_sequence` class but has one additionally slot: `corner`. 
#' It is used specifically when defining the expansion rules. 
#' @export
sfc_unit = function(seq, rot = 0L, universe = NULL) {
	p = sfc_sequence(seq, rot, universe)

	p2 = new("sfc_unit")
	p2@seq = p@seq
	p2@rot = p@rot
	p2@universe = p@universe

	p2
}



# rotation of the direction of approaching
next_rotation = function(letter, rot) {
	if(letter == "R") {
		return(rot - 90L)
	} else if(letter == "L") {
		return(rot + 90L)
	} else if(letter == "U") {
		return(rot + 180L)
	} else {
		return(rot)
	}
}


# subset allows an incomplete curve

#' Utility functions
#' @rdname utility
#' @param x An `sfc_sequence` object.
#' @param i Numeric index.
#' @param value An `sfc_sequence` object.
#' @param ... A list of `sfc_sequence` objects or other arguments.
#' 
#' @export
#' @examples
#' p = sfc_sequence("ABCDEFGH")
#' p
#' p[1:4]
#' p[1:4] = p[8:5]
#' p
#' length(p)
#' c(p, p)
`[.sfc_sequence` = function(x, i) {
	x2 = new("sfc_sequence")
	x2@seq = x@seq[i]
	x2@rot = x@rot[i]
	x2@universe = x@universe
	x2
}


#' @rdname utility
#' @export
`[<-.sfc_sequence` = function(x, i, value) {
	if(sfc_is_compatible(x, value, strict = FALSE)) {
		x@seq[i] = value@seq
		x@rot[i] = value@rot
		x
	} else {
		stop_wrap("`x` and `value` should be compatible.")
	}
}

#' @rdname utility
#' @return 
#' `length.sfc_sequence()` returns an integer scalar.
#' @export
length.sfc_sequence = function(x) {
	length(x@seq)
}

#' @rdname utility
#' @details
#' For efficiency, `c.sfc_sequence()` does not check whether the input `sfc_sequence` objects are compatible.
#' @return
#' `c.sfc_sequence()` returns an `sfc_sequence` object.
#' @export
c.sfc_sequence = function(...) {
	lt = list(...)
	seq = do.call("c", lapply(lt, function(x) x@seq))
	rot = do.call("c", lapply(lt, function(x) x@rot))
	
	p = new("sfc_sequence")
	p@seq = seq
	p@rot = rot
	p@universe = lt[[1]]@universe

	p
}


#' @param object The corresponding object.
#' @rdname show
#' @export
#' @return No value is returned.
setMethod("show",
	signature = "sfc_sequence",
	definition = function(object) {

	n = length(object)
	nr = ceiling(n/8)

	seq = object@seq
	rot = object@rot

	rot_str = function(x) {
		# ifelse(x == 0, "  0", ifelse(x == 90, " 90", x))
		x
	}

	cat("A sequence of ", n, " base pattern", ifelse(n == 1, ".", "s."), "\n", sep = "")
	if(nr > 4) {
		other_lines = paste0(".... other ", nr-4, " line", ifelse(nr-4 > 1, "s", ""), " ....")

		df = matrix("", nrow = 4, ncol = 2)
		for(i in 1:2) {
			ind = seq( (i-1)*8+1, i*8 )
			df[i, 1] = paste(seq[ ind[1:4] ], "(", rot_str(rot[ ind[1:4] ]), ")", sep = "", collapse = "")
			df[i, 2] = paste(seq[ ind[5:8] ], "(", rot_str(rot[ ind[5:8] ]), ")", sep = "", collapse = "")
		}

		ind = seq( (nr-2)*8+1, (nr-1)*8 )
		df[3, 1] = paste(seq[ ind[1:4] ], "(", rot_str(rot[ ind[1:4] ]), ")", sep = "", collapse = "")
		df[3, 2] = paste(seq[ ind[5:8] ], "(", rot_str(rot[ ind[5:8] ]), ")", sep = "", collapse = "")

		ind = seq( (nr-1)*8+1, min((nr)*8, n) )
		if(length(ind) <= 4) {
			df[4, 1] = paste(seq[ ind ], "(", rot_str(rot[ ind ]), ")", sep = "", collapse = "")
			df[4, 2] = ""
		} else if(length(ind) < 8) {
			df[4, 1] = paste(seq[ ind[1:4] ], "(", rot_str(rot[ ind[1:4] ]), ")", sep = "", collapse = "")
			df[4, 2] = paste(seq[ ind[5:length(ind)] ], "(", rot_str(rot[ ind[5:length(ind)] ]), ")", sep = "", collapse = "")
		} else {
			df[4, 1] = paste(seq[ ind[1:4] ], "(", rot_str(rot[ ind[1:4] ]), ")", sep = "", collapse = "")
			df[4, 2] = paste(seq[ ind[5:8] ], "(", rot_str(rot[ ind[5:8] ]), ")", sep = "", collapse = "")
		}

		max_nchar1 = max(nchar(df[, 1]))
		max_nchar2 = max(nchar(df[, 2]))

		for(i in 1:2) {
			cat("  ")
			cat(df[i, 1]); cat(strrep(" ", max_nchar1 - nchar(df[i, 1])))
			cat("  ")
			cat(df[i, 2]); cat("\n")
		}
		cat("  ", other_lines, "\n", sep = "")
		for(i in 3:4) {
			cat("  ")
			cat(df[i, 1]); cat(strrep(" ", max_nchar1 - nchar(df[i, 1])))
			cat("  ")
			cat(df[i, 2]); cat("\n")
		}

	} else {
		df = matrix("", nrow = nr, ncol = 2)
		for(i in 1:nr) {
			ind = seq( (i-1)*8+1, min(i*8, n) )

			if(length(ind) <= 4) {
				df[i, 1] = paste(seq[ ind ], "(", rot_str(rot[ ind ]), ")", sep = "", collapse = "")
				df[i, 2] = ""
			} else if(length(ind) < 8) {
				df[i, 1] = paste(seq[ ind[1:4] ], "(", rot_str(rot[ ind[1:4] ]), ")", sep = "", collapse = "")
				df[i, 2] = paste(seq[ ind[5:length(ind)] ], "(", rot_str(rot[ ind[5:length(ind)] ]), ")", sep = "", collapse = "")
			} else {
				df[i, 1] = paste(seq[ ind[1:4] ], "(", rot_str(rot[ ind[1:4] ]), ")", sep = "", collapse = "")
				df[i, 2] = paste(seq[ ind[5:8] ], "(", rot_str(rot[ ind[5:8] ]), ")", sep = "", collapse = "")
			}
		}

		max_nchar1 = max(nchar(df[, 1]))
		max_nchar2 = max(nchar(df[, 2]))

		for(i in 1:nr) {
			cat("  ")
			cat(df[i, 1]); cat(strrep(" ", max_nchar1 - nchar(df[i, 1])))
			cat("  ")
			cat(df[i, 2]); cat("\n")
		}
	}
})

#' @rdname sfc_universe
#' @export
setMethod("sfc_universe",
	signature = "sfc_sequence",
	definition = function(p) {
	p@universe
})

#' Whether two sfc_sequence objects are compatible
#' @aliases sfc_is_compatible
#' @rdname sfc_is_compatible
#' @param p1 An `sfc_sequence` object.
#' @param p2 An `sfc_sequence` object.
#' @param strict `TRUE` or `FALSE`, see **Details**.
#' 
#' @details
#' The function compares whether the two universe base pattern sets are identical.
#' If `strict` is `TRUE`, the order of the two universe sets should also be the same.
#' If `strict` is `FALSE`, the universe set of `p2` can be a subset of the universe set of `p1`.
#' @export
#' @return A logical scalar.
#' @examples
#' p1 = sfc_2x2("I")
#' p2 = sfc_2x2("R")
#' sfc_is_compatible(p1, p2)
#' 
#' p1 = sfc_2x2("I")
#' p2 = sfc_sequence("R")
#' sfc_is_compatible(p1, p2)
#' sfc_is_compatible(p1, p2, strict = FALSE)
#' 
#' p1 = sfc_sequence("ABC")
#' p2 = sfc_sequence("DEF")
#' sfc_is_compatible(p1, p2)
setMethod("sfc_is_compatible",
	signature = c("sfc_sequence", "sfc_sequence"),
	definition = function(p1, p2, strict = TRUE) {
	if(strict) {
		identical(sfc_universe(p1), sfc_universe(p2))
	} else {
		length(setdiff(sfc_universe(p2), sfc_universe(p1))) == 0
	}
})

#' @rdname sfc_is_compatible
#' @export
setMethod("sfc_is_compatible",
	signature = c("sfc_sequence", "sfc_rules"),
	definition = function(p1, p2) {
	identical(sfc_universe(p1), sfc_universe(p2))
})

#' @rdname sfc_is_compatible
#' @export
setMethod("sfc_is_compatible",
	signature = c("sfc_rules", "sfc_sequence"),
	definition = function(p1, p2) {
	identical(sfc_universe(p1), sfc_universe(p2))
})

#' Constructor of the sfc_rules class
#' 
#' @param rules A list of rules.
#' @param bases A list of base patterns.
#' @param flip A list of rules. They are "flipped" version of `rules`. The value can also simply be `TRUE`, then
#'           the flipped version is automatically generated from `rules`.
#' @param name A self-defined string.
#' 
#' @details
#' It is mainly used internally.
#' 
#' `rules` is a two-level list. It is in a format of `rules[[ base ]][[ expansion_code ]] = sfc_unit()`.
#' In the following example where we define the expansion rules for the 2x2 curve:
#' 
#' ```
#' UNIVERSE_2x2 = c("I", "R", "L", "U", "B", "D", "P", "Q", "C")
#' RULES_2x2 = list()
#' RULES_2x2[["I"]][[1]] = sfc_unit(c("R", "L", "L", "R"), rot = 0, universe = UNIVERSE_2x2)
#' ```
#' 
#' `I` is the level-0 base pattern, `[[1]]` corresponds to the first form of expansion to level-1, and the value
#' assigned is a [`sfc_unit()`] object which is basically a list of base patterns.
#' 
#' Then we also need to provide the base patterns which define how to extend the curve. The list of base patterns
#' is assigned to the `bases` argument. In the same example, we set `bases` as:
#' 
#' ```
#' list("I" = BASE_I, "R" = BASE_R, "L" = BASE_L, "U" = BASE_U, ...)
#' ```
#' 
#' where e.g. [`BASE_I`] is a pre-defined base pattern in the [`sfc_base`] class.
#' 
#' There are the following pre-defined rules:
#' 
#' - [`SFC_RULES_2x2`]
#' - [`SFC_RULES_3x3_PEANO`]
#' - [`SFC_RULES_3x3_MEANDER`]
#' - [`SFC_RULES_3x3_COMBINED`]
#' - [`SFC_RULES_4x4_MEANDER_1`]
#' - [`SFC_RULES_4x4_MEANDER_2`]
#' 
#' Check \url{https://github.com/jokergoo/sfcurve/blob/master/R/zz_global.R} to see how these pre-defined rules are constructed.
#' 
#' @export
#' @return
#' An `sfc_rules` object.
sfc_rules = function(rules, bases, flip = list(), name = "sfc_rules") {
    r = new("sfc_rules")

    for(nm in names(rules)) {
        if(!is.list(rules[[nm]])) {
            rules[[nm]] = list(rules[[nm]])
        }
    }

    universe = sfc_universe(rules[[1]][[1]])
    if( !identical(sort(names(rules)), sort(universe)) ) {
        stop_wrap("`names(rules)` should be identical to the universe of its level-1 patterns.")
    }
    rules = rules[universe]
    r@bases = bases
    r@universe = universe
    r@name = name

    if(identical(flip, TRUE)) {
        flip = lapply(rules, function(x) {
            lapply(x, function(u) sfc_flip_unit(u, rules@base))
        })
    }

    if(length(flip) > 0) {
        for(nm in names(flip)) {
            if(!is.list(flip[[nm]])) {
                flip[[nm]] = list(flip[[nm]])
            }
        }

        universe2 = sfc_universe(flip[[1]][[1]])
        if( !identical(names(flip), universe2) ) {
            stop_wrap("`names(flip)` should be identical to the universe of its level-1 patterns.")
        }
        if( !identical(names(flip), r@universe) ) {
            stop_wrap("`names(flip)` should be identical to `names(rules)`.")
        }

        flip = flip[names(rules)]

        for(nm in names(rules)) {
            if(length(rules[[nm]]) != length(flip[[nm]])) {
                stop_wrap("Length of `rules[['", nm, "']]` should be the same as `flip[['", nm, "']]`.")
            }
        }

        # validate whether rules and flip are compatible, i.e. flip[[i]][[j]] is the flipped version of rules[[i]][[j]]
        for(nm in names(rules)) {
            for(j in seq_along(rules[[nm]])) {
                u1 = flip[[nm]][[j]]
                u2 = sfc_flip_unit(rules[[nm]][[j]], bases)
                if(!identical(u1, u2)) {
                    stop_wrap("`flip[['", nm, "']][[", j, "]]` is not a flip of `rules[[", nm, "]][[", j, "]]`.")
                }
            }
        }
    }

    for(nm in names(rules)) {
        for(i in seq_along(rules[[nm]])) {
            validate_rule(rules[[nm]][[i]], nm, i, bases)
        }
    }

    if(length(flip) > 0) {
        for(nm in names(flip)) {
            for(i in seq_along(flip[[nm]])) {
                validate_rule(flip[[nm]][[i]], nm, i, bases)
            }
        }
    }

    # add corner value for each rule
    .add_corners = function(rules) {
        for(i in seq_along(rules)) {
            for(j in seq_along(rules[[i]])) {
                p = rules[[i]][[j]]

                if(!sfc_is_compatible(rules[[1]][[1]], p)) {
                    stop_wrap(paste0("The universe of rules[[", names(rules)[1], "]][[1]] is not compatible with rules[[", names(rules)[i], "]][[", j, "]]."))
                }

                first_pos = c(0, 0)
                min_x = first_pos[1]
                max_x = first_pos[1]
                min_y = first_pos[2]
                max_y = first_pos[2]
                current_pos = first_pos
                seq = p@seq
                rot = p@rot
                for(k in seq_along(seq)[-1]) {
                    current_pos = sfc_next_point(bases[[ as.character(seq[k-1]) ]], current_pos, rot[k-1])

                    min_x = min(min_x, current_pos[1])
                    max_x = max(max_x, current_pos[1])
                    min_y = min(min_y, current_pos[2])
                    max_y = max(max_y, current_pos[2])
                }
                last_pos = current_pos

                corner = c(0L, 0L)
                if( (equal_to(first_pos[1], min_x) && equal_to(first_pos[2], min_y)) || 
                    (equal_to(first_pos[1], max_x) && equal_to(first_pos[2], max_y)) ) {
                    corner[1] = 1L
                }
                
                if( (equal_to(first_pos[1], min_x) && equal_to(first_pos[2], max_y)) ||
                    (equal_to(first_pos[1], max_x) && equal_to(first_pos[2], min_y)) ) {
                    corner[1] = 2L
                }

                if( (equal_to(last_pos[1], min_x) && equal_to(last_pos[2], min_y)) || 
                    (equal_to(last_pos[1], max_x) && equal_to(last_pos[2], max_y)) ) {
                    corner[2] = 1L
                }

                if( (equal_to(last_pos[1], min_x) && equal_to(last_pos[2], max_y)) ||
                    (equal_to(last_pos[1], max_x) && equal_to(last_pos[2], min_y)) ) {
                    corner[2] = 2L
                }

                rules[[i]][[j]]@corner = corner

            }
        }
        rules
    }

    r@rules = .add_corners(rules)
    if(length(flip)) {
        r@flip = .add_corners(flip)
    }
    
    r
}


# the in- and out-direction of the unit should be the same as the base
validate_rule = function(rule, bp, i, bases) {
    if( !equal_to(sqrt(length(rule)), round(sqrt(length(rule)))) ) {
        stop_wrap("The length of the unit should be k^2 where k = 2, 3, ...")
    }

    if(!is.na(bases[[ bp ]]@in_direction)) {
        prev_p1 = sfc_previous_point(bases[[ as.character(rule@seq[1]) ]], c(0, 0), rule@rot[1])
        prev_p2 = sfc_previous_point(bases[[ bp ]], c(0, 0), 0)

        if(!equal_to(prev_p1, prev_p2)) {
            prev_p1 = sfc_previous_point(bases[[ as.character(rule@seq[1]) ]], c(0, 0), 0)
            if(equal_to(prev_p1, prev_p2)) {
                stop_wrap("The in-direction is different from the base pattern ", bp, " to the level-1 unit. Maybe you need to rotate the unit by ", -rule@rot[1] %% 360, " degrees.")
            } else {
                stop_wrap("The in-direction is different from the base pattern ", bp, " to the level-1 unit.")
            }
        }
    }

    if(!is.na(bases[[ bp ]]@out_direction)) {
        n = length(rule)
        next_p1 = sfc_next_point(bases[[ as.character(rule@seq[n]) ]], c(0, 0), rule@rot[n])
        next_p2 = sfc_next_point(bases[[ bp ]], c(0, 0), 0)

        if(!equal_to(next_p1, next_p2)) {
            stop_wrap("The out-direction is different from the base pattern ", bp, " to the level-1 unit.")
        }
    }
}


#' @export
`[.sfc_rules` = function(x, i) {
    if(is.character(i)) {
        i = intersect(names(x@rules), i)
    } else if(is.numeric(i)) {
        i = intersect(seq_along(x@rules), i)
    }
    x@rules = x@rules[i]
    if(length(x@flip) > 0) {
        x@flip = x@flip[i]
    }
    x
}

#' @rdname show
#' @export
setMethod("show",
    signature = "sfc_rules",
    definition = function(object) {

    rot_str = function(x) {
        # ifelse(x == 0, "  0", ifelse(x == 90, " 90", x))
        x
    }

    .get_output = function(rules) {
        lt = rules
        output = list(character(0), character(0), character(0), character(0))
        for(nm in names(lt)) {
            if(is.null(names(lt[[nm]]))) {
                ind = seq_along(lt[[nm]])
            } else {
                ind = names(lt[[nm]])
            }

            k = 1
            for(i in ind) {
                if(k > 1) {
                    output[[1]] = c(output[[1]], "   ")
                } else {
                    output[[1]] = c(output[[1]], paste0(nm, " |"))
                }
                k = k + 1
                output[[2]] = c(output[[2]], paste0(nm, "_", i))

                seq = lt[[nm]][[i]]@seq
                rot = lt[[nm]][[i]]@rot
                corner = lt[[nm]][[i]]@corner
                output[[3]] = c(output[[3]], paste(seq, "(", rot_str(rot), ")", sep = "", collapse = ""))
                
                output[[4]] = c(output[[4]], paste0(" corner = (", corner[1], ", ", corner[2], ")"))
            }
        }

        max_nchar = sapply(output, function(x) max(nchar(x)))
        for(i in seq_along(output[[1]])) {
            cat(output[[1]][i]); cat(strrep(" ", max_nchar[1] - nchar(output[[1]][i]))); cat(" ")
            cat(output[[2]][i]); cat(strrep(" ", max_nchar[2] - nchar(output[[2]][i]))); cat(" = ")
            cat(output[[3]][i]); cat(strrep(" ", max_nchar[3] - nchar(output[[3]][i]))); cat(" ")
            cat(output[[4]][i]); cat(strrep(" ", max_nchar[4] - nchar(output[[4]][i]))); cat(" ")
            cat("\n")
        }
    }

    cat("Name: ", object@name, "\n", sep = "")
    .get_output(object@rules)
    
    if(length(object@flip)) {
        cat("\n")
        cat("Flipped:\n")
        .get_output(object@flip)
    }
})

#' The universe base pattern set
#' 
#' @param p The corresponding object.
#' @aliases sfc_universe
#' @rdname sfc_universe
#' @return
#' A vector of base patterns.
#' @export
#' @examples
#' sfc_universe(SFC_RULES_2x2)
#' sfc_universe(SFC_RULES_3x3_PEANO)
setMethod("sfc_universe",
    signature = "sfc_rules",
    definition = function(p) {

    p@universe
})


#' The mode of the curve
#' @aliases sfc_mode
#' @rdname sfc_mode
#' @param p The corresponding object.
#' @export
#' @examples
#' sfc_mode(SFC_RULES_2x2)
#' sfc_mode(SFC_RULES_3x3_PEANO)
setMethod("sfc_mode",
    signature = "sfc_rules",
    definition = function(p) {

    as.integer(sqrt(length(p@rules[[1]][[1]])))
})

#' Expand a sequence
#' @rdname sfc_expand_by_rules
#' @aliases sfc_expand_by_rules
#' @param p An `sfc_rules` object.
#' @param seq An `sfc_nxn` object or other objects.
#' @param code The expansion code.
#' @param flip For the Peano curve and the Meander curves, each unit can be flipped without affecting other parts in the curve. This argument
#'        controls whether to flip the unit. Since currently it only works on the Peano curve and the Meander curve, `flip` should be a logical
#'        vector of length one or with the same length as `seq`. Whether it flips horizontally, vertically or against the diagonal line is automatically choosen.
#'        The value of `flip` can also be a function which takes the current curve as the only argument.
#' @param by Which implementation? Only for the testing purpose.
#' 
#' @return
#' If `seq` is an `sfc_nxn` object, the function also returns an "expanded" `sfc_nxn` object. Or else it returns an `sfc_sequence` object.
#' @export
#' @examples
#' sfc_expand_by_rules(SFC_RULES_2x2, sfc_2x2("I"))
setMethod("sfc_expand_by_rules", 
    signature = c("sfc_rules", "sfc_nxn"),
    definition = function(p, seq, code = 1L, flip = FALSE, by = "Cpp") {

    if(!identical(sfc_universe(seq), sfc_universe(p))) {
        stop_wrap("The universe of `p` and `seq` should be identical.")
    }
   
    n = length(seq)
    letters = seq@seq

    do_flipping = TRUE
    if(length(flip) == 0) {
        flip = rep(FALSE, n)
        do_flipping = FALSE
    } else {
        if(is.logical(flip)) {
            if(length(flip) == 1) {
                flip = rep(flip, n)
            } else if(length(flip) == n) {

            } else {
                flip = rep(flip, times = n/sfc_mode(p)^(2*sfc_level(seq)))
            }
            if(length(flip) != n) {
                stop_wrap("The logical vector `flip` should have a length of one or with the same length as `seq`.")
            }
        } else {

            flip = flip(seq)
            if(length(flip) != n) {
                stop_wrap("The self-deifned function `flip` should return a logical vector with current curve as its argument.")
            }
        }
    }

    if(by == "R") {
        if(length(code) == 1) {
            pl = lapply(seq_along(letters), function(i) {
                if(flip[i]) {
                    p@flip[[ letters[i] ]][[ code ]]
                } else {
                    p@rules[[ letters[i] ]][[ code ]]
                }
            }) # x is as an integer
        } else if(length(letters) == length(code)) {
            pl = lapply(seq_along(letters), function(i) {
                if(flip[i]) {
                    p@flip[[ letters[i] ]][[ code[i] ]]
                } else {
                    p@rules[[ letters[i] ]][[ code[i] ]]
                }
            })
        }
        p2 = do.call("c", pl)
    } else {
        lt_rules = lapply(p@rules, function(r) {
            lapply(r, function(x) {
                list(seq = as.integer(x@seq), rot = x@rot)
            })
        })

        if(!do_flipping) {
            lt = expand_by_rules_cpp(lt_rules, as.integer(letters), code)
        } else {
            lt_flip = lapply(p@flip, function(r) {
                lapply(r, function(x) {
                    list(seq = as.integer(x@seq), rot = x@rot)
                })
            })

            lt = expand_by_rules_2_cpp(lt_rules, lt_flip, as.integer(letters), code, flip)
        }
        attributes(lt[[1]]) = list(levels = sfc_universe(p), class = "factor")
        p2 = sfc_sequence(seq = lt[[1]], rot = lt[[2]])
    }

    p2 = sfc_rotate(p2, rep(seq@rot, each = length(p@rules[[1]][[1]])))

    if(inherits(seq, "sfc_nxn")) {
        p3 = new(class(seq))
        p3@seq = p2@seq
        p3@rot = p2@rot
        p3@universe = p2@universe
        p3@level = seq@level + 1L
        p3@mode = seq@mode
        p3@seed = seq@seed
        p3@rules = seq@rules
        p3
    } else {
        p2
    }
    
})

#' @rdname sfc_expand_by_rules
#' @export
setMethod("sfc_expand_by_rules", 
    signature = c("sfc_rules", "factor"),
    definition = function(p, seq, code = 1L, flip = FALSE, by = "Cpp") {

    seq = sfc_sequence(seq, universe = sfc_universe(p))
    sfc_expand_by_rules(p, seq, code = code, flip = flip, by = by)

})

#' @rdname sfc_expand_by_rules
#' @export
setMethod("sfc_expand_by_rules", 
    signature = c("sfc_rules", "character"),
    definition = function(p, seq, code = 1L, flip = FALSE, by = "Cpp") {

    seq = sfc_sequence(seq, universe = sfc_universe(p))
    sfc_expand_by_rules(p, seq, code = code, flip = flip, by = by)
})


grob_math = function(label, x, y, gp = gpar(), ...) {
    expr = parse(text = label)
    gp$fontfamily = "Times"
    textGrob(expr, x, y, ..., gp = gp)
}

tex_pattern = function(x, which, p) {
    seq = p@seq
    rot = p@rot
    l = rot > 0
    rot[l] = paste0("^", rot[l], "")
    rot[!l] = ""
    
    seq = paste0("italic(", seq, ")")
    expr = paste(seq, rot, sep = "", collapse = ",")
    paste0("italic(", x, ")[", which, "] == paste(", expr, ")")
}

grob_single_base_rule = function(p, bp, equation_max_width, flip = FALSE, ...) {
    rules = p@rules
    bases = rules@bases

    if(inherits(p, "sfc_2x2")) {
        level0 = sfc_2x2(bp)
    } else if(inherits(p, "sfc_3x3_peano")) {
        level0 = sfc_3x3_peano(bp, flip = flip)
    } else if(inherits(p, "sfc_3x3_meander")) {
        level0 = sfc_3x3_meander(bp, flip = flip)
    } else if(inherits(p, "sfc_3x3_combined")) {
        level0 = sfc_3x3_combined(bp, level = 0, flip = flip)
    } else if(inherits(p, "sfc_4x4_meander")) {
        level0 = sfc_4x4_meander(bp, type = p@type, flip = flip)
    } else {
        cl = getClass(class(p))
        fun = get(class(p), envir = as.environment(cl@package))
        level0 = fun(bp, flip = flip)
    }

    pl = rules@rules[[bp]]

    for(i in seq_along(pl)) {
        pl[[i]] = sfc_expand(level0, i, flip = flip)
    }

    n = length(pl)
    nr = ceiling(n/2)
    nc = ceiling(n/nr)

    k = as.integer(sqrt(length(pl[[1]]@seq)))

    vp_xscale = c(0, 4 + (nc*k + nc) +7)
    vp_yscale = c(-(nr-1)*(k+2)-1, k+1)

    vp_width = (4 + (nc*k + nc))*size + equation_max_width + unit(8, "mm")
    vp_height = diff(vp_yscale)*size

    vp_xscale[2] = convertWidth(vp_width, "mm", valueOnly = TRUE)/as.numeric(size)
    vp = viewport(xscale = vp_xscale, yscale = vp_yscale, width = vp_width, height = vp_height, ...)

    gbl = list()

    gbl[[1]] = grob_math(paste0("italic(", bp, ")"), 1, k/2, default.units = "native")

    gbl[[2]] = sfc_grob(bases[[bp]])
    gbl[[2]]$vp$x = unit(2.5, "native")
    gbl[[2]]$vp$y = unit(k/2, "native")

    ii = 2;
    for(i in 1:nr) {
        ii = ii + 1
        gbl[[ii]] = sfc_grob(pl[[i*2-1]], x = 4 + k/2, y = k/2 - (i-1)*(k+2), default.units = "native", extend = TRUE)
        gbl[[ii]]$children[[1]]$gp = gpar(col = "grey", lwd = 1)
        gbl[[ii]]$children[[2]]$gp = gpar(col = "grey", lwd = 1)
        gbl[[ii]]$children[[3]]$gp = gpar(col = "black", lwd = 1)
        class(gbl[[ii]]) = setdiff(class(gbl[[ii]]), "grob_sfc_sequence")
        gbl[[ii]]$vp$width = (k+2)*size
        gbl[[ii]]$vp$height = (k+2)*size

        if(i*2 <= n) {
            ii = ii + 1
            gbl[[ii]] = sfc_grob(pl[[i*2]], x = 4 + k+1 + k/2, y = k/2 - (i-1)*(k+2), default.units = "native", extend = TRUE)
            gbl[[ii]]$children[[1]]$gp = gpar(col = "grey", lwd = 1)
            gbl[[ii]]$children[[2]]$gp = gpar(col = "grey", lwd = 1)
            gbl[[ii]]$children[[3]]$gp = gpar(col = "black", lwd = 1)
            class(gbl[[ii]]) = setdiff(class(gbl[[ii]]), "grob_sfc_sequence")
            gbl[[ii]]$vp$width = (k+2)*size
            gbl[[ii]]$vp$height = (k+2)*size
        }
    }

    ii = length(gbl)
    for(i in 1:n) {
        ii  = ii + 1
        gbl[[ii]] = grob_math(tex_pattern(bp, i, pl[[i]]), x = 6 + ifelse(n == 1, 1, 2)*k, k - 0.5 - (i-1)*1.2, just = "left", default.units = "native")
    }

    gbl[[ii+1]] = rectGrob(gp = gpar(fill = NA, col = "black"), x = 0, just = "left")

    arg = gbl
    arg$vp = vp
    
    do.call(grobTree, arg)
}

#' Draw the expansion rules
#' 
#' @rdname draw_rules
#' @details
#' The expansion rules define how the curve is expanded from level-0 to level-1.
#' @export
#' @return No value is returned.
#' @examples
#' draw_rules_2x2()
draw_rules_2x2 = function() {

    p = sfc_2x2("I")
    grid.newpage()

    rules = p@rules@rules
    equation_max_width = max(do.call("unit.c", lapply(names(rules), function(nm) {
        do.call("unit.c", lapply(seq_along(rules[[nm]]), function(i) {
            convertWidth(grobWidth(grob_math(tex_pattern(nm, i,  rules[[nm]][[i]]), x = 0, y = 0)), "mm")
        }))
    })))

    gb1 = grob_single_base_rule(p, "I", equation_max_width = equation_max_width, x = size, y = unit(1, "npc") - size, just = c("left", "top"))
    grid.draw(gb1)

    gb2 = grob_single_base_rule(p, "R", equation_max_width = equation_max_width, x = size, y = unit(1, "npc") - size - gb1$vp$height, just = c("left", "top"))
    grid.draw(gb2)

    gb3 = grob_single_base_rule(p, "L", equation_max_width = equation_max_width, x = size, y = unit(1, "npc") - size - gb1$vp$height - gb2$vp$height, just = c("left", "top"))
    grid.draw(gb3)

    gb4 = grob_single_base_rule(p, "U", equation_max_width = equation_max_width, x = size, y = unit(1, "npc") - size - gb1$vp$height - gb2$vp$height - gb3$vp$height, just = c("left", "top"))
    grid.draw(gb4)

    gb5 = grob_single_base_rule(p, "B", equation_max_width = equation_max_width, x = size + gb1$vp$width + size, y = unit(1, "npc") - size, just = c("left", "top"))
    grid.draw(gb5)

    gb6 = grob_single_base_rule(p, "D", equation_max_width = equation_max_width, x = size + gb1$vp$width + size, y = unit(1, "npc") - size - gb5$vp$height, just = c("left", "top"))
    grid.draw(gb6)

    gb7 = grob_single_base_rule(p, "P", equation_max_width = equation_max_width, x = size + gb1$vp$width + size, y = unit(1, "npc") - size - gb5$vp$height - gb6$vp$height, just = c("left", "top"))
    grid.draw(gb7)

    gb8 = grob_single_base_rule(p, "Q", equation_max_width = equation_max_width, x = size + gb1$vp$width + size, y = unit(1, "npc") - size - gb5$vp$height - gb6$vp$height - gb7$vp$height, just = c("left", "top"))
    grid.draw(gb8)

    gb9 = grob_single_base_rule(p, "C", equation_max_width = equation_max_width, x = size + gb1$vp$width + size, y = unit(1, "npc") - size - gb5$vp$height - gb6$vp$height - gb7$vp$height - gb8$vp$height, just = c("left", "top"))
    grid.draw(gb9)
}

#' @rdname draw_rules
#' @param flip Whether to use the "flipped" rules? For the Peano curve and the Meander curve, there is also a "fliiped" version 
#'      of the curve expansion rules. See the vignettes for details.
#' @export
#' @examples
#' # the units in the main rules of the Peano curve are vertical
#' draw_rules_3x3_peano()
#' # the units in the flipped rules of the Peano curve are horizontal
#' draw_rules_3x3_peano(flip = TRUE)
draw_rules_3x3_peano = function(flip = FALSE) {

    p = sfc_3x3_peano("I", flip = flip)

    grid.newpage()

    if(flip) {
        rules = p@rules@flip
    } else {
        rules = p@rules@rules
    }
    equation_max_width = max(do.call("unit.c", lapply(names(rules), function(nm) {
        do.call("unit.c", lapply(seq_along(rules[[nm]]), function(i) {
            convertWidth(grobWidth(grob_math(tex_pattern(nm, i,  rules[[nm]][[i]]), x = 0, y = 0)), "mm")
        }))
    })))

    gb1 = grob_single_base_rule(p, "I", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size, just = c("left", "top"))
    nc = length(gb1$children)
    gb1$children[[nc]]$width = gb1$children[[nc]]$width
    grid.draw(gb1)

    gb2 = grob_single_base_rule(p, "J", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size - gb1$vp$height, just = c("left", "top"))
    nc = length(gb2$children)
    gb2$children[[nc]]$width = gb2$children[[nc]]$width
    grid.draw(gb2)

    gb3 = grob_single_base_rule(p, "R", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size - gb1$vp$height - gb2$vp$height, just = c("left", "top"))
    nc = length(gb3$children)
    gb3$children[[nc]]$width = gb3$children[[nc]]$width
    grid.draw(gb3)

    gb4 = grob_single_base_rule(p, "L", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size - gb1$vp$height - gb2$vp$height - gb3$vp$height, just = c("left", "top"))
    nc = length(gb4$children)
    gb4$children[[nc]]$width = gb4$children[[nc]]$width
    grid.draw(gb4)
}

#' @rdname draw_rules
#' @export
#' @examples
#' # the units in the main rules of the Meander curve are "forward"
#' # i.e. the direction of the "wave" is the same as the direction of the curve
#' draw_rules_3x3_meander()
#' # the units in the flipped rules of the Meander curve are "backward"
#' draw_rules_3x3_meander(flip = TRUE)
draw_rules_3x3_meander = function(flip = FALSE) {

    p = sfc_3x3_meander("I", flip = flip)

    grid.newpage()

    if(flip) {
        rules = p@rules@flip
    } else {
        rules = p@rules@rules
    }
    equation_max_width = max(do.call("unit.c", lapply(names(rules), function(nm) {
        do.call("unit.c", lapply(seq_along(rules[[nm]]), function(i) {
            convertWidth(grobWidth(grob_math(tex_pattern(nm, i,  rules[[nm]][[i]]), x = 0, y = 0)), "mm")
        }))
    })))

    gb1 = grob_single_base_rule(p, "I", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size, just = c("left", "top"))
    nc = length(gb1$children)
    gb1$children[[nc]]$width = gb1$children[[nc]]$width
    grid.draw(gb1)

    gb2 = grob_single_base_rule(p, "R", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size - gb1$vp$height, just = c("left", "top"))
    nc = length(gb2$children)
    gb2$children[[nc]]$width = gb2$children[[nc]]$width
    grid.draw(gb2)

    gb3 = grob_single_base_rule(p, "L", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size - gb1$vp$height - gb2$vp$height, just = c("left", "top"))
    nc = length(gb3$children)
    gb3$children[[nc]]$width = gb3$children[[nc]]$width
    grid.draw(gb3)

    gb4 = grob_single_base_rule(p, "U", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size - gb1$vp$height - gb2$vp$height - gb3$vp$height, just = c("left", "top"))
    nc = length(gb4$children)
    gb4$children[[nc]]$width = gb4$children[[nc]]$width
    grid.draw(gb4)

    gb5 = grob_single_base_rule(p, "B", equation_max_width = equation_max_width, flip = flip, x = size + gb1$vp$width + size, y = unit(1, "npc") - size, just = c("left", "top"))
    nc = length(gb5$children)
    gb5$children[[nc]]$width = gb5$children[[nc]]$width
    grid.draw(gb5)

    gb6 = grob_single_base_rule(p, "D", equation_max_width = equation_max_width, flip = flip, x = size + gb1$vp$width + size, y = unit(1, "npc") - size - gb5$vp$height, just = c("left", "top"))
    nc = length(gb6$children)
    gb6$children[[nc]]$width = gb6$children[[nc]]$width
    grid.draw(gb6)

    gb7 = grob_single_base_rule(p, "P", equation_max_width = equation_max_width, flip = flip, x = size + gb1$vp$width + size, y = unit(1, "npc") - size - gb5$vp$height - gb6$vp$height, just = c("left", "top"))
    nc = length(gb7$children)
    gb7$children[[nc]]$width = gb7$children[[nc]]$width
    grid.draw(gb7)

    gb8 = grob_single_base_rule(p, "Q", equation_max_width = equation_max_width, flip = flip, x = size + gb1$vp$width + size, y = unit(1, "npc") - size - gb5$vp$height - gb6$vp$height - gb7$vp$height, just = c("left", "top"))
    nc = length(gb8$children)
    gb8$children[[nc]]$width = gb8$children[[nc]]$width
    grid.draw(gb8)

    gb9 = grob_single_base_rule(p, "C", equation_max_width = equation_max_width, flip = flip, x = size + gb1$vp$width + size, y = unit(1, "npc") - size - gb5$vp$height - gb6$vp$height - gb7$vp$height - gb8$vp$height, just = c("left", "top"))
    nc = length(gb9$children)
    gb9$children[[nc]]$width = gb9$children[[nc]]$width
    grid.draw(gb9)

}

######

#' All traverse paths of a sequence
#' @rdname traverse_path
#' @param rules An `sfc_rules` object.
#' @param p An `sfc_sequence` sequence. `p` and `rules` should have the same universe base set.
#'        Please provide `p` as a small sequence because the total number of all traverse paths might be very huge.
#' @export
#' @details
#' Given an input sequence with rotations, `all_traverse_paths()` lists all combinations of expansion 
#' codes from the first letter to the last letter in `p` (i.e. all possible traverse paths).
#' @return
#' `all_traverse_paths()` returns a list of integer vectors.
all_traverse_paths = function(rules, p) {

    if(!identical(sfc_universe(rules), sfc_universe(p))) {
        stop_wrap("Universe set of `rules` and `p` should be identical.")
    }

    rules = rules@rules
    seq = p@seq
    rot = p@rot
    n = length(seq)

    pl = list()
    prev_code = seq_along(rules[[ as.character(seq[1]) ]])
    if(n == 1) {
        return(lapply(prev_code, function(x) x))
    }

    for(i in seq_along(seq)[-1]) {

        # pick current code accroding to prev_code
        link = list(prev = integer(0), curr = integer(0))
        prev_code2 = NULL
        for(pc in prev_code) {
            prev_corner = rules[[ as.character(seq[i-1]) ]][[pc]]@corner
            prev_corner = adjust_corner(prev_corner, rot[i-1])
            for(j in seq_along(rules[[ as.character(seq[i]) ]])) {
                curr_corner = rules[[ as.character(seq[i]) ]][[j]]@corner
                curr_corner = adjust_corner(curr_corner, rot[i])
                
                if(prev_corner[2] != curr_corner[[1]]) {
                    link$prev = c(link$prev, paste0(i-1, "-", pc))
                    link$curr = c(link$curr, paste0(i, "-", j))
                    prev_code2 = c(prev_code2, j)
                }
                
            }
        }
        prev_code2 = unique(prev_code2)
        if(length(prev_code2) == 0) {
            stop_wrap("Sequence cannot expand completely at `p[", i, "]` (letter = \"", seq[i], "\", rot = ", rot[i], ").")
        }
        prev_code = prev_code2
        link = cbind(link$prev, link$curr)
        pl[[i-1]] = link
    }
    m = do.call(rbind, pl)
    g = igraph::graph_from_edgelist(m)

    e = which(igraph::degree(g, mode = "out") == 0 & grepl(paste0("^", n, "-"), igraph::V(g)$name))
    s = which(grepl(paste0("^1-"), igraph::V(g)$name))

    do.call("c", lapply(s, function(x) {
        lapply(igraph::all_shortest_paths(g, x, to = e)$vpath, function(pp) {
            as.integer(gsub("^\\d+-", "", names(pp)))
        })
    }))
}

adjust_corner = function(corner, rot) {
    if(rot == 90 || rot == 270) {
        rev_corner(corner)
    } else {
        corner
    }
}

rev_corner = function(corner) {
    ifelse(corner == 1L, 2L, 1L)
}

#' @rdname traverse_path
#' @export
#' @details
#' `get_one_traverse_path()` returns one random traverse path.
#' @return
#' `get_one_traverse_path()` returns an integer vector.
#' @examples
#' # expansion rules for the general 3x3 curves
#' p = SFC_RULES_3x3_COMBINED@rules$I[[3]]
#' get_one_traverse_path(SFC_RULES_3x3_COMBINED, p)
#' get_one_traverse_path(SFC_RULES_3x3_COMBINED, p)
#' get_one_traverse_path(SFC_RULES_3x3_COMBINED, p)
#' get_one_traverse_path(SFC_RULES_3x3_COMBINED, p)
get_one_traverse_path = function(rules, p) {

    if(!identical(sfc_universe(rules), sfc_universe(p))) {
        stop_wrap("Universe set of `rules` and `p` should be identical.")
    }

    rules = rules@rules
    seq = p@seq
    rot = p@rot
    n = length(seq)

    pl = list()
    prev_code = sample(length(rules[[ as.character(seq[1]) ]]), 1)
    if(n == 1) {
        return(prev_code)
    }

    for(i in seq_along(seq)[-1]) {

        # pick current code accroding to prev_code
        link = list(prev = integer(0), curr = integer(0))
        prev_code2 = NULL
        
        pc = prev_code
        prev_corner = rules[[ as.character(seq[i-1]) ]][[pc]]@corner
        prev_corner = adjust_corner(prev_corner, rot[i-1])
        for(j in seq_along(rules[[ as.character(seq[i]) ]])) {
            curr_corner = rules[[ as.character(seq[i]) ]][[j]]@corner
            curr_corner = adjust_corner(curr_corner, rot[i])
            
            if(prev_corner[2] != curr_corner[[1]]) {
                link$prev = c(link$prev, paste0(i-1, "-", pc))
                link$curr = c(link$curr, paste0(i, "-", j))
                prev_code2 = c(prev_code2, j)
            }
            
        }
        
        if(length(prev_code2) == 0) {
            return(integer(0))
        }

        kk = sample(length(prev_code2), 1)

        prev_code = prev_code2[kk]
        link = cbind(link$prev[kk], link$curr[kk])
        pl[[i-1]] = link
    }
    pp = c(pl[[1]][[1]], sapply(pl, function(x) x[2]))
    as.integer(gsub("^\\d+-", "", pp))
}

#' @rdname traverse_path
#' @param type If the value is `"11|22"`, it highlights the paths only via 1-1/2-2 corners. If the value is `"12|21"`, it highlights the paths
#'       only via 1-2/2-1 corners.
#' @export
#' @examples
#' # 
#' p = SFC_RULES_3x3_COMBINED@rules$I[[3]]
#' plot_traverse_paths(SFC_RULES_3x3_COMBINED, p)
#' plot_traverse_paths(SFC_RULES_3x3_COMBINED, p, type = "11|22")
#' plot_traverse_paths(SFC_RULES_3x3_COMBINED, p, type = "12|21")
#' 
#' # 2x2 curve
#' p = sfc_2x2("I", 11)
#' plot_traverse_paths(SFC_RULES_2x2, p)
#' 
#' # Peano curve
#' p = sfc_3x3_peano("I", 1)
#' plot_traverse_paths(SFC_RULES_3x3_PEANO, p)
#' 
#' # Meander curve
#' p = sfc_3x3_meander("I", 1)
#' plot_traverse_paths(SFC_RULES_3x3_MEANDER, p)
plot_traverse_paths = function(rules, p, type = c("all", "11|22", "12|21")) {

    if(!identical(sfc_universe(rules), sfc_universe(p))) {
        stop_wrap("Universe set of `rules` and `p` should be identical.")
    }

    paths = all_traverse_paths(rules, p)

    rules = rules@rules
    seq = p@seq
    rot = p@rot
    n = length(seq)

    maxr = max(sapply(rules, length))

    grid.newpage()
    pushViewport(viewport(xscale = c(1, max(2, n)), yscale = c(1, max(2, maxr)), x = unit(25, "mm"), width = unit(1, "npc") - unit(40, "mm"), y = unit(25, "mm"), height = unit(1, "npc") - unit(35, "mm"), just = c("left", "bottom")))
    grid.text(1:maxr, unit(1, "native") - unit(10, "mm"), seq(1, maxr), default.units = "native", gp = gpar(fontsize = 10))
    grid.text("Expansion code", unit(1, "native") - unit(18, "mm"), (1+maxr)/2, default.units = "native", rot = 90)
    # grid.text(paste0(seq, "(", rot, ")"), 1:n, unit(1, "native") - unit(14, "mm"), default.units = "native", gp = gpar(fontsize = 10))
    for(i in 1:n) {
        grid.draw(grob_math(paste0("italic(", seq[i], ")^", rot[i]), i, unit(1, "native") - unit(14, "mm"), default.units = "native", gp = gpar(fontsize = 12)))
    }
    
    type = match.arg(type)
    l2 = lapply(paths, function(path) {
        l = rep(FALSE, length(path))
        for(i in seq_along(path)) {
            corner = adjust_corner( rules[[ seq[i] ]][[ path[i] ]]@corner, rot[i])
            if(type == "11|22") {
                if(corner[1] == corner[2]) {
                    l[i] = TRUE
                }
            }else if(type == "12|21") {
                if(corner[1] != corner[2]) {
                    l[i] = TRUE
                }
            } else {
                l[i] = TRUE
            }
        }
        l
    })
    l2 = sapply(l2, all)

    for(i in order(l2)) {
        path = paths[[i]]
        np = length(path)
        pos = cbind(1:(np-1), path[-np], 2:np, path[-1])
        theta = atan( (pos[, 4] - pos[, 2])/(pos[, 3] - pos[, 1]) )
        len_x = 2/convertHeight(unit(1, "npc"), "mm", valueOnly = TRUE)*(maxr-1)
        len_y = len_x
        pos[, 1] = pos[, 1] + len_x*cos(theta)
        pos[, 2] = pos[, 2] + len_y*sin(theta)
        pos[, 3] = pos[, 3] - len_x*cos(theta)
        pos[, 4] = pos[, 4] - len_y*sin(theta)
        col = ifelse(l2[i], "black", "#DDDDDD")
        grid.segments(pos[, 1], pos[, 2], pos[, 3], pos[, 4], default.units = "native", gp = gpar(col = col, fill = col), arrow = arrow(length = unit(6, "pt"), angle = 15, type = "closed"))
    }

    for(i in seq_len(n)) {
        k = length(rules[[ seq[i] ]])
        for(j in 1:k) {
            corner = adjust_corner( rules[[ seq[i] ]][[j]]@corner, rot[i])
            col = "#DDDDDD"
            pch = 1
            if(type == "11|22") {
                if(corner[1] == corner[2]) {
                    col = "black"
                    pch = 16
                }
            } else if(type == "12|21") {
                if(corner[1] != corner[2]) {
                    col = "black"
                    pch = 16
                }
            } else {
                col = "black"
                pch = 16
            }
            grid.points(i, j, default.units = "native", pch = pch, size = unit(8, "pt"), gp = gpar(col = col))
            grid.text(paste0("(", corner[1], ",", corner[2], ")"), i, unit(j, "native")-unit(4, "mm"), default.units = "native", gp = gpar(fontsize = 8, col = col))
        }
    }
    popViewport()
}

#' Generate a nxn curve based on expansion rules
#' 
#' @param rules An `sfc_rules` object.
#' @param name Name of the curve. The name will be used for the functions that will be generated.
#' @param envir Environment where the functions are exported.
#' @param flippable Whether `rules` can have flipped version? If it is `TRUE`, the generated function also accepts the `flip` argument.
#' @param verbose Whether to print messages?
#' 
#' @details
#' Two functions will be exported:
#' 
#' - `sfc_\{name\}()`
#' - `draw_rules_\{name\}()`
#' 
#' For the simplicity, flipping is not supported yet.
#' 
#' @export
#' @return No value is returned.
#' @examples
#' UNIVERSE_4x4_PEANO = c("I", "R", "L")
#'
#' RULES_4x4_PEANO = list()
#' RULES_4x4_PEANO[["I"]][[1]] = sfc_unit("RIILLIIRRIILLIIR", rot = 0, universe = UNIVERSE_4x4_PEANO)
#' RULES_4x4_PEANO[["I"]][[2]] = sfc_hflip(RULES_4x4_PEANO[["I"]][[1]])
#' RULES_4x4_PEANO[["R"]][[1]] = sfc_unit("IIIRRIILLIIRRIIL", rot = 0, universe = UNIVERSE_4x4_PEANO)      
#' RULES_4x4_PEANO[["R"]][[2]] = sfc_rotate(sfc_unit("LIIRRIILLIIRRIII", 
#'                                 rot = 270, universe = UNIVERSE_4x4_PEANO), 90)
#' RULES_4x4_PEANO[["L"]][[1]] = sfc_hflip(RULES_4x4_PEANO[["R"]][[2]])
#' RULES_4x4_PEANO[["L"]][[2]] = sfc_hflip(RULES_4x4_PEANO[["R"]][[1]])
#'
#' SFC_RULES_4x4_PEANO = sfc_rules(rules = RULES_4x4_PEANO,
#'         name = "Peano 4x4",
#'         bases = BASE_LIST[UNIVERSE_4x4_PEANO])
#' 
#' sfc_generator(SFC_RULES_4x4_PEANO, "4x4_peano")
#' draw_rules_4x4_peano()
#' sfc_4x4_peano("I", 111) |> plot()
sfc_generator = function(rules, name, envir = topenv(parent.frame()), 
	flippable = FALSE, verbose = TRUE) {

	RULES = rules
	NAME = name

	if(!(identical(sort(names(RULES@rules)), c("I", "L", "R")) || identical(sort(names(RULES@rules)), c("I", "J", "L", "R")))) {
		stop_wrap("`rules` should only contain base patterns I/J/R/L.")
	}

	if(any(sapply(RULES@rules, length) > 2)) {
		stop_wrap("`rules` should only contain one or two traverse codes for a single base pattern.")
	}

	for(nm in names(RULES@rules)) {
		for(i in seq_along(RULES@rules[[nm]])) {
			pa = get_one_traverse_path(RULES, RULES@rules[[nm]][[i]])
			if(length(pa) == 0) {
				stop_wrap("Cannot find a complete traverse path for rule ", nm, "_", i)
			}
		}
	}

	code = as.vector(unique(t(do.call(cbind, lapply(RULES@rules, function(x) sapply(x, function(y) sort(y@corner)))))))
	if(!(identical(code, c(1L, 2L)) || identical(code, c(1L, 2L, 1L, 2L)) || identical(code, c(2L, 1L, 2L, 1L)))) {
		stop_wrap("`rules` can only contain 11|22 or 12|21 level-1 units.")
	}

	if(identical(code, c(1L, 2L))) {
		code_num = 2
	} else {
		code_num = 1
	}

	if(code_num == 2) {
		for(nm in names(RULES@rules)) {
			if(!identical(RULES@rules[[nm]][[1]]@corner, c(1L, 2L))) {
				stop_wrap("The first pattern should have corner value of [1, 2].")
			}
		}
	}

	if(flippable) {
		flip = lapply(RULES@rules, function(x) {
            lapply(x, function(u) sfc_flip_unit(u, RULES@bases))
        })

        if(identical(RULES@rules[[1]][[1]], flip[[1]][[1]])) {
        	if(verbose) {
        		message("The flipped rules are identical to the original rules. Flipping will not be supported.")
        	}
        } else {
        	RULES@flip = flip
        }
	}

	cl = paste0("sfc_", name)

	setClass(cl,
		contains = "sfc_nxn",
		prototype = list(mode = sfc_mode(RULES)), where = envir)

	sfc_fun = function(seed, code = integer(0), rot = 0L, flip = FALSE) {

		code = .parse_code(code, 1:2)

		if(inherits(seed, "character")) {
			seed = sfc_seed(seed, rot = rot, universe = sfc_universe(RULES))
		} else if(inherits(seed, "sfc_seed")) {
			seed@seq = factor(as.vector(seed@seq), levels = sfc_universe(RULES))
		} else {
			seed = sfc_seed(seq = seed@seq, rot = seed@rot)
		}

		p = as(seed, cl)
		p@seed = seed
		p@level = 0L
		p@mode = sfc_mode(RULES)

		if(length(p@rules@flip) == 0) {
			for(i in seq_along(code)) {
				p = sfc_expand(p, code[i])
			}
		} else {
			if(is.logical(flip)) {
				if(!(length(flip) == length(seed) || length(flip) == p@mode^2 || length(flip) == 1)) {
					stop_wrap(paste0("If `flip` is a logical vector, it should have a length the same as `seed` or ", p@mode^2, "\n"))
				}
			}

			if(is.function(flip)) {
				for(i in seq_along(code)) {
					p = sfc_expand(p, code[i], flip = flip(p))
				}
			} else {
				for(i in seq_along(code)) {
				
					if(i == 1) {
						if(length(flip) == length(seed)) {
							p = sfc_expand(p, code[i], flip = flip)
						} else {
							p = sfc_expand(p, code[i], flip = flip[1])
						}
					} else if (i == 2) {
						if(length(flip) == p@mode^2) {

						} else {
							flip = rep(flip, each = p@mode^2)
						}
						
						p = sfc_expand(p, code[i], flip = flip)
					} else {
						flip = rep(flip, each = p@mode^2)
						p = sfc_expand(p, code[i], flip = flip)
					}
				}
			}
		}
		
		
		p@expansion = as.integer(code)
		p@universe = sfc_universe(RULES)
		p
		
	}

	setAs("sfc_sequence", cl, function(from) {
		p = new(cl)
		p@seq = from@seq
		levels(p@seq) = sfc_universe(RULES)
		if(any(is.na(p@seq))) {
			stop_wrap("Base letters should all be in `sfc_universe(RULES)`.")
		}
		p@rot = from@rot
		p@universe = sfc_universe(RULES)
		p@level = 0L
		p@mode = sfc_mode(RULES)
		p@rules = RULES

		p
	}, where = envir)

	setMethod("sfc_expand",
		signature = cl,
		definition = function(p, code, flip = FALSE) {

		seq = p@seq
		rot = p@rot
		n = length(p@seq)

		rules = p@rules
		if(length(rules@flip) == 0) {
			flip = FALSE
		}

		tl = integer(n)

		if(code_num == 2) {
			tl[1] = code
			if(n > 1) {
				for(i in 2:n) {
					tl[i] = traverse_type_2x2(tl[i-1], rot[i-1], rot[i])
				}
			}
		} else {
			tl = rep(1L, n)
		}

		sfc_expand_by_rules(rules, p, code = tl, flip = flip)

	}, where = envir)

	draw_rules = function(flip = FALSE) {

	    p = sfc_fun("I", flip = flip)

	    if(length(p@rules@flip) == 0) {
	    	flip = FALSE
	    }

	    grid.newpage()

	    if(flip) {
	        rules = p@rules@flip
	    } else {
	        rules = p@rules@rules
	    }
	    equation_max_width = max(do.call("unit.c", lapply(names(rules), function(nm) {
	        do.call("unit.c", lapply(seq_along(rules[[nm]]), function(i) {
	            convertWidth(grobWidth(grob_math(tex_pattern(nm, i,  rules[[nm]][[i]]), x = 0, y = 0)), "mm")
	        }))
	    })))

	    gb1 = grob_single_base_rule(p, "I", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size, just = c("left", "top"))
	    nc = length(gb1$children)
	    gb1$children[[nc]]$width = gb1$children[[nc]]$width
	    grid.draw(gb1)

	    if("J" %in% sfc_universe(p)) {
	    	gb2 = grob_single_base_rule(p, "J", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size - gb1$vp$height, just = c("left", "top"))
		    nc = length(gb2$children)
		    gb2$children[[nc]]$width = gb2$children[[nc]]$width
		    grid.draw(gb2)

		    gb3 = grob_single_base_rule(p, "R", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size - gb1$vp$height - gb2$vp$height, just = c("left", "top"))
		    nc = length(gb3$children)
		    gb3$children[[nc]]$width = gb3$children[[nc]]$width
		    grid.draw(gb3)

		    gb4 = grob_single_base_rule(p, "L", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size - gb1$vp$height - gb2$vp$height - gb3$vp$height, just = c("left", "top"))
		    nc = length(gb4$children)
		    gb4$children[[nc]]$width = gb4$children[[nc]]$width
		    grid.draw(gb4)
	    } else {

		    gb2 = grob_single_base_rule(p, "R", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size - gb1$vp$height, just = c("left", "top"))
		    nc = length(gb2$children)
		    gb2$children[[nc]]$width = gb2$children[[nc]]$width
		    grid.draw(gb2)

		    gb3 = grob_single_base_rule(p, "L", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size - gb1$vp$height - gb2$vp$height, just = c("left", "top"))
		    nc = length(gb3$children)
		    gb3$children[[nc]]$width = gb3$children[[nc]]$width
		    grid.draw(gb3)
		}

	}

	assign(paste0("sfc_", NAME), sfc_fun, envir = envir)
	assign(paste0("draw_rules_", NAME), draw_rules, envir = envir)
	if(verbose) {
		cat("The following two functions are exported to the current top environment:\n", sep = "")
		cat("  - ", paste0("sfc_", NAME), "()\n", sep = "")
		cat("  - ", paste0("draw_rules_", NAME), "()\n", sep = "")

		if(length(RULES@flip) > 0) {
			cat("")
			cat("flipping is supported.\n")
		}
	}

}

#' Transformations of a sequence
#' 
#' @param p,e1 An `sfc_sequence` object.
#' @param rot,e2 Rotation measured in the polar coordinate system, in degrees.
#' 
#' @rdname sfc_transformation
#' @aliases sfc_rotate
#' 
#' @details
#' - `sfc_rotate()` and `^()` rotate each base pattern.
#' - `sfc_hflip()` flips a sequence horizontally.
#' - `sfc_vflip()` flips a sequence vertically.
#' - `sfc_dflip()` flips a sequence against a diagonal line (with slop 1 or -1).
#' @export
#' @return An `sfc_sequence` object.
setMethod("sfc_rotate",
	signature = "sfc_sequence",
	definition = function(p, rot) {

	p@rot = (p@rot + as.integer(rot)) %% 360L
	p
})


#' @rdname sfc_transformation
#' @export
`^.sfc_sequence` = function(e1, e2) {
	sfc_rotate(e1, e2)
}

validate_in_directions = function(p2, in_direction) {
	 
	corner = guess_unit_corner(p2, "first")
	if(corner  == "bottomleft") {
		if(in_direction == 180) {
			in_direction = in_direction + 90
		} else if(in_direction == 270) {
			in_direction = in_direction - 90
		}
	} else if(corner == "bottomright") {
		if(in_direction == 0) {
			in_direction = in_direction - 90
		} else if(in_direction == 270) {
			in_direction = in_direction + 90
		}
	} else if(corner == "topleft") {
		if(in_direction == 90) {
			in_direction = in_direction + 90
		} else if(in_direction == 180) {
			in_direction = in_direction - 90
		}
	} else if(corner == "topright") {
		if(in_direction == 0) {
			in_direction = in_direction + 90
		} else if(in_direction == 90) {
			in_direction = in_direction - 90
		}
	}
	in_direction %% 360
}

validate_out_directions = function(p2, out_direction) {
	 
	corner = guess_unit_corner(p2, "last")
	if(corner  == "bottomleft") {
		if(out_direction == 0) {
			out_direction = out_direction + 90
		} else if(out_direction == 90) {
			out_direction = out_direction - 90
		}
	} else if(corner == "bottomright") {
		if(out_direction == 180) {
			out_direction = out_direction - 90
		} else if(out_direction == 90) {
			out_direction = out_direction + 90
		}
	} else if(corner == "topleft") {
		if(out_direction == 270) {
			out_direction = out_direction + 90
		} else if(out_direction == 0) {
			out_direction = out_direction - 90
		}
	} else if(corner == "topright") {
		if(out_direction == 180) {
			out_direction = out_direction + 90
		} else if(out_direction == 270) {
			out_direction = out_direction - 90
		}
	}
	out_direction %% 360
}


#' @rdname sfc_transformation
#' @aliases sfc_hflip
#' @param fix_ends By default, the curve is flipped as a complete whole, which means, the associated entry and exit directions of the curve
#'       is also adjusted accordingly. When flipping subunits in a curve, e.g. level-1 subunits in a Peano curve, we want the 
#'       entry and exit direction of the subunit not changed so that the subunits are still connected in the curve after the flipping. In this 
#'       case, `fix_ends` can be set to `TRUE`, then only the subunits are flipped while the connections to neighbouring subunits are not affected. 
#'       See the **Examples** section.
#' @param bases A list of base patterns, consider to use [`BASE_LIST`]. It is only used when `fix_ends = TRUE`.
#' @export
#' @examples
#' p = sfc_3x3_meander("R", 2, rot = -90)
#' draw_multiple_curves(
#'     p, 
#'     sfc_hflip(p), 
#'     sfc_hflip(p, fix_ends = TRUE), 
#'     nrow = 1)
setMethod("sfc_hflip",
	signature = "sfc_sequence",
	definition = function(p, fix_ends = FALSE, bases = NULL) {

	if(fix_ends) {
		if(is.null(bases)) {
			if(!inherits(p, "sfc_nxn")) {
				stop_wrap("`fix_ends = TURE` only works on the sfc_nxn objects.")
			}
			bases = p@rules@bases
		}

		p2 = sfc_hflip(sfc_reverse(p))

		in_direction = base_in_direction(bases[[ as.character(p@seq[1]) ]], p@rot[1])
		out_direction = base_out_direction(bases[[ as.character(p2@seq[1]) ]], p2@rot[1])
		if(abs(in_direction - out_direction) == 180) {
			in_direction = validate_in_directions(p2, in_direction)
		}
		lt = guess_base_pattern(bases, in_direction, out_direction)
		p2@seq[1] = lt[[1]]
		p2@rot[1] = as.integer(round(lt[[2]]))
		
		n = length(p)
		in_direction = base_in_direction(bases[[ as.character(p2@seq[n]) ]], p2@rot[n])
		out_direction = base_out_direction(bases[[ as.character(p@seq[n]) ]], p@rot[n])
		if(abs(in_direction - out_direction) == 180) {
			out_direction = validate_out_directions(p2, out_direction)
		}
		lt = guess_base_pattern(bases, in_direction, out_direction)
		p2@seq[n] = lt[[1]]
		p2@rot[n] = as.integer(round(lt[[2]]))

		if("J" %in% sfc_universe(p)) {
			if(p2@seq[2] == "I" && p2@seq[1] == "I") {
				p2@seq[1] = "J"
			} else if(p2@seq[2] == "J" && p2@seq[1] == "J") {
				p2@seq[1] = "I"
			}

			if(p2@seq[n-1] == "I" && p2@seq[n] == "I") {
				p2@seq[n] = "J"
			} else if(p2@seq[n-1] == "J" && p2@seq[n] == "J") {
				p2@seq[n] = "I"
			}
		}

		p2
	} else {
		p2 = p
		p2@seq[p@seq == "L"] = "R"
		p2@seq[p@seq == "R"] = "L"

		if("J" %in% sfc_universe(p)) {
			p2@seq[p@seq == "I"] = "J"
			p2@seq[p@seq == "J"] = "I"
		}

		p2@rot[p@rot == 90] = 270L
		p2@rot[p@rot == 270] = 90L

		p2@rot = p2@rot %% 360L

		if(fix_ends) {
			n = length(p)
			p2@rot[c(1, n)] = p@rot[c(1, n)]
		}
		p2
	}
})


#' @rdname sfc_transformation
#' @aliases sfc_vflip
#' @export
#' @examples
#' p = sfc_3x3_meander("L", 2, rot = -90)
#' draw_multiple_curves(
#'     p, 
#'     sfc_vflip(p), 
#'     sfc_vflip(p, fix_ends = TRUE), 
#'     nrow = 1)
setMethod("sfc_vflip",
	signature = "sfc_sequence",
	definition = function(p, fix_ends = FALSE, bases = NULL) {

	if(fix_ends) {
		if(is.null(bases)) {
			if(!inherits(p, "sfc_nxn")) {
				stop_wrap("`fix_ends = TURE` only works on the sfc_nxn objects.")
			}
			bases = p@rules@bases
		}

		p2 = sfc_vflip(sfc_reverse(p))

		in_direction = base_in_direction(bases[[ as.character(p@seq[1]) ]], p@rot[1])
		out_direction = base_out_direction(bases[[ as.character(p2@seq[1]) ]], p2@rot[1])
		if(abs(in_direction - out_direction) == 180) {
			in_direction = validate_in_directions(p2, in_direction)
		}
		lt = guess_base_pattern(bases, in_direction, out_direction)
		p2@seq[1] = lt[[1]]
		p2@rot[1] = as.integer(round(lt[[2]]))

		n = length(p)
		in_direction = base_in_direction(bases[[ as.character(p2@seq[n]) ]], p2@rot[n])
		out_direction = base_out_direction(bases[[ as.character(p@seq[n]) ]], p@rot[n])
		if(abs(in_direction - out_direction) == 180) {
			out_direction = validate_out_directions(p2, out_direction)
		}
		lt = guess_base_pattern(bases, in_direction, out_direction)
		p2@seq[n] = lt[[1]]
		p2@rot[n] = as.integer(round(lt[[2]]))

		if("J" %in% sfc_universe(p)) {
			if(p2@seq[2] == "I" && p2@seq[1] == "I") {
				p2@seq[1] = "J"
			} else if(p2@seq[2] == "J" && p2@seq[1] == "J") {
				p2@seq[1] = "I"
			}

			if(p2@seq[n-1] == "I" && p2@seq[n] == "I") {
				p2@seq[n] = "J"
			} else if(p2@seq[n-1] == "J" && p2@seq[n] == "J") {
				p2@seq[n] = "I"
			}
		}

		p2
	} else {
		p2 = sfc_hflip(sfc_rotate(p, 180L))

		if(fix_ends) {
			n = length(p)
			p2@rot[c(1, n)] = p@rot[c(1, n)]
		}
		p2
	}
})

#' @rdname sfc_transformation
#' @aliases sfc_dflip
#' 
#' @param slop Slop of the diagonal. Value can only be 1 or -1.
#' @export
#' @examples
#' p = sfc_3x3_peano("I", 2)
#' draw_multiple_curves(
#'     p, 
#'     sfc_dflip(p, 1), 
#'     sfc_dflip(p, 1, fix_ends = TRUE), 
#'     nrow = 1)
setMethod("sfc_dflip",
	signature = "sfc_sequence",
	definition = function(p, slop = 1L, fix_ends = FALSE, bases = NULL) {

	slop = as.integer(slop)
	if(fix_ends) {
		if(is.null(bases)) {
			if(!inherits(p, "sfc_nxn")) {
				stop_wrap("`fix_ends = TURE` only works on the sfc_nxn objects.")
			}
			bases = p@rules@bases
		}

		p2 = sfc_dflip(p, slop = slop)

		in_direction = base_in_direction(bases[[ as.character(p@seq[1]) ]], p@rot[1])
		out_direction = base_out_direction(bases[[ as.character(p2@seq[1]) ]], p2@rot[1])
		lt = guess_base_pattern(bases, in_direction, out_direction)
		p2@seq[1] = lt[[1]]
		p2@rot[1] = as.integer(round(lt[[2]]))

		n = length(p)
		in_direction = base_in_direction(bases[[ as.character(p2@seq[n]) ]], p2@rot[n])
		out_direction = base_out_direction(bases[[ as.character(p@seq[n]) ]], p@rot[n])
		lt = guess_base_pattern(bases, in_direction, out_direction)
		p2@seq[n] = lt[[1]]
		p2@rot[n] = as.integer(round(lt[[2]]))

		if("J" %in% sfc_universe(p)) {
			if(p2@seq[2] == "I" && p2@seq[1] == "I") {
				p2@seq[1] = "J"
			} else if(p2@seq[2] == "J" && p2@seq[1] == "J") {
				p2@seq[1] = "I"
			}

			if(p2@seq[n-1] == "I" && p2@seq[n] == "I") {
				p2@seq[n] = "J"
			} else if(p2@seq[n-1] == "J" && p2@seq[n] == "J") {
				p2@seq[n] = "I"
			}
		}

		p2
	} else {
		if(slop == 1L) {
			p2 = sfc_rotate(sfc_hflip(p), -90L)
		} else {
			p2 = sfc_rotate(sfc_hflip(p), 90L)
		}

		if(fix_ends) {
			n = length(p)
			p2@rot[c(1, n)] = p@rot[c(1, n)]
		}
		p2
	}
})


#' @rdname sfc_transformation
#' @aliases sfc_reverse
#' 
#' @export
setMethod("sfc_reverse",
	signature = "sfc_sequence", 
	definition = function(p) {
	p2 = p
	p2@seq = rev(p2@seq)
	p2@rot = rev(p2@rot)

	l1 = p2@seq == "I" | p2@seq == "J"
	if(any(l1)) {
		p2@rot[l1] = p2@rot[l1] + 180L
	}

	l2 = p2@seq == "R"
	l3 = p2@seq == "L"

	if(any(l2)) {
		p2@seq[l2] = "L"
		p2@rot[l2] = p2@rot[l2] + 90L
	}
	if(any(l3)) {
		p2@seq[l3] = "R"
		p2@rot[l3] = p2@rot[l3] - 90L
	}

	p2@rot = p2@rot %% 360L
	p2
})

#' @rdname sfc_transformation
#' @param x An `sfc_sequence` object.
#' @export
rev.sfc_sequence = function(x) {
	sfc_reverse(x)
}

#' @rdname sfc_index
#' @param x An `sfc_nxn` object.
#' @param i,index A string of digits representing the path on the hierarchy of the curve. The left side
#'       corresponds to the top level and the right side corresponds to the bottom level on the curve. For the
#'       2x2 curve, the digits can only be 1-4, and for the Peano and Meander curves, the digites can be 1-9.
#'       The hierarchical index should be specified in a format of `i1:i2:i3:...`
#'         where `:` can be replaced by any non-number character. For 2x2 and 3x3 curves, `:` can be omitted and the
#'         hierarchical index can be specified as `i1i2i3...`. See the **Examples** section.
#' @param j A value of `TRUE` or `FALSE` that controls whether to keep the `sfc_nxn` class or degenerate to the `sfc_sequence` class.
#' @param ... Ignore.
#' @param drop A value of `TRUE` or `FALSE` that controls whether to keep the `sfc_nxn` class or degenerate to the `sfc_sequence` class.
#' @export
#' @return
#' `sfc_index()` returns an integer vector.
#' @examples
#' p = sfc_2x2("I", "11111")
#' p["3:2:1"]
#' # for 2x2 and 3x3 curves, ":" can be omitted
#' p["321"]
#' p["3:2:1", TRUE] # or p["3:2:1", drop = FALSE]
`[.sfc_nxn` = function(x, i, j, ..., drop = TRUE) {

	if(missing(i)) {
		return(x)
	}
	if(length(i) == 0) {
		return(x)
	}

	if(missing(j)) {
		if(drop) {
			`[.sfc_sequence`(x, sfc_index(x, i))
		} else {
			x2 = x
			ind = sfc_index(x, i)
			x2@seq = x@seq[ind]
			x2@rot = x@rot[ind]
			x2
		}
	} else {
		if(!is.logical(j)) {
			stop_wrap("The second index should be TURE or FALSE.")
		}
		if(j) {
			x2 = x
			ind = sfc_index(x, i)
			x2@seq = x@seq[ind]
			x2@rot = x@rot[ind]
			x2
		} else {
			`[.sfc_sequence`(x, sfc_index(x, i))
		}
	}
}

get_index_from_nxn = function(index, level, n) {
	if(grepl("\\D", index)) {
		index = as.integer(strsplit(index, "\\D+")[[1]])
		index = index[!is.na(index)]
	} else {
		if(n <= 3) {
			index = as.integer(strsplit(as.character(index), "")[[1]])
		}
	}
	index = as.integer(index)

	ind = seq_len( (n^2)^level )
	for(i in index) {
		if(is.na(i)) {
			stop_wrap("It should only contain digits.")
		}
		if(i > n^2) {
			stop_wrap( paste0("Each digit in `index` should be smaller than ", n^2, ".") )
		}
		if(i < 1) {
			stop_wrap("Each digit in `index` should be larger than zero.")
		}
		portion = length(ind)/(n^2)

		if(portion < 1) {
			stop_wrap( paste0("Depth in `index` should be not larger than ", level, ".") )
		}
		ind = ind[seq( (i-1)*portion+1, i*portion )]
	}
	ind
}

#' Subunit in the curve
#' @aliases sfc_index
#' @rdname sfc_index
#' @param p An `sfc_nxn` object.
#' @export
#' @details
#' `sfc_index()` only works on square curves (i.e. a curve with a single base letter as seed.)
#' @examples
#' # only for testing
#' p = sfc_2x2("I", "11111")
#' om = par(no.readonly = TRUE)
#' par(mfrow = c(2, 2))
#' test_sfc_index(p, "3")
#' test_sfc_index(p, "3:2")
#' test_sfc_index(p, "3:2:1")
#' test_sfc_index(p, "3:2:1:1")
#' par(om)
#' 
#' p = sfc_3x3_meander("I", "11111")
#' om = par(no.readonly = TRUE)
#' par(mfrow = c(2, 2))
#' test_sfc_index(p, "7")
#' test_sfc_index(p, "7:5")
#' test_sfc_index(p, "7:5:9")
#' test_sfc_index(p, "7:5:9:2")
#' par(om)
setMethod("sfc_index",
	signature = "sfc_nxn",
	definition = function(p, index = "") {

	if(length(p@seed) != 1) {
		stop_wrap("`sfc_index()` only works on curves with a single base letter as seed.")
	}

	if(is.numeric(index)) {
		return(index)
	} else if(length(index) == 0) {
		return(seq_len(length(p)))
	} else if(identical(index, "")) {
		return(seq_len(length(p)))
	} else {
		get_index_from_nxn(as.character(index), p@level, p@mode)
	}
})

#' @rdname sfc_index
#' @import graphics
#' @export
#' @details
#' `test_sfc_index()` is a helper function for demonstrating `sfc_index()`.
test_sfc_index = function(p, index) {
	loc = sfc_segments(p)

	om = par("mar")
	on.exit(par(mar = om))
	par(mar = c(1, 1, 4, 1))
	plot(loc, type = "l", col = "grey", ann = FALSE, axes = FALSE, asp = 1)
	txt = index
	index = sfc_index(p, index)

	if(length(index) == 1) {
		points(loc[index, 1], loc[index, 2], pch = 16, col = "black")
	} else {
		lines(loc[index, 1], loc[index, 2], lwd = 2, col = "black")
	}
	title(paste0(class(p), ": level = ", p@level, ", ", p@mode, "x", p@mode, ", index = ", txt))
}

rotate_90 = function(pos) {
	n = as.integer(round(sqrt(nrow(pos))))
	pos2 = pos
	pos2[, 1] = -pos[, 2] + n - 1
	pos2[, 2] = pos[, 1]
	pos2
}

#' H-curve
#' @rdname h_curve
#' @param h The seed of the H-curve. The value should be one of [`H0`], [`H1`] or [`H2`].
#' @param iteration Number of iterations.
#' @param connect How the four subunits are connected to form the H-curve on the next level. See **Details**.
#' @param random Whether to generate subunits randomly on each iteration.
#' @details
#' An H-curve on level k is composed with four subunits on level k-1. If we number the four subunits in the following order:
#' 
#' ```
#' 2  3
#' 1  4
#' ```
#' 
#' where subunit 1 connects to subunit 2, subunit 2 connects to subunit 3, et al., and e.g. subunit 1 connects to subunit 2 
#' via its toprigth corner. Since H-curve can be thought of as a closed curve, to, e.g. let subunit 1 to connect to  
#' subunit 2, its topright corner needs to be opened. There are two segments on subunit 1 that can be removed/opened: the horizontal
#' segment and the vertical segment on the topright corner in subunit 1.
#' 
#' In this way, in `sfc_h()`, the argument `connect` only accepts a single value of `"h"` or `"v"` where the types of segments for 
#' all the four subunits are the same, i.e. whether all the horizontal corner segments are opened or whether all the vertical corner
#' segments are opened. In `expand_h()`, the argument `connect` can be set to a string with four letters or a vector of length four where the type of segments of the
#' four subunits can be set separately.
#' 
#' In the random mode, each subunit is generated randomly, the type of the open segment is choosen randomly, also each subunit has a probability of 0.5
#' to rotate by 90 degrees.
#' 
#' @return
#' A two-column matrix of coordinates of points on the curve.
#' @export
#' @importFrom stats runif
#' @examples
#' draw_multiple_curves(
#'     sfc_h(H0, iteration = 2),
#'     sfc_h(H2, iteration = 2),
#'     closed = TRUE, nrow =1
#' )
#' draw_multiple_curves(
#'     sfc_h(H1, iteration = 3, random = TRUE),
#'     sfc_h(H1, iteration = 3, random = TRUE),
#'     closed = TRUE, nrow = 1
#' )
sfc_h = function(h, iteration = 1, connect = c("h", "v"), random = FALSE) {

	if(!(identical(h, H0) || identical(h, H1) || identical(h, H2))) {
		stop_wrap("`h` can only take values from the three pre-defined objects: `H0/H1/H2`.")
	}

	if(random) {
		hl = list(h, h, h, h)
		cl = c("h", "v")

		for(i in seq_len(iteration)) {

			
			for(i in 1:4) {
				if(runif(1) > 0.5) {
					hl[[i]] = rotate_90(hl[[i]])
				}
			}
			h1 = expand_h(hl[[sample(4, 1)]], hl[[sample(4, 1)]], hl[[sample(4, 1)]], hl[[sample(4, 1)]], connect = sample(cl, 1, replace = TRUE))
			
			for(i in 1:4) {
				if(runif(1) > 0.5) {
					hl[[i]] = rotate_90(hl[[i]])
				}
			}
			h2 = expand_h(hl[[sample(4, 1)]], hl[[sample(4, 1)]], hl[[sample(4, 1)]], hl[[sample(4, 1)]], connect = sample(cl, 1, replace = TRUE))
			
			for(i in 1:4) {
				if(runif(1) > 0.5) {
					hl[[i]] = rotate_90(hl[[i]])
				}
			}
			h3 = expand_h(hl[[sample(4, 1)]], hl[[sample(4, 1)]], hl[[sample(4, 1)]], hl[[sample(4, 1)]], connect = sample(cl, 1, replace = TRUE))
			
			for(i in 1:4) {
				if(runif(1) > 0.5) {
					hl[[i]] = rotate_90(hl[[i]])
				}
			}
			h4 = expand_h(hl[[sample(4, 1)]], hl[[sample(4, 1)]], hl[[sample(4, 1)]], hl[[sample(4, 1)]], connect = sample(cl, 1, replace = TRUE))

			hl = list(h1, h2, h3, h4)
		}

		hl[[sample(4, 1)]]

	} else {
		connect = match.arg(connect)
		for(i in seq_len(iteration)) {
			h = expand_h(h, connect = connect)
		}
		h
	}
}

#' @rdname h_curve
#' @param h1 The first subunit on the bottom left.
#' @param h2 The second subunit on the top left.
#' @param h3 The third subunit on the top right.
#' @param h4 The fourth subunit on the bottom right.
#' @export
#' @examples
#' draw_multiple_curves(
#'     expand_h(H0, connect = "hvvh"),
#'     expand_h(H1, connect = "vvhh"),
#'     closed = TRUE, nrow = 1
#' )
#' 
#' # set the four subunits separately
#' h1 = expand_h(H0, connect = "hhhh")
#' h2 = expand_h(H0, connect = "vvvv")
#' h3 = expand_h(H0, connect = "hvhv")
#' h4 = expand_h(H0, connect = "hvvh")
#' expand_h(h1, h2, h3, h4, connect = "vhvh") |> 
#'     plot_segments(closed = TRUE)
#' 
#' fun = function(h, iteration) {
#'     for(i in 1:iteration) h = expand_h(h, connect = "vhvh")
#'     h
#' }
#' fun(H0, 4) |> plot_segments(closed = TRUE)
expand_h = function(h1, h2 = h1, h3 = h1, h4 = h1, connect = "hhhh") {

	n = as.integer(round(sqrt(nrow(h1))))

	
	if(length(connect) == 1) {
		connect = strsplit(connect, "")[[1]]
		if(length(connect) == 1) {
			connect = rep(connect, 4)
		}
	}
	if(length(connect) != 4) {
		stop_wrap("`connect` should be a string or a vector with four letters.")
	}

	h1 = open_h(h1, 3, connect[1])
	h2 = open_h(h2, 4, connect[2])
	h3 = open_h(h3, 1, connect[3])
	h4 = open_h(h4, 2, connect[4])

	h2[, 2] = h2[, 2] + n

	h3[, 1] = h3[, 1] + n
	h3[, 2] = h3[, 2] + n

	h4[, 1] = h4[, 1] + n

	rbind(h1, h2, h3, h4)
}


# h is a two-column matrix, coordinate clockwise, starts from center, the lower segment
# it returns a list of points with clockwise direction
#. 2 3
#. 1 4
open_h = function(h, where, how) {
	min_x = min(h[, 1])
	max_x = max(h[, 1])
	min_y = min(h[, 2])
	max_y = max(h[, 2])

	n = nrow(h)

	if(where == 1) {
		i = which(h[, 1] == min_x & h[, 2] == min_y)
		if(how == "h") {
			part1 = h[seq_len(i-1), , drop = FALSE]
			part2 = h[seq(i, n), , drop = FALSE]
		} else {
			part1 = h[seq(1, i), , drop = FALSE]
			part2 = h[i+seq_len(n-i), , drop = FALSE]
		}
		h2 = rbind(part2, part1)
	} else if(where == 2) {
		i = which(h[, 1] == min_x & h[, 2] == max_y)
		if(how == "h") {
			part1 = h[seq(1, i), , drop = FALSE]
			part2 = h[i+seq_len(n-i), , drop = FALSE]
		} else {
			part1 = h[seq_len(i-1), , drop = FALSE]
			part2 = h[seq(i, n), , drop = FALSE]
		}
		h2 = rbind(part2, part1)
	} else if(where == 3) {
		i = which(h[, 1] == max_x & h[, 2] == max_y)
		if(how == "h") {
			part1 = h[seq_len(i-1), , drop = FALSE]
			part2 = h[seq(i, n), , drop = FALSE]
		} else {
			part1 = h[seq(1, i), , drop = FALSE]
			part2 = h[i+seq_len(n-i), , drop = FALSE]
		}
		h2 = rbind(part2, part1)
	} else if(where == 4) {
		i = which(h[, 1] == max_x & h[, 2] == min_y)
		if(how == "h") {
			part1 = h[seq(1, i), , drop = FALSE]
			part2 = h[i+seq_len(n-i), , drop = FALSE]
		} else {
			part1 = h[seq_len(i-1), , drop = FALSE]
			part2 = h[seq(i, n), , drop = FALSE]
		}
		h2 = rbind(part2, part1)
	}

	h2
}




H0 = cbind(c(0L, 0L, 1L, 1L), 
	       c(0L, 1L, 1L, 0L))

H1 = expand_h(H0, connect = "h")
H2 = expand_h(H0, connect = "v")

#' Seed sequences of the H-curve
#' @rdname h_seed
#' @export
#' @details
#' The three objects simply contain coordinates of points on the three base H-curves.
#' @return Two-column matrices.
#' @examples
#' H0
#' draw_multiple_curves(H0, H1, H2, nrow = 1, closed = TRUE)
"H0"

#' @rdname h_seed
#' @export
"H1"

#' @rdname h_seed
#' @export
"H2"

#' Draw multiple curves
#' 
#' @param ... A list of `sfc_sequence` objects or objects in its child classes, a list of [`grid::grob`] objects 
#'          or a list of two-column coordinate matrices, i.e., all the forms that can represent curves in this package.
#' @param list The list of curve object can also be directly specified as a "list" object.
#' @param nrow Number of rows in the layout.
#' @param ncol Number of columns in the layout.
#' @param extend Whether to draw the entry and exit segments? It is only used when input is a list of `sfc_sequence` objects.
#' @param title Whether to add titles on each panel? The title is constructed in the form of `initial_seed|expansion_codes`, e.g. `I|111`.
#'   The value can be a vector of user-defined strings.
#' @param closed Whether the curves are closed? The value should be a logical vector. If it is `TRUE`, the last point
#'           is connected to the first point in the curve to form a closed curve. Length of `closed` can be 1 or the number of curves.
#' @param padding Space around each curve. The value should be a [`grid::unit`] object with length 1.
#' @param lwd Line width with length 1.
#' @param col Color for the segments with length 1. If the value is `NULL`, it uses the "Spectral" color palettes from the **RColorBrewer** package.
#' 
#' @details
#' This function is used for quick comparison on curves.
#' @return No value is returned.
#' @export
#' @examples
#' # for all forms of curves initialized by base pattern 'R', with rotation 0, and on level 3
#' draw_multiple_curves(
#'     sfc_2x2("R", "111"),
#'     sfc_2x2("R", "112"),
#'     sfc_2x2("R", "121"),
#'     sfc_2x2("R", "122"),
#'     sfc_2x2("R", "211"),
#'     sfc_2x2("R", "212"),
#'     sfc_2x2("R", "221"),
#'     sfc_2x2("R", "222"),
#'     nrow = 2, title = TRUE)
#' 
#' # simply a list of sequences
#' # note they only contain I/R/L, so the base patterns I/R/L are internally used
#' draw_multiple_curves(
#'     sfc_sequence("IIII"),
#'     sfc_sequence("RRRR"),
#'     sfc_sequence("RRLL"),
#'     nrow = 1
#' )
draw_multiple_curves = function(..., list = NULL, nrow = NULL, ncol = NULL, extend = TRUE, 
	title = FALSE, closed = FALSE, padding = unit(0, "pt"),
	lwd = 4, col = NULL) {

	pl = list(...)
	if(length(pl) == 1) {
		if(is.list(pl[[1]])) {
			pl = pl[[1]]
		}
	}
	if(length(list)) {
		pl = c(pl, list)
	}

	n = length(pl)
	if(n < 1) {
		stop_wrap("No curve is specified.")
	}

	if(is.null(nrow) && is.null(ncol)) {
		nrow = floor(sqrt(n))
	}
	if(!is.null(nrow) && is.null(ncol)) {
		ncol = ceiling(n/nrow)
	} else if(is.null(nrow) && !is.null(ncol)) {
		nrow = ceiling(n/ncol)
	} else if(!is.null(nrow) && !is.null(ncol)) {
		ncol = ceiling(n/nrow)
	}

	if(length(title) == 1) {
		title = rep(title, n)
	}
	if(length(closed) == 1) {
		closed = rep(closed, n)
	}
	
	gbl = lapply(seq_along(pl), function(i) {
		if(inherits(pl[[i]], "sfc_sequence")) {
			sfc_grob(pl[[i]], extend = extend, title = title[i], closed = closed[i], lwd = lwd, col = col)
		} else if(inherits(pl[[i]], "matrix")) {
			sfc_grob(pl[[i]], title = title[i], closed = closed[i], lwd = lwd, col = col)
		} else if(inherits(pl[[i]], "grob")) {
			pl[[i]]$children[[1]]$gp$lwd = lwd
			if(!is.null(col)) {
				pl[[i]]$children[[1]]$gp$col = col
			}
			pl[[i]]
		} else if(inherits(pl[[i]], "sfc_base")) {
			sfc_grob(pl[[i]])
		} else {
			stop_wrap("Wrong data type.")
		}
	})
	
	grid.newpage()
	pushViewport(viewport(layout = grid.layout(nrow = nrow, ncol = ncol)))
	for(i in seq_len(nrow)) {
		for(j in seq_len(ncol)) {
			ind = (i-1)*ncol + j
			if(ind > n) {
				break
			}
			pushViewport(viewport(layout.pos.row = i, layout.pos.col = j))
			pushViewport(viewport(width = unit(1, "npc") - padding, height = unit(1, "npc") - padding))
			grid.draw(gbl[[ind]])
			popViewport()
			popViewport()
		}
	}
	popViewport()
}


rotate_coord = function(x, theta, center = c(0, 0)) {
	if(length(x) == 2) {
		dim(x) = c(1, 2)
	}

	y = x[, 2]
	x = x[, 1]

	x = x - center[1]
	y = y - center[2]

	theta = theta/180*pi
	x2 = x*cos(theta) - y*sin(theta)
	y2 = x*sin(theta) + y*cos(theta)

	x2 = x2 + center[1]
	y2 = y2 + center[2]

	cbind(x2, y2)
}

reverse_coord = function(x) {
	if(length(x) == 2) {
		dim(x) = c(1, 2)
	}
	x[seq(nrow(x), 1), , drop = FALSE]
}

move_coord = function(x, offset) {
	if(length(x) == 2) {
		dim(x) = c(1, 2)
	}
	x[, 1] = x[, 1] + offset[1]
	x[, 2] = x[, 2] + offset[2]
	x
}

hflip_coord = function(x, v = 0) {
	x2 = move_coord(x, c(-v, 0))
	x2[, 1] = -x2[, 1]
	x2 = move_coord(x2, c(v, 0))
	x2
}

dflip_coord = function(x, center = c(0, 0), type = 1) {
	x2 = move_coord(x, -center)
	if(type == 1) {
		x2 = hflip_coord(rotate_coord(x2, 90))
	} else {
		x2 = hflip_coord(rotate_coord(x2, -90))
	}
	x2 = move_coord(x2, center)
	x2
}

get_circular_index = function(ind, n) {
	(ind-1) %% n + 1
}


stop_wrap = function (...) {
    x = paste0(...)
    x = paste(strwrap(x), collapse = "\n")
    stop(x, call. = FALSE)
}

warning_wrap = function (...) {
    x = paste0(...)
    x = paste(strwrap(x), collapse = "\n")
    warning(x, call. = FALSE)
}

message_wrap = function (...) {
    x = paste0(...)
    x = paste(strwrap(x), collapse = "\n")
    message(x)
}

equal_to = function(x, y) {
	all(abs(x - y) < 1e-6)
}


convert_to_child_class = function(x, class) {
	x2 = new(class)
	for(nm in slotNames(x)) {
		slot(x2, nm) = slot(x, nm)
	}
	x2
}

in_range = function(x, range, others = NULL) {
	if(is.null(others)) {
		(x >= range[1] && x <= range[2])
	} else {
		(x >= range[1] && x <= range[2]) || x %in% others
	}
}

library(testthat)

guess_unit_corner = sfcurve:::guess_unit_corner


.rotate = function(x, rot = 0, base = c("bottomleft", "bottomright", "topright", "topleft")) {
	i = which(base == x)
	j = get_circular_index(i+rot/90, 4)
	base[j]
}


test_that("test guess_unit_corner on SFC_RULES_2x2", {
	
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$I[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$I[[1]], rot), which = "last"), .rotate("topleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$I[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$I[[2]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$R[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$R[[1]], rot), which = "last"), .rotate("bottomright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$R[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$R[[2]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$L[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$L[[1]], rot), which = "last"), .rotate("topleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$L[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$L[[2]], rot), which = "last"), .rotate("bottomleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$U[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$U[[1]], rot), which = "last"), .rotate("bottomright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$U[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$U[[2]], rot), which = "last"), .rotate("bottomleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$B[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$B[[1]], rot), which = "last"), .rotate("topleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$B[[2]], rot), which = "first"), .rotate("topright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$B[[2]], rot), which = "last"), .rotate("topleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$D[[1]], rot), which = "first"), .rotate("topleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$D[[1]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$D[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$D[[2]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$P[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$P[[1]], rot), which = "last"), .rotate("topleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$P[[2]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$P[[2]], rot), which = "last"), .rotate("bottomright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$Q[[1]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$Q[[1]], rot), which = "last"), .rotate("bottomleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$Q[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$Q[[2]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$C[[1]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$C[[1]], rot), which = "last"), .rotate("bottomleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$C[[2]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_2x2@rules$C[[2]], rot), which = "last"), .rotate("bottomright", rot))
	}
})

test_that("test guess_unit_corner on SFC_RULES_3x3_PEANO", {
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_PEANO@rules$I[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_PEANO@rules$I[[1]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_PEANO@rules$J[[1]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_PEANO@rules$J[[1]], rot), which = "last"), .rotate("topleft", rot))
	}
	
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_PEANO@rules$R[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_PEANO@rules$R[[1]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_PEANO@rules$L[[1]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_PEANO@rules$L[[1]], rot), which = "last"), .rotate("topleft", rot))
	}
})


test_that("test guess_unit_corner on SFC_RULES_3x3_PEANO, flip", {
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_PEANO@flip$I[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_PEANO@flip$I[[1]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_PEANO@flip$J[[1]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_PEANO@flip$J[[1]], rot), which = "last"), .rotate("topleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_PEANO@flip$R[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_PEANO@flip$R[[1]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_PEANO@flip$L[[1]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_PEANO@flip$L[[1]], rot), which = "last"), .rotate("topleft", rot))
	}

})


test_that("test guess_unit_corner on SFC_RULES_3x3_MEANDER", {
	
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$I[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$I[[1]], rot), which = "last"), .rotate("topleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$I[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$I[[2]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$R[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$R[[1]], rot), which = "last"), .rotate("bottomright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$R[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$R[[2]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$L[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$L[[1]], rot), which = "last"), .rotate("topleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$L[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$L[[2]], rot), which = "last"), .rotate("bottomleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$U[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$U[[1]], rot), which = "last"), .rotate("bottomright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$U[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$U[[2]], rot), which = "last"), .rotate("bottomleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$B[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$B[[1]], rot), which = "last"), .rotate("topleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$B[[2]], rot), which = "first"), .rotate("topright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$B[[2]], rot), which = "last"), .rotate("topleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$D[[1]], rot), which = "first"), .rotate("topleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$D[[1]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$D[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$D[[2]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$P[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$P[[1]], rot), which = "last"), .rotate("bottomright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$P[[2]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$P[[2]], rot), which = "last"), .rotate("topleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$Q[[1]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$Q[[1]], rot), which = "last"), .rotate("bottomleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$Q[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$Q[[2]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$C[[1]], rot), which = "first"), .rotate("topright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$C[[1]], rot), which = "last"), .rotate("bottomright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$C[[2]], rot), which = "first"), .rotate("topleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@rules$C[[2]], rot), which = "last"), .rotate("bottomleft", rot))
	}
})


test_that("test guess_unit_corner on SFC_RULES_3x3_MEANDER, flip", {
	
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$I[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$I[[1]], rot), which = "last"), .rotate("topleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$I[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$I[[2]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$R[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$R[[1]], rot), which = "last"), .rotate("bottomright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$R[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$R[[2]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$L[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$L[[1]], rot), which = "last"), .rotate("topleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$L[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$L[[2]], rot), which = "last"), .rotate("bottomleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$U[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$U[[1]], rot), which = "last"), .rotate("bottomright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$U[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$U[[2]], rot), which = "last"), .rotate("bottomleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$B[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$B[[1]], rot), which = "last"), .rotate("topleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$B[[2]], rot), which = "first"), .rotate("topright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$B[[2]], rot), which = "last"), .rotate("topleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$D[[1]], rot), which = "first"), .rotate("topleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$D[[1]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$D[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$D[[2]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$P[[1]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$P[[1]], rot), which = "last"), .rotate("bottomright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$P[[2]], rot), which = "first"), .rotate("bottomleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$P[[2]], rot), which = "last"), .rotate("topleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$Q[[1]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$Q[[1]], rot), which = "last"), .rotate("bottomleft", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$Q[[2]], rot), which = "first"), .rotate("bottomright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$Q[[2]], rot), which = "last"), .rotate("topright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$C[[1]], rot), which = "first"), .rotate("topright", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$C[[1]], rot), which = "last"), .rotate("bottomright", rot))
	}

	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$C[[2]], rot), which = "first"), .rotate("topleft", rot))
	}
	for(rot in c(0, 90, 180, 270)) {
		expect_equal(guess_unit_corner(sfc_rotate(SFC_RULES_3x3_MEANDER@flip$C[[2]], rot), which = "last"), .rotate("bottomleft", rot))
	}
})

# adjust the segment, start point must be in the bottom-left unit and end point
# must be in the bottom-right unit

#' Shape of the curve
#' @aliases sfc_shape
#' @rdname sfc_shape
#' 
#' @param p An `sfc_2x2` object.
#' @details
#' The shape of the curve is defined as a form of the curve without considering entry/exit directions, rotation, 
#' flipping (reflection) nor reversing.
#' 
#' ## 2x2 curve
#' 
#' The process of selecting the shape segment of the curve denoted as `P` is:
#' 
#' 1. The entry-point should locate in the bottom left subunit and the exit-point should 
#'    locate in the bottom right subunit. We try the four rotations (0, 90, 180, 270), and 
#'    the four rotations on the horizontally flipped curve. Once we find the transformed curve
#'    that satisfies this criterion, we name it as `P2`.
#' 2. We also generate `P3` which is a horizontally flipped version of `rev(P2)`.
#' 3. We compare the first point `p` of `P2` and `P3`, and select the one whose `p` has the smaller
#'    x-coordinate (i.e. more to the left of the curve). If the x-coordinates of `p` are the same in
#'    `P2` and `P3`, we select the one whose `p` has the smaller y-coordinate.
#' 
#' @return
#' `sfc_shape()` returns a two-column data frame of the xy-coordinates of the shape curve.
#' @export
#' @examples
#' p1 = sfc_2x2("I", 11)
#' p2 = sfc_2x2("R", 22)
#' draw_multiple_curves(
#'     p1, p2, 
#'     sfc_shape(p1), sfc_shape(p2), 
#'     col = "black")
setMethod("sfc_shape", 
	signature = "sfc_2x2",
	definition = function(p) {

	loc = sfc_segments(p)
	np = nrow(loc)

	# center of the curve set to (0, 0)
	loc[, 1] = loc[, 1] - mean(loc[, 1])
	loc[, 2] = loc[, 2] - mean(loc[, 2])

	ok = FALSE
	for(theta in c(0, 90, 180, 270)) {
		loc2 = rotate_coord(loc, theta)
		if(loc2[1, 1] < 0 & loc2[1, 2] < 0 & loc2[np, 1] > 0 & loc2[np, 2] < 0) {
			ok = TRUE
			break
		}
	}

	if(!ok) {
		loc = hflip_coord(loc, 0)
		for(theta in c(0, 90, 180, 270)) {
			loc2 = rotate_coord(loc, theta)
			if(loc2[1, 1] < 0 & loc2[1, 2] < 0 & loc2[np, 1] > 0 & loc2[np, 2] < 0) {
				ok = TRUE
				break
			}
		}
	}

	loc_rev = hflip_coord(reverse_coord(loc2))

	if(equal_to(loc2[1, 1], loc_rev[1, 1])) {
		if(loc2[1, 2] <= loc_rev[1, 2]) {
			loc_final = loc2
		} else {
			loc_final = loc_rev
		}
	} else {
		if(loc2[1, 1] < loc_rev[1, 1]) {
			loc_final = loc2
		} else {
			loc_final = loc_rev
		}
	}

	loc_final[, 1] = round(loc_final[, 1] - min(loc_final[, 1]))
	loc_final[, 2] = round(loc_final[, 2] - min(loc_final[, 2]))
	loc_final = as.integer(loc_final)

	dim(loc_final) = c(nrow(loc), 2)
	loc_final
})

#' @rdname sfc_shape
#' @param level Level of the 2x2 curve.
#' @export
#' @return
#' `all_2x2_shapes()` returns a list of `n` two-column data frames where each data frame corresponds to
#' the xy-coordnates of the corresponding shape curve.
#' @examples
#' sl = all_2x2_shapes(2)
#' draw_multiple_curves(list = sl, lwd = 2, col = "black")
#' sl = all_2x2_shapes(3)
#' draw_multiple_curves(list = sl, lwd = 2, col = "black")
all_2x2_shapes = function(level = 2) {

	pl = list()
	for(bp in names(SFC_RULES_2x2@rules)) {
		pl = c(pl, lapply(1:2^level, function(x) {
			sfc_2x2(bp, code = int_to_binary(x-1, level)+1)
		}))
	}

	sl = lapply(pl, sfc_shape)

	hash = sapply(sl, digest::digest)
	l = !duplicated(hash)
	sl[l]
}

uid_3x3_peano = function(loc, mode) {

	v = c("vertical" = 1L, "horizontal" = 2L)

	unit_orientation = function(loc) {
		if(equal_to(loc[1, 1], loc[2, 1]) && equal_to(loc[2, 1], loc[3, 1])) {
	        "vertical"
	    } else {
	        "horizontal"
	    }
	}

	go_next_level = function(loc, res) {
		n = nrow(loc)
		if(n == 9) {
			return(res)
		} else {
			block_size = n/mode^2
			for(i in 1:9) {
				loc2 = loc[seq((i-1)*block_size+1, i*block_size), ]
				loc1 = sfc_reduce(loc2, to = 1)
				res = c(res, v[unit_orientation(loc1)])
				res = go_next_level(loc2, res)
			}
			return(res)
		}
	}

	if(nrow(loc) == 1) {
		return(integer(0))
	}

	if(nrow(loc) == 9) {
		return(v[unit_orientation(loc)])
	}

	loc1 = sfc_reduce(loc, to = 1)
	res = v[unit_orientation(loc1)]
	res = go_next_level(loc, res)

	paste(res, collapse = "")
}

#' @rdname sfc_shape
#' @export
#' @details
#' ## 3x3 Peano curve
#' 
#' The process of selecting the shape segment of the curve denoted as `P` is:
#' 
#' 1. The entry-point should locate in the bottom left subunit and the exit-point should 
#'    locate in the top right subunit. We try the four rotations (0, 90, 180, 270). Once we find the transformed curve
#'    that satisfies this criterion, we name it as `P2`.
#' 2. We also generate `P3` which is a 180 degrees rotation on the reversed `P2`, `P4` which is a diagonal flip with slop of 1 on `P2`
#'    and `P5` which is a diagonal flip with slop of 1 on `P3`.
#' 3. We calculate the "UID" of `P(2-5)` and pick the one with the smallest UID as the final curve.
#' 
#' The UID of a 3x3 Peano curve is based on the hierarchical indices of the units on it. The hierarchy of the
#' Peano curve is traversed in a depth-first manner. On each node, the orientation of the corresponding unit is
#' calculated where vertical is 1 and horizontal is 2. The digits are concatenated into a long string.
#' 
#' @examples
#' p = sfc_3x3_peano("I", 11)
#' draw_multiple_curves(
#'     p, sfc_dflip(p),
#'     sfc_shape(p), sfc_shape(sfc_dflip(p)),
#'     col = "black")
setMethod("sfc_shape", 
	signature = "sfc_3x3_peano",
	definition = function(p) {

	loc = sfc_segments(p)
	sfc_shape_3x3_peano_internal(loc, p@mode)
})

sfc_shape_3x3_peano_internal = function(loc, mode) {
	np = nrow(loc)

	# center of the curve set to (0, 0)
	loc[, 1] = loc[, 1] - mean(loc[, 1])
	loc[, 2] = loc[, 2] - mean(loc[, 2])

	ok = FALSE
	for(theta in c(0, 90, 180, 270)) {
		loc2 = rotate_coord(loc, theta)
		if(loc2[1, 1] < 0 & loc2[1, 2] < 0 & loc2[np, 1] > 0 & loc2[np, 2] > 0) {
			ok = TRUE
			break
		}
	}

	loc_flip1 = rotate_coord(reverse_coord(loc2), 180)
	loc_flip2 = dflip_coord(loc2, type = 1)
	loc_flip3 = dflip_coord(loc_flip1, type = 1)
	
	uid1 = uid_3x3_peano(loc2, mode)
	uid2 = uid_3x3_peano(loc_flip1, mode)
	uid3 = uid_3x3_peano(loc_flip2, mode)
	uid4 = uid_3x3_peano(loc_flip3, mode)

	lt = list(loc2, loc_flip1, loc_flip2, loc_flip3)

	ind = which.min(c(uid1, uid2, uid3, uid4))
	loc_final = lt[[ind]]

	loc_final[, 1] = round(loc_final[, 1] - min(loc_final[, 1]))
	loc_final[, 2] = round(loc_final[, 2] - min(loc_final[, 2]))
	loc_final = as.integer(loc_final)

	dim(loc_final) = c(nrow(loc), 2)
	loc_final
}

#' @rdname sfc_shape
#' @details
#' `all_3x3_peano_shapes()` only calculates all shapes for Peano curve on level 2.
#' @export
#' @examples
#' \donttest{
#' sl = all_3x3_peano_shapes()
#' length(sl)
#' # the first 8 shapes
#' draw_multiple_curves(sl[1:8], col = "black")
#' }
all_3x3_peano_shapes = function(level = 2) {

	pl = list()
	p = c("v", "h")
	for(x in 1:2^9) {
		ind = int_to_binary(x-1, 9)+1
		code = paste(p[ind], collapse = "")
		pl[[code]] = peano_curve(level = 2, pattern = code)
	}

	sl = lapply(pl, sfc_shape_3x3_peano_internal, mode = 3)

	hash = sapply(sl, digest::digest)
	l = !duplicated(hash)
	sl[l]
}




uid_3x3_meander = function(loc, mode) {

	v = c("forward" = 1L, "backward" = 2L)

	unit_orientation = function(loc) {
		if(equal_to(loc[1, 1], loc[2, 1]) && equal_to(loc[2, 1], loc[3, 1])) {
	        "forward"
	    } else {
	        "backward"
	    }
	}

	go_next_level = function(loc, res) {
		n = nrow(loc)
		if(n == 9) {
			return(res)
		} else {
			block_size = n/mode^2
			for(i in 1:9) {
				loc2 = loc[seq((i-1)*block_size+1, i*block_size), ]
				loc1 = sfc_reduce(loc2, to = 1)
				res = c(res, v[unit_orientation(loc1)])
				res = go_next_level(loc2, res)
			}
			return(res)
		}
	}

	if(nrow(loc) == 1) {
		return(integer(0))
	}

	if(nrow(loc) == 9) {
		return(v[unit_orientation(loc)])
	}

	loc1 = sfc_reduce(loc, to = 1)
	res = v[unit_orientation(loc1)]
	res = go_next_level(loc, res)

	paste(res, collapse = "")
}

#' @rdname sfc_shape
#' @export
setMethod("sfc_shape", 
	signature = "sfc_3x3_meander",
	definition = function(p) {

	loc = sfc_segments(p)
	sfc_shape_3x3_meander_internal(loc, p@mode)
	
})

sfc_shape_3x3_meander_internal = function(loc, mode) {
	np = nrow(loc)

	# center of the curve set to (0, 0)
	loc[, 1] = loc[, 1] - mean(loc[, 1])
	loc[, 2] = loc[, 2] - mean(loc[, 2])

	ok = FALSE
	for(theta in c(0, 90, 180, 270)) {
		loc2 = rotate_coord(loc, theta)
		if(loc2[1, 1] < 0 & loc2[1, 2] < 0 & loc2[np, 1] > 0 & loc2[np, 2] < 0) {
			ok = TRUE
			break
		}
	}

	if(!ok) {
		loc = hflip_coord(loc, 0)
		for(theta in c(0, 90, 180, 270)) {
			loc2 = rotate_coord(loc, theta)
			if(loc2[1, 1] < 0 & loc2[1, 2] < 0 & loc2[np, 1] > 0 & loc2[np, 2] < 0) {
				ok = TRUE
				break
			}
		}
	}

	loc_flip1 = rotate_coord(reverse_coord(loc2), 180)
	loc_flip2 = dflip_coord(loc2, type = 1)
	loc_flip3 = dflip_coord(loc_flip1, type = 1)
	
	uid1 = uid_3x3_meander(loc2, mode)
	uid2 = uid_3x3_meander(loc_flip1, mode)
	uid3 = uid_3x3_meander(loc_flip2, mode)
	uid4 = uid_3x3_meander(loc_flip3, mode)

	lt = list(loc2, loc_flip1, loc_flip2, loc_flip3)

	ind = which.min(c(uid1, uid2, uid3, uid4))
	loc_final = lt[[ind]]

	loc_final[, 1] = round(loc_final[, 1] - min(loc_final[, 1]))
	loc_final[, 2] = round(loc_final[, 2] - min(loc_final[, 2]))
	loc_final = as.integer(loc_final)

	dim(loc_final) = c(nrow(loc), 2)
	loc_final
}

#' @rdname sfc_shape
#' @details
#' `all_3x3_meander_shapes()` only considers the Meander curve with all subunits on all levels in forward orientation.
#' @export
#' @examples
#' sl = all_3x3_meander_shapes(2)
#' draw_multiple_curves(list = sl, lwd = 2, col = "black")
all_3x3_meander_shapes = function(level = 2) {

	pl = list()
	for(bp in names(SFC_RULES_3x3_MEANDER@rules)) {
		pl = c(pl, lapply(1:2^level, function(x) {
			sfc_3x3_meander(bp, code = int_to_binary(x-1, level)+1)
		}))
	}

	sl = lapply(pl, sfc_shape)

	hash = sapply(sl, digest::digest)
	l = !duplicated(hash)
	sl[l]
}

r = 0.3
s = 0.6
arrow_length = grid::unit(0.2, "native")
size =  grid::unit(5, "mm")

I_base = function(...) {
    gbl = list()

    gbl[[1]] = linesGrob(c(0, 0), c(-s, s), default.units = "native", gp = gpar(col = "grey"), arrow = arrow(angle = 15, length = arrow_length))
    gbl[[2]] = pointsGrob(0, 0, default.units = "native", pch = 16, size = unit(4, "pt"))

    grobTree(gbl[[1]], gbl[[2]], vp = viewport(xscale = c(-1, 1), yscale = c(-1, 1), width = 2*size, height = 2*size, ...))
}

R_base = function(...) {
    gbl = list()

    gbl[[1]] = linesGrob(c(0, 0, s), c(-s, 0, 0), default.units = "native", gp = gpar(col = "grey"), arrow = arrow(angle = 15, length = arrow_length))
    gbl[[2]] = pointsGrob(0, 0, default.units = "native", pch = 16, size = unit(4, "pt"))

    grobTree(gbl[[1]], gbl[[2]], vp = viewport(xscale = c(-1, 1), yscale = c(-1, 1), width = 2*size, height = 2*size, ...))
}

L_base = function(...) {
    gbl = list()

    gbl[[1]] = linesGrob(c(0, 0, -s), c(-s, 0, 0), default.units = "native", gp = gpar(col = "grey"), arrow = arrow(angle = 15, length = arrow_length))
    gbl[[2]] = pointsGrob(0, 0, default.units = "native", pch = 16, size = unit(4, "pt"))

    grobTree(gbl[[1]], gbl[[2]], vp = viewport(xscale = c(-1, 1), yscale = c(-1, 1), width = 2*size, height = 2*size, ...))
}


U_base = function(...) {
    l = convertWidth(arrow_length, "native", valueOnly = TRUE)*1.2
    gbl = list()

    gbl[[1]] = linesGrob(c(0-0.05, 0-0.05, 0-0.05-l*sin(30/2/180*pi)), c(0, -s, l*cos(30/2/180*pi)-s), default.units = "native", gp = gpar(col = "grey"))
    gbl[[2]] = linesGrob(c(0+0.05, 0+0.05, 0+0.05+l*sin(30/2/180*pi)), c(-s, 0,  - l*cos(30/2/180*pi)), default.units = "native", gp = gpar(col = "grey"))
    gbl[[3]] = pointsGrob(0, 0, default.units = "native", pch = 16, size = unit(4, "pt"))

    grobTree(gbl[[1]], gbl[[2]], gbl[[3]], vp = viewport(xscale = c(-1, 1), yscale = c(-1, 1), width = 2*size, height = 2*size, ...))
}

C_base = function(...) {
    theta = seq(0, 360)/180*pi
    x = cos(theta)*r
    y = sin(theta)*r
    n = length(x)

    gbl = list()
    gbl[[1]] = linesGrob(x, y, default.units = "native", gp = gpar(col = "grey"))
    gbl[[2]] = linesGrob(c(x[n-1]+0.002, x[n]), c(y[n-1], y[n]), default.units = "native", gp = gpar(col = "grey"), arrow = arrow(angle = 30, length = arrow_length, ends = "first"))
    gbl[[3]] = pointsGrob(0, 0, default.units = "native", pch = 16, size = unit(4, "pt"))

    grobTree(gbl[[1]], gbl[[2]], gbl[[3]], vp = viewport(xscale = c(-1, 1), yscale = c(-1, 1), width = 2*size, height = 2*size, ...))
}


B_base = function(...) {
    gbl = list()

    gbl[[1]] = linesGrob(c(0, 0), c(0, s), default.units = "native", gp = gpar(col = "grey"), arrow = arrow(angle = 15, length = arrow_length))
    theta = seq(0, 360)/180*pi
    x = cos(theta)*r
    y = sin(theta)*r
    n = length(x)
    gbl[[2]] = linesGrob(x+r, y, default.units = "native", gp = gpar(col = "grey"))
    gbl[[3]] = pointsGrob(0, 0, default.units = "native", pch = 16, size = unit(4, "pt"))

    grobTree(gbl[[1]], gbl[[2]], gbl[[3]], vp = viewport(xscale = c(-1, 1), yscale = c(-1, 1), width = 2*size, height = 2*size, ...))
}

D_base = function(...) {
    gbl = list()

    gbl[[1]] = linesGrob(c(0, 0), c(0, s), default.units = "native", gp = gpar(col = "grey"), arrow = arrow(angle = 15, length = arrow_length))
    theta = seq(0, 360)/180*pi
    x = cos(theta)*r
    y = sin(theta)*r
    n = length(x)
    gbl[[2]] = linesGrob(x-r, y, default.units = "native", gp = gpar(col = "grey"))
    gbl[[3]] = pointsGrob(0, 0, default.units = "native", pch = 16, size = unit(4, "pt"))

    grobTree(gbl[[1]], gbl[[2]], gbl[[3]], vp = viewport(xscale = c(-1, 1), yscale = c(-1, 1), width = 2*size, height = 2*size, ...))
}


P_base = function(...) {
    gbl = list()

    gbl[[1]] = linesGrob(c(0, 0), c(-s, 0), default.units = "native", gp = gpar(col = "grey"))
    gbl[[2]] = linesGrob(c(0, 0), c(-s, -s/2), default.units = "native", gp = gpar(col = "grey"), arrow = arrow(angle = 15, length = arrow_length))
    theta = seq(0, 360)/180*pi
    x = cos(theta)*r
    y = sin(theta)*r
    n = length(x)
    gbl[[3]] = linesGrob(x+r, y, default.units = "native", gp = gpar(col = "grey"))
    gbl[[4]] = pointsGrob(0, 0, default.units = "native", pch = 16, size = unit(4, "pt"))

    grobTree(gbl[[1]], gbl[[2]], gbl[[3]], gbl[[4]], vp = viewport(xscale = c(-1, 1), yscale = c(-1, 1), width = 2*size, height = 2*size, ...))
}


Q_base = function(...) {
    gbl = list()

    gbl[[1]] = linesGrob(c(0, 0), c(-s, 0), default.units = "native", gp = gpar(col = "grey"))
    gbl[[2]] = linesGrob(c(0, 0), c(-s, -s/2), default.units = "native", gp = gpar(col = "grey"), arrow = arrow(angle = 15, length = arrow_length))
    theta = seq(0, 360)/180*pi
    x = cos(theta)*r
    y = sin(theta)*r
    n = length(x)
    gbl[[3]] = linesGrob(x-r, y, default.units = "native", gp = gpar(col = "grey"))
    gbl[[4]] = pointsGrob(0, 0, default.units = "native", pch = 16, size = unit(4, "pt"))

    grobTree(gbl[[1]], gbl[[2]], gbl[[3]], gbl[[4]], vp = viewport(xscale = c(-1, 1), yscale = c(-1, 1), width = 2*size, height = 2*size, ...))
}


pdf(NULL)
oe = try({
    BASE_I = sfc_base("I", in_direction = 90, out_direction = 90, primary = TRUE, open = TRUE, grob = I_base())
    BASE_J = sfc_base("J", in_direction = 90, out_direction = 90, primary = TRUE, open = TRUE, grob = I_base())
    BASE_R = sfc_base("R", in_direction = 90, out_direction = 0, primary = TRUE, open = TRUE, grob = R_base())
    BASE_L = sfc_base("L", in_direction = 90, out_direction = 180, primary = TRUE, open = TRUE, grob = L_base())
    BASE_U = sfc_base("U", in_direction = 90, out_direction = -90, primary = FALSE, open = TRUE, grob = U_base())
    BASE_B = sfc_base("B", in_direction = NA, out_direction = 90, primary = FALSE, open = FALSE, grob = B_base())
    BASE_D = sfc_base("D", in_direction = NA, out_direction = 90, primary = FALSE, open = FALSE, grob = D_base())
    BASE_P = sfc_base("P", in_direction = 90, out_direction = NA, primary = FALSE, open = FALSE, grob = P_base())
    BASE_Q = sfc_base("Q", in_direction = 90, out_direction = NA, primary = FALSE, open = FALSE, grob = Q_base())
    BASE_C = sfc_base("C", in_direction = NA, out_direction = NA, primary = FALSE, open = FALSE, grob = C_base())
}, silent = TRUE)
dev.off()

if(inherits(oe, "try-error")) {
    stop(oe)
}

BASE_LIST = list(
    "I" = BASE_I,
    "J" = BASE_J,
    "R" = BASE_R,
    "L" = BASE_L,
    "U" = BASE_U,
    "B" = BASE_B,
    "D" = BASE_D,
    "P" = BASE_P,
    "Q" = BASE_Q,
    "C" = BASE_C
)

UNIVERSE_GLOBAL = c("I", "J", "R", "L", "U", "B", "D", "P", "Q", "C")


UNIVERSE_2x2 = c("I", "R", "L", "U", "B", "D", "P", "Q", "C")
RULES_2x2 = list()
RULES_2x2[["I"]][[1]] = sfc_unit(c("R", "L", "L", "R"), rot = 0, universe = UNIVERSE_2x2)
RULES_2x2[["I"]][[2]] = sfc_unit(c("L", "R", "R", "L"), rot = 0, universe = UNIVERSE_2x2)
RULES_2x2[["R"]][[1]] = sfc_unit(c("I", "R", "R", "L"), rot = 0, universe = UNIVERSE_2x2)
RULES_2x2[["R"]][[2]] = sfc_unit(c("L", "R", "R", "I"), rot = 0, universe = UNIVERSE_2x2)
RULES_2x2[["L"]][[1]] = sfc_unit(c("R", "L", "L", "I"), rot = 0, universe = UNIVERSE_2x2)
RULES_2x2[["L"]][[2]] = sfc_unit(c("I", "L", "L", "R"), rot = 0, universe = UNIVERSE_2x2)
RULES_2x2[["U"]][[1]] = sfc_unit(c("I", "R", "R", "I"), rot = 0, universe = UNIVERSE_2x2)
RULES_2x2[["U"]][[2]] = sfc_unit(c("I", "L", "L", "I"), rot = 0, universe = UNIVERSE_2x2)
RULES_2x2[["B"]][[1]] = sfc_unit(c("L", "L", "L", "R"), rot = 180, universe = UNIVERSE_2x2)
RULES_2x2[["B"]][[2]] = sfc_unit(c("R", "R", "R", "I"), rot = 270, universe = UNIVERSE_2x2)
RULES_2x2[["D"]][[1]] = sfc_unit(c("L", "L", "L", "I"), rot = 90, universe = UNIVERSE_2x2)
RULES_2x2[["D"]][[2]] = sfc_unit(c("R", "R", "R", "L"), rot = 180, universe = UNIVERSE_2x2)
RULES_2x2[["P"]][[1]] = sfc_unit(c("R", "L", "L", "L"), rot = 0, universe = UNIVERSE_2x2)
RULES_2x2[["P"]][[2]] = sfc_unit(c("I", "R", "R", "R"), rot = 0, universe = UNIVERSE_2x2)
RULES_2x2[["Q"]][[1]] = sfc_unit(c("I", "L", "L", "L"), rot = 0, universe = UNIVERSE_2x2)
RULES_2x2[["Q"]][[2]] = sfc_unit(c("L", "R", "R", "R"), rot = 0, universe = UNIVERSE_2x2)
RULES_2x2[["C"]][[1]] = sfc_unit(c("L", "L", "L", "L"), rot = 270, universe = UNIVERSE_2x2)
RULES_2x2[["C"]][[2]] = sfc_unit(c("R", "R", "R", "R"), rot = 90, universe = UNIVERSE_2x2)



###=========================================

UNIVERSE_3x3_PEANO = c("I", "J", "R", "L")

RULES_3x3_PEANO = list()
RULES_3x3_PEANO[["I"]][[1]] = sfc_unit("IJRRILLJI", rot = 0, universe = UNIVERSE_3x3_PEANO)  
RULES_3x3_PEANO[["J"]][[1]] = sfc_unit("JILLJRRIJ", rot = 0, universe = UNIVERSE_3x3_PEANO)
RULES_3x3_PEANO[["R"]][[1]] = sfc_unit("IJRRILLJR", rot = 0, universe = UNIVERSE_3x3_PEANO)      
RULES_3x3_PEANO[["L"]][[1]] = sfc_unit("JILLJRRIL", rot = 0, universe = UNIVERSE_3x3_PEANO) 


RULES_3x3_PEANO_FLIP = list()
RULES_3x3_PEANO_FLIP[["I"]][[1]] = sfc_unit("RILLJRRIL", rot = 0, universe = UNIVERSE_3x3_PEANO)  
RULES_3x3_PEANO_FLIP[["J"]][[1]] = sfc_unit("LJRRILLJR", rot = 0, universe = UNIVERSE_3x3_PEANO)
RULES_3x3_PEANO_FLIP[["R"]][[1]] = sfc_unit("RILLJRRIJ", rot = 0, universe = UNIVERSE_3x3_PEANO)      
RULES_3x3_PEANO_FLIP[["L"]][[1]] = sfc_unit("LJRRILLJI", rot = 0, universe = UNIVERSE_3x3_PEANO) 





#===================================

UNIVERSE_3x3_MEANDER = c("I", "R", "L", "U", "B", "D", "P", "Q", "C")

RULES_3x3_MEANDER = list()
RULES_3x3_MEANDER[["I"]][[1]] = sfc_unit("RILILLRRI", rot = 0, universe = UNIVERSE_3x3_MEANDER)  
RULES_3x3_MEANDER[["I"]][[2]] = sfc_unit("LIRIRRLLI", rot = 0, universe = UNIVERSE_3x3_MEANDER)
RULES_3x3_MEANDER[["R"]][[1]] = sfc_unit("IIRIRRLLI", rot = 0, universe = UNIVERSE_3x3_MEANDER)      
RULES_3x3_MEANDER[["R"]][[2]] = sfc_unit("LIRIRRLLR", rot = 0, universe = UNIVERSE_3x3_MEANDER)
RULES_3x3_MEANDER[["L"]][[1]] = sfc_unit("RILILLRRL", rot = 0, universe = UNIVERSE_3x3_MEANDER)                    # 1, 2
RULES_3x3_MEANDER[["L"]][[2]] = sfc_unit("IILILLRRI", rot = 0, universe = UNIVERSE_3x3_MEANDER) 
RULES_3x3_MEANDER[["U"]][[1]] = sfc_unit("IIRIRRLLR", rot = 0, universe = UNIVERSE_3x3_MEANDER)                    # 1, 2
RULES_3x3_MEANDER[["U"]][[2]] = sfc_unit("IILILLRRL", rot = 0, universe = UNIVERSE_3x3_MEANDER) 
RULES_3x3_MEANDER[["B"]][[1]] = sfc_unit("LILILLRRI", rot = 180, universe = UNIVERSE_3x3_MEANDER)                    # 1, 2
RULES_3x3_MEANDER[["B"]][[2]] = sfc_unit("RIRIRRLLR", rot = 270, universe = UNIVERSE_3x3_MEANDER) 
RULES_3x3_MEANDER[["D"]][[1]] = sfc_unit("LILILLRRL", rot = 90, universe = UNIVERSE_3x3_MEANDER)                    # 1, 2
RULES_3x3_MEANDER[["D"]][[2]] = sfc_unit("RIRIRRLLI", rot = 180, universe = UNIVERSE_3x3_MEANDER) 
RULES_3x3_MEANDER[["P"]][[1]] = sfc_unit("IIRIRRLLL", rot = 0, universe = UNIVERSE_3x3_MEANDER)                    # 1, 2
RULES_3x3_MEANDER[["P"]][[2]] = sfc_unit("RILILLRRR", rot = 0, universe = UNIVERSE_3x3_MEANDER) 
RULES_3x3_MEANDER[["Q"]][[1]] = sfc_unit("IILILLRRR", rot = 0, universe = UNIVERSE_3x3_MEANDER)                    # 1, 2
RULES_3x3_MEANDER[["Q"]][[2]] = sfc_unit("LIRIRRLLL", rot = 0, universe = UNIVERSE_3x3_MEANDER) 
RULES_3x3_MEANDER[["C"]][[1]] = sfc_unit("LILILLRRR", rot = 0, universe = UNIVERSE_3x3_MEANDER)                    # 1, 2
RULES_3x3_MEANDER[["C"]][[2]] = sfc_unit("RIRIRRLLL", rot = 0, universe = UNIVERSE_3x3_MEANDER)                    # 1, 2


####

UNIVERSE_3x3_COMBINED = c("I", "R", "L", "U", "B", "D", "P", "Q", "C")
RULES_3x3_COMBINED = list()

RULES_3x3_COMBINED[["I"]][[1]] = sfc_unit("RILILLRRI", rot = 0, universe = UNIVERSE_3x3_COMBINED)  
RULES_3x3_COMBINED[["I"]][[2]] = sfc_unit("LIRIRRLLI", rot = 0, universe = UNIVERSE_3x3_COMBINED)
RULES_3x3_COMBINED[["R"]][[1]] = sfc_unit("IIRIRRLLI", rot = 0, universe = UNIVERSE_3x3_COMBINED)      
RULES_3x3_COMBINED[["R"]][[2]] = sfc_unit("LIRIRRLLR", rot = 0, universe = UNIVERSE_3x3_COMBINED)
RULES_3x3_COMBINED[["L"]][[1]] = sfc_unit("RILILLRRL", rot = 0, universe = UNIVERSE_3x3_COMBINED)                    # 1, 2
RULES_3x3_COMBINED[["L"]][[2]] = sfc_unit("IILILLRRI", rot = 0, universe = UNIVERSE_3x3_COMBINED) 

RULES_3x3_COMBINED[["I"]][[3]] = sfc_unit("IIRRILLII", rot = 0, universe = UNIVERSE_3x3_COMBINED)  
RULES_3x3_COMBINED[["I"]][[4]] = sfc_unit("IILLIRRII", rot = 0, universe = UNIVERSE_3x3_COMBINED)
RULES_3x3_COMBINED[["R"]][[3]] = sfc_unit("IIRRILLIR", rot = 0, universe = UNIVERSE_3x3_COMBINED)      
RULES_3x3_COMBINED[["L"]][[3]] = sfc_unit("IILLIRRIL", rot = 0, universe = UNIVERSE_3x3_COMBINED) 

RULES_3x3_COMBINED[["U"]][[1]] = sfc_unit("IIRIRRLLR", rot = 0, universe = UNIVERSE_3x3_COMBINED)                    # 1, 2
RULES_3x3_COMBINED[["U"]][[2]] = sfc_unit("IILILLRRL", rot = 0, universe = UNIVERSE_3x3_COMBINED) 
RULES_3x3_COMBINED[["B"]][[1]] = sfc_unit("LILILLRRI", rot = 180, universe = UNIVERSE_3x3_COMBINED)  
RULES_3x3_COMBINED[["B"]][[2]] = sfc_unit("RIRIRRLLR", rot = 270, universe = UNIVERSE_3x3_COMBINED)                    # 1, 2
RULES_3x3_COMBINED[["B"]][[3]] = sfc_unit("LILLIRRII", rot = 270, universe = UNIVERSE_3x3_COMBINED)                    # 1, 2

RULES_3x3_COMBINED[["D"]][[1]] = sfc_unit("LILILLRRL", rot = 90, universe = UNIVERSE_3x3_COMBINED)  
RULES_3x3_COMBINED[["D"]][[2]] = sfc_unit("RIRIRRLLI", rot = 180, universe = UNIVERSE_3x3_COMBINED)                   # 1, 2
RULES_3x3_COMBINED[["D"]][[3]] = sfc_unit("RIRRILLII", rot = 90, universe = UNIVERSE_3x3_COMBINED)                    # 1, 2
RULES_3x3_COMBINED[["P"]][[1]] = sfc_unit("IIRIRRLLL", rot = 0, universe = UNIVERSE_3x3_COMBINED)                    # 1, 2
RULES_3x3_COMBINED[["P"]][[2]] = sfc_unit("RILILLRRR", rot = 0, universe = UNIVERSE_3x3_COMBINED)
RULES_3x3_COMBINED[["P"]][[3]] = sfc_unit("IIRRILLIL", rot = 0, universe = UNIVERSE_3x3_COMBINED) 
RULES_3x3_COMBINED[["Q"]][[1]] = sfc_unit("IILILLRRR", rot = 0, universe = UNIVERSE_3x3_COMBINED)                    # 1, 2
RULES_3x3_COMBINED[["Q"]][[2]] = sfc_unit("LIRIRRLLL", rot = 0, universe = UNIVERSE_3x3_COMBINED) 
RULES_3x3_COMBINED[["Q"]][[3]] = sfc_unit("IILLIRRIR", rot = 0, universe = UNIVERSE_3x3_COMBINED) 
RULES_3x3_COMBINED[["C"]][[1]] = sfc_unit("LILILLRRR", rot = 0, universe = UNIVERSE_3x3_COMBINED)                  # 1, 2
RULES_3x3_COMBINED[["C"]][[2]] = sfc_unit("RIRIRRLLL", rot = 0, universe = UNIVERSE_3x3_COMBINED)                    # 1, 2
RULES_3x3_COMBINED[["C"]][[3]] = sfc_unit("RIRRILLIL", rot = 90, universe = UNIVERSE_3x3_COMBINED)                    # 1, 2
RULES_3x3_COMBINED[["C"]][[4]] = sfc_unit("LILLIRRIR", rot = 270, universe = UNIVERSE_3x3_COMBINED)                   # 1, 2


###

.onLoad = function(libname, pkgname) {

    parent = parent.env(environment())
    SFC_RULES_2x2 = sfc_rules(rules = RULES_2x2, name = "2x2",
        bases = BASE_LIST[UNIVERSE_2x2])

    SFC_RULES_3x3_PEANO = sfc_rules(rules = RULES_3x3_PEANO, flip = RULES_3x3_PEANO_FLIP,
        name = "3x3 Peano", bases = BASE_LIST[UNIVERSE_3x3_PEANO])

    RULES_3x3_MEANDER_FLIP = lapply(RULES_3x3_MEANDER, function(x) {
        lapply(x, function(u) sfc_flip_unit(u, BASE_LIST))
    })
    SFC_RULES_3x3_MEANDER = sfc_rules(rules = RULES_3x3_MEANDER, flip = RULES_3x3_MEANDER_FLIP,
        name = "3x3 Meander",
        bases = BASE_LIST[UNIVERSE_3x3_MEANDER])

    RULES_3x3_COMBINED_FLIP = lapply(RULES_3x3_COMBINED, function(x) {
        lapply(x, function(u) sfc_flip_unit(u, BASE_LIST))
    })

    SFC_RULES_3x3_COMBINED = sfc_rules(rules = RULES_3x3_COMBINED, flip = RULES_3x3_COMBINED_FLIP,
        name = "3x3 combined",
        bases = BASE_LIST[UNIVERSE_3x3_COMBINED])


    RULES_4x4_MEANDER_1_FLIP = lapply(RULES_4x4_MEANDER_1, function(x) {
        lapply(x, function(u) sfc_flip_unit(u, BASE_LIST))
    })

    SFC_RULES_4x4_MEANDER_1 = sfc_rules(rules = RULES_4x4_MEANDER_1, flip = RULES_4x4_MEANDER_1_FLIP,
        name = "4x4 Meander type 1",
        bases = BASE_LIST[UNIVERSE_4x4_MEANDER])

    RULES_4x4_MEANDER_2_FLIP = lapply(RULES_4x4_MEANDER_2, function(x) {
        lapply(x, function(u) sfc_flip_unit(u, BASE_LIST))
    })
    SFC_RULES_4x4_MEANDER_2 = sfc_rules(rules = RULES_4x4_MEANDER_2, flip = RULES_4x4_MEANDER_2_FLIP,
        name = "4x4 Meander type 2",
        bases = BASE_LIST[UNIVERSE_4x4_MEANDER])

    assign("SFC_RULES_2x2", SFC_RULES_2x2, envir = parent)
    assign("SFC_RULES_3x3_PEANO", SFC_RULES_3x3_PEANO, envir = parent)
    assign("SFC_RULES_3x3_MEANDER", SFC_RULES_3x3_MEANDER, envir = parent)
    assign("SFC_RULES_3x3_COMBINED", SFC_RULES_3x3_COMBINED, envir = parent)
    assign("SFC_RULES_4x4_MEANDER_1", SFC_RULES_4x4_MEANDER_1, envir = parent)
    assign("SFC_RULES_4x4_MEANDER_2", SFC_RULES_4x4_MEANDER_2, envir = parent)
}

#' Base patterns
#' 
#' A list of pre-defined base patterns. See the **Examples** section.
#' 
#' `BASE_I` and `BASE_J` are identical. They are only used to distinguish the two
#' "going forward" patterns for the level-1 units with 11/22 corner values, i.e. bottom-left to top-right, and bottom-right
#' to top-left.
#' @rdname base_patterns
#' @return `sfc_base` objects.
#' @export
#' @examples
#' BASE_I
#' BASE_J
#' BASE_R
#' BASE_L
#' BASE_U
#' BASE_B
#' BASE_D
#' BASE_P
#' BASE_Q
#' BASE_C
#' draw_multiple_curves(
#'     BASE_I, BASE_J, BASE_R, BASE_L, BASE_U,
#'     BASE_B, BASE_D, BASE_P, BASE_Q, BASE_C,
#'     nrow = 2
#' )
"BASE_I"

#' @rdname base_patterns
#' @export
"BASE_J"

#' @rdname base_patterns
#' @export
"BASE_R"

#' @rdname base_patterns
#' @export
"BASE_L"

#' @rdname base_patterns
#' @export
"BASE_U"

#' @rdname base_patterns
#' @export
"BASE_B"

#' @rdname base_patterns
#' @export
"BASE_D"

#' @rdname base_patterns
#' @export
"BASE_P"

#' @rdname base_patterns
#' @export
"BASE_Q"

#' @rdname base_patterns
#' @export
"BASE_C"

#' @rdname base_patterns
#' @export
"BASE_LIST"


#' Rules
#' 
#' A list of pre-defined expansion rules for different curves.
#' 
#' @details
#' `SFC_RULES_3x3_PEANO`, `SFC_RULES_3x3_MEANDER` and `SFC_RULES_3x3_COMBINED`, `SFC_RULES_MEANDER_4x4_1`, `SFC_RULES_MEANDER_4x4_2` also contain the "flipped" expansion rules.
#' @rdname pre_defined_rules
#' @return `sfc_rules` objects.
#' @export
#' @examples
#' SFC_RULES_2x2
#' SFC_RULES_3x3_PEANO
#' SFC_RULES_3x3_MEANDER
#' SFC_RULES_3x3_COMBINED
#' SFC_RULES_4x4_MEANDER_1
#' SFC_RULES_4x4_MEANDER_2
"SFC_RULES_2x2"

#' @rdname pre_defined_rules
#' @export
"SFC_RULES_3x3_PEANO"

#' @rdname pre_defined_rules
#' @export
"SFC_RULES_3x3_MEANDER"

#' @rdname pre_defined_rules
#' @details
#' `SFC_RULES_3x3_COMBINED` is a combination of `SFC_RULES_3x3_PEANO` and `SFC_RULES_3x3_MEANDER` where
#' in `SFC_RULES_3x3_PEANO`, `J` is replaced by its original pattern `I`.
#' @export
"SFC_RULES_3x3_COMBINED"

#' @rdname pre_defined_rules
#' @export
"SFC_RULES_4x4_MEANDER_1"

#' @rdname pre_defined_rules
#' @details
#' `SFC_RULES_4x4_MEANDER_1` and `SFC_RULES_4x4_MEANDER_2` are extension rules of Meander curves (3x3)
#' on the 4x4 mode It is only for the demonstration purpose, thus only `I/R/L` are supported.
#' @export
"SFC_RULES_4x4_MEANDER_2"

UNIVERSE_4x4_MEANDER = c("I", "R", "L")

RULES_4x4_MEANDER_1 = list()
RULES_4x4_MEANDER_1[["I"]][[1]] = sfc_unit("IIRIRRLLILIILIIR", rot = 0, universe = UNIVERSE_4x4_MEANDER)  
RULES_4x4_MEANDER_1[["I"]][[2]] = sfc_unit("IILILLRRIRIIRIIL", rot = 0, universe = UNIVERSE_4x4_MEANDER)
RULES_4x4_MEANDER_1[["R"]][[1]] = sfc_unit("RILILLRRIRIIRIIL", rot = 0, universe = UNIVERSE_4x4_MEANDER)      
RULES_4x4_MEANDER_1[["R"]][[2]] = sfc_unit("IILILLRRIRIIRIII", rot = 0, universe = UNIVERSE_4x4_MEANDER)
RULES_4x4_MEANDER_1[["L"]][[1]] = sfc_unit("IIRIRRLLILIILIII", rot = 0, universe = UNIVERSE_4x4_MEANDER)
RULES_4x4_MEANDER_1[["L"]][[2]] = sfc_unit("LIRIRRLLILIILIIR", rot = 0, universe = UNIVERSE_4x4_MEANDER) 

RULES_4x4_MEANDER_2 = list()
RULES_4x4_MEANDER_2[["I"]][[1]] = sfc_unit("RIILIILLIRIRRLLR", rot = 0, universe = UNIVERSE_4x4_MEANDER)  
RULES_4x4_MEANDER_2[["I"]][[2]] = sfc_unit("LIIRIIRRILILLRRL", rot = 0, universe = UNIVERSE_4x4_MEANDER)
RULES_4x4_MEANDER_2[["R"]][[1]] = sfc_unit("IIIRIIRRILILLRRL", rot = 0, universe = UNIVERSE_4x4_MEANDER)      
RULES_4x4_MEANDER_2[["R"]][[2]] = sfc_unit("LIIRIIRRILILLRRI", rot = 0, universe = UNIVERSE_4x4_MEANDER)
RULES_4x4_MEANDER_2[["L"]][[1]] = sfc_unit("RIILIILLIRIRRLLI", rot = 0, universe = UNIVERSE_4x4_MEANDER)
RULES_4x4_MEANDER_2[["L"]][[2]] = sfc_unit("IIILIILLIRIRRLLR", rot = 0, universe = UNIVERSE_4x4_MEANDER) 

#' Expand the curve to the next level
#' @aliases sfc_expand
#' @rdname sfc_expand
#' @param p An `sfc_2x2` object or other related objects.
#' @param code Expansion code, a single integer.
#' 
#' @details
#' For the Hilbert curve and Meander curve, as long as the expansion code of the first base pattern in the sequence is determinted, 
#' the expansion codes for other base patterns in the sequence are all determined. For the Peano curve, since
#' there is only one traverse path on any level, `code` is ignored.
#' 
#' These functions are mainly used internally.
#' 
#' @return An object in the same class as the input.
#' @export
#' @examples
#' p = sfc_2x2("I", 11)
#' sfc_expand(p, 2) # I|211
setMethod("sfc_expand",
	signature = "sfc_2x2",
	definition = function(p, code, flip = FALSE) {
    # argument flip is not used
	seq = p@seq
	rot = p@rot
	n = length(p@seq)

	rules = p@rules

	tl = integer(n)
	tl[1] = code

	if(n > 1) {
		for(i in 2:n) {
			# here defines how a single pattern is expanded into four or more patterns
			tl[i] = traverse_type_2x2(tl[i-1], rot[i-1], rot[i])
		}
	}

	sfc_expand_by_rules(rules, p, code = tl, flip = FALSE)

})


# t: previous type code
# r: prevous rotation
# r_next: next rotation
traverse_type_2x2 = function(t, r, r_next) {
	if((r_next - r) %% 180L == 0L) {
		t
	} else {
		c(2, 1)[t]
	}
}


#' @rdname sfc_expand
#' @param flip Whethe to flip level-1 units? The value should be a logical vector of length one or the same as the length of `p`.
#' @export
#' @examples
#' p = sfc_3x3_peano("I", 11)
#' sfc_expand(p, 2) # I|211
setMethod("sfc_expand",
	signature = "sfc_3x3_peano",
	definition = function(p, code = 1, flip = FALSE) {

	seq = p@seq
	rot = p@rot
	n = length(p@seq)

	rules = p@rules

	code = 1

	tl = integer(n)
	tl[1] = code

	if(n > 1) {
		for(i in 2:n) {
			tl[i] = code
		}
	}

	sfc_expand_by_rules(rules, p, code = tl, flip = flip)

})


#' @rdname sfc_expand
#' @export
#' @examples
#' p = sfc_3x3_meander("I", 11)
#' sfc_expand(p, 2) # I|211
setMethod("sfc_expand",
	signature = "sfc_3x3_meander",
	definition = function(p, code, flip = FALSE) {

	seq = p@seq
	rot = p@rot
	n = length(p@seq)

	rules = p@rules

	tl = integer(n)
	tl[1] = code

	if(n > 1) {
		for(i in 2:n) {
			tl[i] = traverse_type_2x2(tl[i-1], rot[i-1], rot[i])
		}
	}
	
	sfc_expand_by_rules(rules, p, code = tl, flip = flip)

})

setClass("sfc_3x3_combined",
	contains = "sfc_nxn",
	prototype = list(mode = 3L))


#' General 3x3 space-filling curves 
#' 
#' @rdname sfc_3x3_combined
#' @param seed The seed sequence. In most cases, the seed sequence is a single base pattern, which can be specified as a single letter, then `rot` controls
#'      the initial rotation of the base pattern. It also supports a sequence with more than one base patterns as the seed sequence. In this case,
#'      it can be specified as a string of more than one base letters, then `rot` can be set to a single rotation scalar which controls the rotation of the
#'      first letter, or a vector with the same length as the number of base letters.
#' @param rot Rotation of the seed sequence, measured in the polar coordinate system, in degrees.
#' @param level Level of the curve. Currently it is restricted to an integer no bigger than 5.
#' @param flip The same setting as in [`sfc_3x3_peano()`] or [`sfc_3x3_meander()`].
#' 
#' @details
#' This type of 3x3 curve uses the combintation of base patterns from both the Peano curve and the Meander curve.
#' On each level, the traverse path is randomly selected.
#' @return
#' `sfc_3x3_combined()` returns an `sfc_3x3_combined` object.
#' @export
#' @examples
#' draw_multiple_curves(
#'     sfc_3x3_combined("I", level = 3),
#'     sfc_3x3_combined("I", level = 3),
#'     sfc_3x3_combined("I", level = 3),
#'     nrow = 1
#' )
sfc_3x3_combined = function(seed, level = 0, rot = 0L, flip = FALSE) {

	if(level > 5) {
		stop_wrap("`level` should be not larger than 5.")
	}

	if(inherits(seed, "character")) {
		seed = sfc_seed(seed, rot = rot, universe = sfc_universe(SFC_RULES_3x3_COMBINED))
	} else if(inherits(seed, "sfc_seed")) {
		levels(seed@seq) = sfc_universe(SFC_RULES_3x3_COMBINED)
	} else {
		seed = sfc_seed(seq = seed@seq, rot = seed@rot)
	}

	p = as(seed, "sfc_3x3_combined")
	p@seed = seed
	p@level = 0L
	p@mode = 3L

	if(is.logical(flip)) {
		if(!(length(flip) == length(seed) || length(flip) == 9 || length(flip) == 1)) {
			stop_wrap("If `flip` is a logical vector, it should have a length the same as `seed` or 9\n")
		}
	}

	code = rep(1, level)
	if(is.function(flip)) {
		for(i in seq_along(code)) {
			p = sfc_expand(p, NULL, flip = flip(p))
		}
	} else {
		for(i in seq_along(code)) {
		
			if(i == 1) {
				if(length(flip) == length(seed)) {
					p = sfc_expand(p, NULL, flip = flip)
				} else {
					p = sfc_expand(p, NULL, flip = flip[1])
				}
			} else if (i == 2) {
				if(length(flip) == 9) {

				} else {
					flip = rep(flip, each = 9)
				}
				
				p = sfc_expand(p, NULL, flip = flip)
			} else {
				flip = rep(flip, each = 9)
				p = sfc_expand(p, NULL, flip = flip)
			}
		}
	}

	p@universe = sfc_universe(seed)
	p
	
}

setAs("sfc_seed", "sfc_3x3_combined", function(from) {
	p = new("sfc_3x3_combined")
	p@seq = from@seq
	levels(p@seq) = sfc_universe(SFC_RULES_3x3_COMBINED)
	if(any(is.na(p@seq))) {
		stop_wrap("Base letters should all be in `sfc_universe(SFC_RULES_3x3_COMBINED)`.")
	}
	p@rot = from@rot
	p@universe = sfc_universe(SFC_RULES_3x3_COMBINED)
	p@level = 0L
	p@mode = 3L
	p@rules = SFC_RULES_3x3_COMBINED

	p
})

#' @rdname sfc_3x3_combined
#' @param p An `sfc_3x3_combined` object.
#' @param code Ignore. The traverse codes are selected randomly.
#' @export
setMethod("sfc_expand",
	signature = "sfc_3x3_combined",
	definition = function(p, code = NULL, flip = FALSE) {

	seq = p@seq
	rot = p@rot
	n = length(p@seq)

	rules = p@rules

	tl = get_one_traverse_path(rules, p)
	if(!is.null(code)) {
		tl = as.integer(code)
	}

	k = 0
	while(length(tl) == 0 && k < 10) {
		tl = get_one_traverse_path(rules, p)
		k = k + 1
	}

	sfc_expand_by_rules(rules, p, code = tl, flip = flip)

})

#' @rdname sfc_3x3_combined
#' @export
#' @examples
#' draw_rules_3x3_combined()
#' draw_rules_3x3_combined(flip = TRUE)
draw_rules_3x3_combined = function(flip = FALSE) {

    p = sfc_3x3_combined("I", flip = flip)

    grid.newpage()

    if(flip) {
        rules = p@rules@flip
    } else {
        rules = p@rules@rules
    }
    equation_max_width = max(do.call("unit.c", lapply(names(rules), function(nm) {
        do.call("unit.c", lapply(seq_along(rules[[nm]]), function(i) {
            convertWidth(grobWidth(grob_math(tex_pattern(nm, i,  rules[[nm]][[i]]), x = 0, y = 0)), "mm")
        }))
    })))

    gb1 = grob_single_base_rule(p, "I", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size, just = c("left", "top"))
    nc = length(gb1$children)
    gb1$children[[nc]]$width = gb1$children[[nc]]$width
    grid.draw(gb1)

    gb2 = grob_single_base_rule(p, "R", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size - gb1$vp$height, just = c("left", "top"))
    nc = length(gb2$children)
    gb2$children[[nc]]$width = gb2$children[[nc]]$width
    grid.draw(gb2)

    gb3 = grob_single_base_rule(p, "L", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size - gb1$vp$height - gb2$vp$height, just = c("left", "top"))
    nc = length(gb3$children)
    gb3$children[[nc]]$width = gb3$children[[nc]]$width
    grid.draw(gb3)

    gb4 = grob_single_base_rule(p, "U", equation_max_width = equation_max_width, flip = flip, x = size + gb1$vp$width + size, y = unit(1, "npc") - size, just = c("left", "top"))
    nc = length(gb4$children)
    gb4$children[[nc]]$width = gb4$children[[nc]]$width
    grid.draw(gb4)

	gb5 = grob_single_base_rule(p, "B", equation_max_width = equation_max_width, flip = flip, x = size + gb1$vp$width + size, y = unit(1, "npc") - size - gb4$vp$height, just = c("left", "top"))
    nc = length(gb5$children)
    gb5$children[[nc]]$width = gb5$children[[nc]]$width
    grid.draw(gb5)

    gb6 = grob_single_base_rule(p, "D", equation_max_width = equation_max_width, flip = flip, x = size + gb1$vp$width + size, y = unit(1, "npc") - size - gb4$vp$height - gb5$vp$height, just = c("left", "top"))
    nc = length(gb6$children)
    gb6$children[[nc]]$width = gb6$children[[nc]]$width
    grid.draw(gb6)

    gb7 = grob_single_base_rule(p, "P", equation_max_width = equation_max_width, flip = flip, x = size + gb1$vp$width + size, y = unit(1, "npc") - size - gb4$vp$height - gb5$vp$height - gb6$vp$height, just = c("left", "top"))
    nc = length(gb7$children)
    gb7$children[[nc]]$width = gb7$children[[nc]]$width
    grid.draw(gb7)

	gb8 = grob_single_base_rule(p, "Q", equation_max_width = equation_max_width, flip = flip, x = size + gb1$vp$width + size, y = unit(1, "npc") - size - gb4$vp$height - gb5$vp$height - gb6$vp$height - gb7$vp$height, just = c("left", "top"))
    nc = length(gb8$children)
    gb8$children[[nc]]$width = gb8$children[[nc]]$width
    grid.draw(gb8)

    gb9 = grob_single_base_rule(p, "C", equation_max_width = equation_max_width, flip = flip, x = size + gb1$vp$width + size, y = unit(1, "npc") - size - gb4$vp$height - gb5$vp$height - gb6$vp$height - gb7$vp$height - gb8$vp$height, just = c("left", "top"))
    nc = length(gb9$children)
    gb9$children[[nc]]$width = gb9$children[[nc]]$width
    grid.draw(gb9)

}

#' Apply to every unit in the sfc_nxn curve
#' @aliases sfc_apply
#' @rdname sfc_apply
#' @param p An `sfc_nxn` object.
#' @param depth An integer between 0 and `level-1` of the curve.
#' @param fun A function of which the argument `x` is a subunit in the curve. The subunit is an `sfc_nxn` object but only contains the current sub-sequence.
#'       The function should return an `sfc_seuqence` object with the same length as of `x`. The function can take an optional second argument which
#'       the index of the current subunit in the curve.
#' 
#' @details
#' This function is mainly used to flip subunits on various levels on the curve, thus mainly on the Peano curve and the Meander curve.
#' A depth of 0 corresponds to the complete curve. A depth of 1 corresponds to the nine first-level units, et al.
#' 
#' Currently, `sfc_apply()` only works on curves with a single base pattern as the seed.
#' 
#' @export
#' @return
#' An `sfc_nxn` object.
#' @examples
#' p = sfc_3x3_peano("I", level = 3)
#' # flip the global curve
#' draw_multiple_curves(
#'     p, 
#'     sfc_apply(p, 0, sfc_flip_unit),
#'     nrow = 1
#' )
#' 
#' # flip all the subunits on depth = 1
#' draw_multiple_curves(
#'     p, 
#'     sfc_apply(p, 1, sfc_flip_unit),
#'     nrow = 1
#' )
#' 
#' # flip all the subunits on depth = 2
#' draw_multiple_curves(
#'     p, 
#'     sfc_apply(p, 2, sfc_flip_unit),
#'     nrow = 1
#' )
#' 
#' # flip all level-1 patterns on the Peano curve to horizontal
#' # only works on the lowest subunit, 
#' p2 = sfc_apply(p, 2, function(x) {
#'     if(level1_unit_orientation(x) == "vertical") {
#'         sfc_flip_unit(x)
#'     } else {
#'         x
#'     }
#' })
#' # then on depth=1, only flip the unit with odd index
#' p3 = sfc_apply(p2, 1, function(x, i) {
#'     if(i %% 2 == 1) {
#'         sfc_flip_unit(x)
#'     } else {
#'         x
#'     }
#' })
#' draw_multiple_curves(p2, p3, nrow = 1)
#' 
#' # flip all level-1 patterns to vertical
#' p3 = sfc_apply(p, 2, function(x) {
#'     if(level1_unit_orientation(x) == "horizontal") {
#'         sfc_flip_unit(x)
#'     } else {
#'         x
#'     }
#' })
#' draw_multiple_curves(p, p3, nrow = 1)
setMethod("sfc_apply",
	signature = "sfc_nxn",
	definition = function(p, depth = 1, fun = function(x) x) {
	if(depth >= p@level) {
		stop_wrap("`depth` should only be smaller than the level of the curve.")
	}
	if(depth < 0) {
		stop_wrap("`depth` should be zero or a positive integer.")
	}
	if(length(p@seed) != 1) {
		stop_wrap("Currently, `sfc_apply()` only works on curves with a single base pattern as the seed.")
	}

	mode = p@mode

	n_block = (mode^2)^depth
	block_size = (mode^2)^(p@level - depth)

	fm = formals(fun)
	narg = length(fm)

	for(i in seq_len(n_block)) {
		ind = seq( (i-1)*block_size + 1, i*block_size )
		unit = p[ind, TRUE]

		if(narg == 1) {
			p[ind] = fun(unit)
		} else if(narg == 2) {
			if(names(fm)[2] %in% c("i", "j", "k")) {
				p[ind] = fun(unit, i)
			} else {
				p[ind] = fun(unit)
			}
		}
	}
	p
})

#' Coordinates of the points on the curve
#' @aliases sfc_segments
#' @rdname sfc_segments
#' @param p An `sfc_nxn` or `sfc_sequence` object.
#' @param bases A list of base patterns, consider to use [`BASE_LIST`].
#' @param start Coordinate of the start point.
#' @param ... Other argument.
#' 
#' @return
#' A two-column matrix of coordinates of points on the curve.
#' @export
#' @examples
#' p = sfc_2x2("I", "11")
#' loc = sfc_segments(p)
#' plot(loc, type = "l", asp = 1)
setMethod("sfc_segments", 
	signature = "sfc_nxn", 
	definition = function(p, bases = p@rules@bases, start = c(0, 0), ...) {

	callNextMethod(p, bases, start, ...)
})


#' @rdname sfc_segments
#' @details
#' For the `sfc_segments()` on the `sfc_sequence` object, if `bases` is not set,
#' it uses [`BASE_LIST`] internally. Make sure the sequence only contains the pre-defined base patterns.
#' @param by Which implementation? Only for the testing purpose.
#' @export
setMethod("sfc_segments", 
	signature = "sfc_sequence", 
	definition = function(p, bases = NULL, start = c(0, 0), by = "Cpp") {

	if(is.null(bases)) {
		bases = BASE_LIST[sfc_universe(p)]
	}

	if(any(sapply(bases, is.null))) {
		stop_wrap("`bases` does not cover all base patterns in `p`.")
	}

	bases = bases[ sfc_universe(p) ]
	seq = as.integer(p@seq)
	rot = p@rot
	
	if(by == "Cpp") {
		pos = sfc_segments_cpp(seq, rot, bases, start)
	} else {
		n = length(seq)
		pos = matrix(NA, nrow = n, ncol = 2)

		pos[1, ] = start
		for(i in seq_len(n)[-1]) {
			pos[i, ] = sfc_next_point(bases[[ seq[i-1] ]], pos[i-1, ], rot[i-1])
		}
	}

	pos2 = as.integer(round(pos))
	dim(pos2) = dim(pos)

	pos2
})


as_sfc_segment = function(x) {
	class(x) = union("sfc_segment", class(x))
	x
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

hilbert_curve_3d_cpp <- function(level) {
    .Call(`_sfcurve_hilbert_curve_3d_cpp`, level)
}

expand_by_rules_cpp <- function(rules, letters, code) {
    .Call(`_sfcurve_expand_by_rules_cpp`, rules, letters, code)
}

expand_by_rules_2_cpp <- function(rules, flip, letters, code, l_flip) {
    .Call(`_sfcurve_expand_by_rules_2_cpp`, rules, flip, letters, code, l_flip)
}

sfc_segments_cpp <- function(seq, rot, bases, start) {
    .Call(`_sfcurve_sfc_segments_cpp`, seq, rot, bases, start)
}

hilbert_curve_cpp <- function(level, type = 1L) {
    .Call(`_sfcurve_hilbert_curve_cpp`, level, type)
}

peano_curve_cpp <- function(level, type = 0L) {
    .Call(`_sfcurve_peano_curve_cpp`, level, type)
}

int_to_binary <- function(x, len) {
    .Call(`_sfcurve_int_to_binary`, x, len)
}

#' Reduce a curve
#' @aliases sfc_reduce
#' @rdname sfc_reduce
#' @param p An `sfc_nxn` object.
#' @param to Which level to reduce to? Value should be between 1 and `sfc_level(p) - 1`.
#' @details
#' The reduction is applied on the coordinates of points.
#' @return
#' A two-column matrix of coordinates of the reduced curve.
#' @export
#' @examples
#' p = sfc_3x3_peano("I", level = 3)
#' draw_multiple_curves(
#'     p, 
#'     sfc_reduce(p, 2), 
#'     sfc_reduce(p, 1), 
#'     nrow = 1)
setMethod("sfc_reduce",
	signature = "sfc_nxn",
	definition = function(p, to = sfc_level(p) - 1) {

	if(!(to <= sfc_level(p) - 1 && to >= 1)) {
		stop_wrap("`to` should take values in [", 1, ", ", sfc_level(p) - 1, "]")
	}

	mode = p@mode
	level = sfc_level(p)

	n_block = (mode^2)^to
	block_size = (mode^2)^(level - to)

	loc = sfc_segments(p)
	loc2 = matrix(NA_real_, nrow = n_block, ncol = 2)

	for(i in seq_len(n_block)) {
		ind = seq( (i-1)*block_size + 1, i*block_size )
		loc2[i, ] = colMeans(loc[ind, , drop = FALSE])
	}

	loc2

})

#' @rdname sfc_reduce
#' @param mode Mode of the curve.
#' @export
#' @examples
#' p = hilbert_curve(level = 4)
#' draw_multiple_curves(
#'     p, 
#'     sfc_reduce(p, 3), 
#'     sfc_reduce(p, 2), 
#'     nrow = 1)
setMethod("sfc_reduce",
	signature = "matrix",
	definition = function(p, to = level - 1, mode = NULL) {

	N = nrow(p)
	if(is.null(mode)) {
		if( abs(round(log(N)/2/log(2)) - log(N)/2/log(2)) < 1e-4) {
			mode = 2
		} else if( abs(round(log(N)/2/log(3)) - log(N)/2/log(3)) < 1e-4) {
			mode = 3
		}
	}
	level = round(log(N)/2/log(mode))
	
	n_block = (mode^2)^to
	block_size = (mode^2)^(level - to)

	loc = p
	loc2 = matrix(NA_real_, nrow = n_block, ncol = 2)

	for(i in seq_len(n_block)) {
		ind = seq( (i-1)*block_size + 1, i*block_size )
		loc2[i, ] = colMeans(loc[ind, , drop = FALSE])
	}

	loc2

})


#' @rdname sfc_reduce
#' @param gb A `grob` object returned by [`sfc_grob()`] or a `sfc_nxn` object then [`sfc_grob()`] is internally applied.
#' @param level The level of the unit.
#' @export
#' @details
#' `add_base_structure()` adds a base structure on a certain level to the curve.
#' @examples
#' p = hilbert_curve(3)
#' draw_multiple_curves(
#'     add_base_structure(p, level = 1),
#'     add_base_structure(p, level = 2),
#'     nrow = 1
#' )
add_base_structure = function(gb, level = 1) {
	if(inherits(gb, "sfc_nxn")) {
		gb = sfc_grob(gb)
	} else if(inherits(gb, "matrix")) {
		gb = sfc_grob(gb)
	}
	x = c(gb$children[[1]]$x0[1], gb$children[[1]]$x1)
	y = c(gb$children[[1]]$y0[1], gb$children[[1]]$y1)

	loc = cbind(x, y)

	loc_reduced = sfc_reduce(loc, level)
	n = nrow(loc_reduced)

	gb$children[["base_structure"]] = segmentsGrob(x0 = unit(loc_reduced[1:(n-1), 1], "native"), y0 = unit(loc_reduced[1:(n-1), 2], "native"),
		                               x1 = unit(loc_reduced[2:n, 1], "native"), y1 = unit(loc_reduced[2:n, 2], "native"),
		                               gp = gpar(col = "black", lwd = 2, lty = 3))
	gb$childrenOrder = c(gb$childrenOrder, "base_structure")

	gb
}

#' Create space-filling curves
#' 
#' @param seed The seed sequence. In most cases, the seed sequence is a single base pattern, which can be specified as a single letter, then `rot` controls
#'      the initial rotation of the base pattern. It also supports a sequence with more than one base patterns as the seed sequence. In this case,
#'      it can be specified as a string of more than one base letters, then `rot` can be set to a single rotation scalar which controls the rotation of the
#'      first letter, or a vector with the same length as the number of base letters.
#' @param code A vector of the expansion code. The left side corresponds to the top levels of the curve and the right side corresponds to the bottom level of the curve.
#'      The value can be set as a vector e.g. `c(1, 2, 1)`, or as a string e.g. `"121"`, or as a number e.g. `121`.
#' @param rot Rotation of the seed sequence, measured in the polar coordinate system, in degrees.
#' 
#' @details
#' - `sfc_2x2()` generates the Hilbert curve from the seed sequence.
#' - `sfc_3x3_peano()` generates the Peano curve from the seed sequence.
#' - `sfc_3x3_meander()` generates the Meander curve from the seed sequence.
#' 
#' @rdname spacefilling
#' @return
#' 
#' - `sfc_hilbert()` returns an `sfc_2x2` object.
#' - `sfc_peano()` returns an `sfc_3x3_peano` object.
#' - `sfc_meander()` returns an `sfc_3x3_meander` object.
#' 
#' These three classes are child classes of `sfc_nxn`.
#' @export
#' @import methods
#' @examples
#' sfc_2x2("I", "111") |> plot()
#' sfc_2x2("I", "111", rot = 90) |> plot()
#' sfc_2x2("IR", "111", rot = 90) |> plot()
sfc_2x2 = function(seed, code = integer(0), rot = 0L) {

	code = .parse_code(code, 1:2)

	if(inherits(seed, "character")) {
		seed = sfc_seed(seed, rot = rot, universe = sfc_universe(SFC_RULES_2x2))
	} else if(inherits(seed, "sfc_seed")) {
		seed@seq = factor(as.vector(seed@seq), levels = sfc_universe(SFC_RULES_2x2))
	} else {
		seed = sfc_seed(seq = seed@seq, rot = seed@rot)
	}

	p = as(seed, "sfc_2x2")
	p@seed = seed
	p@level = 0L
	p@mode = 2L

	for(i in seq_along(code)) {
		p = sfc_expand(p, code[i])
	}

	p@expansion = as.integer(code)
	p@universe = sfc_universe(SFC_RULES_2x2)
	p
	
}

.parse_code = function(code, full) {
	if(length(code) == 1) {
		if(is.numeric(code)) {
			if(code > 2) {
				code = as.character(code)
			}
		}
	}
	if(is.character(code)) {
		code = as.integer(strsplit(code[1], "")[[1]])
	}

	if(any(!code %in% 1:2)) {
		stop_wrap(paste0("`code` should only contain ", paste(1:2, collapse = ", "), "."))
	}

	code
}

#' @rdname spacefilling
#' @param flip Whether to use the "flipped" rules? For the Peano curve and the Meander curve, there is also a "fliiped" version 
#'      of curve expansion rules. On each level expansion in the Peano curve and the Meander curve, a point expands to nine points in 
#'      3x3 grids. Thus the value of `flip` can be set as a logical vector of length of nine that controls whether to use the flipped expansion
#'      for the corresponding unit. Besides such "1-to-9" mode, `flip` can also be set as a function which acccepts the number of current points in the curve and return
#'      a logical vector with the same length, i.e. the "all-to-all*9" mode.
#' @param level Specifically for `sfc_3x3_peano()`, since there is only one expansion code 1, it can also be generated by `rep(1, level)`.
#' @export
#' @examples
#' sfc_3x3_peano("I", "111") |> plot()
#' sfc_3x3_peano("I", "111", 
#'     flip = c(FALSE, FALSE, FALSE, TRUE, FALSE, FALSE, TRUE, FALSE, FALSE)) |> plot()
#' sfc_3x3_peano("IJ", "111") |> plot()
#'
#' sfc_3x3_peano("I", level = 4, flip = function(p) {
#'     p@rot %in% c(90, 270)
#' }) |> plot(lwd = 1)
#' 
#' level = 4
#' sfc_3x3_peano("I", level = level, flip = function(p) {
#'      if(length(p) == 9^(level-1)) {
#'          l = rep(FALSE, length(p))
#'          ind = 1:9^2 + 9^2*rep(c(0, 2, 4, 6, 8), each = 9^2)
#'          l[ind] = p@rot[ind] %in% c(90, 270)
#' 
#'          ind = 1:9^2 + 9^2*rep(c(1, 3, 5, 7), each = 9^2)
#'          l[ind] = p@rot[ind] %in% c(0, 180)
#' 
#'          l
#'     } else {
#'          rep(FALSE, length(p))
#'     }
#' }) |> plot(lwd = 1)
#' 
sfc_3x3_peano = function(seed, code = integer(0), rot = 0L, level = NULL, flip = FALSE) {

	if(!is.null(level)) {
		code = rep(1, level)
	} else {
		code = .parse_code(code, 1L)
	}

	if(inherits(seed, "character")) {
		seed = sfc_seed(seed, rot = rot, universe = sfc_universe(SFC_RULES_3x3_PEANO))
	} else if(inherits(seed, "sfc_seed")) {
		levels(seed@seq) = sfc_universe(SFC_RULES_3x3_PEANO)
	} else {
		seed = sfc_seed(seq = seed@seq, rot = seed@rot)
	}

	p = as(seed, "sfc_3x3_peano")
	p@seed = seed
	p@level = 0L
	p@mode = 3L

	if(is.logical(flip)) {
		if(!(length(flip) == length(seed) || length(flip) == 9 || length(flip) == 1)) {
			stop_wrap("If `flip` is a logical vector, it should have a length the same as `seed` or 9\n")
		}
	}

	if(is.function(flip)) {
		for(i in seq_along(code)) {
			p = sfc_expand(p, code[i], flip = flip(p))
		}
	} else {
		for(i in seq_along(code)) {
		
			if(i == 1) {
				if(length(flip) == length(seed)) {
					p = sfc_expand(p, code[i], flip = flip)
				} else {
					p = sfc_expand(p, code[i], flip = flip[1])
				}
			} else if (i == 2) {
				if(length(flip) == 9) {

				} else {
					flip = rep(flip, each = 9)
				}
				
				p = sfc_expand(p, code[i], flip = flip)
			} else {
				flip = rep(flip, each = 9)
				p = sfc_expand(p, code[i], flip = flip)
			}
		}
	}

	p@expansion = as.integer(code)
	p@universe = sfc_universe(seed)

	p
	
}

#' @rdname spacefilling
#' @export
#' @examples
#' sfc_3x3_meander("I", "111") |> plot()
#' sfc_3x3_meander("I", "111", 
#'     flip = c(TRUE, FALSE, TRUE, FALSE, FALSE, TRUE, FALSE, FALSE, TRUE)) |> plot()
#' sfc_3x3_meander("IR", "111") |> plot()
sfc_3x3_meander = function(seed, code = integer(0), rot = 0L, flip = FALSE) {

	code = .parse_code(code, 1:2)

	if(inherits(seed, "character")) {
		seed = sfc_seed(seed, rot = rot, universe = sfc_universe(SFC_RULES_3x3_MEANDER))
	} else if(inherits(seed, "sfc_seed")) {
		levels(seed@seq) = sfc_universe(SFC_RULES_3x3_MEANDER)
	} else {
		seed = sfc_seed(seq = seed@seq, rot = seed@rot)
	}

	p = as(seed, "sfc_3x3_meander")
	p@seed = seed
	p@level = 0L
	p@mode = 3L

	if(is.logical(flip)) {
		if(!(length(flip) == length(seed) || length(flip) == 9 || length(flip) == 1)) {
			stop_wrap("If `flip` is a logical vector, it should have a length the same as `seed` or 9\n")
		}
	}

	if(is.function(flip)) {
		for(i in seq_along(code)) {
			p = sfc_expand(p, code[i], flip = flip(p))
		}
	} else {
		for(i in seq_along(code)) {
		
			if(i == 1) {
				if(length(flip) == length(seed)) {
					p = sfc_expand(p, code[i], flip = flip)
				} else {
					p = sfc_expand(p, code[i], flip = flip[1])
				}
			} else if (i == 2) {
				if(length(flip) == 9) {

				} else {
					flip = rep(flip, each = 9)
				}
				
				p = sfc_expand(p, code[i], flip = flip)
			} else {
				flip = rep(flip, each = 9)
				p = sfc_expand(p, code[i], flip = flip)
			}
		}
	}

	p@expansion = as.integer(code)
	p@universe = sfc_universe(seed)
	p
	
}

#' The level of the curve
#' @aliases sfc_level
#' @rdname sfc_level
#' @param p An `sfc_nxn` object.
#'
#' @return An integer.
#' @export
#' @examples
#' p = sfc_2x2("I", "11")
#' sfc_level(p)
#' 
#' p = sfc_2x2("I", "1111")
#' sfc_level(p)
setMethod("sfc_level",
	signature = "sfc_nxn", 
	definition = function(p) {
	p@level
})

#' @rdname sfc_mode
#' @export
#' @return An integer.
#' @examples
#' p = sfc_2x2("I", "1")
#' sfc_mode(p)
#' p = sfc_3x3_peano("I", "1")
#' sfc_mode(p)
setMethod("sfc_mode",
    signature = "sfc_nxn",
    definition = function(p) {

    sfc_mode(p@rules)
})



setAs("sfc_seed", "sfc_2x2", function(from) {
	p = new("sfc_2x2")
	p@seq = from@seq
	levels(p@seq) = sfc_universe(SFC_RULES_2x2)
	if(any(is.na(p@seq))) {
		stop_wrap("Base letters should all be in `sfc_universe(SFC_RULES_2x2)`.")
	}
	p@rot = from@rot
	p@universe = sfc_universe(SFC_RULES_2x2)
	p@level = 0L
	p@mode = 2L
	p@rules = SFC_RULES_2x2

	p
})


setAs("sfc_sequence", "sfc_3x3_peano", function(from) {
	p = new("sfc_3x3_peano")
	p@seq = from@seq
	levels(p@seq) = sfc_universe(SFC_RULES_3x3_PEANO)
	if(any(is.na(p@seq))) {
		stop_wrap("Base letters should all be in `sfc_universe(SFC_RULES_3x3_PEANO)`.")
	}
	p@rot = from@rot
	p@universe = sfc_universe(SFC_RULES_3x3_PEANO)
	p@level = 0L
	p@mode = 3L
	p@rules = SFC_RULES_3x3_PEANO

	p
})


setAs("sfc_sequence", "sfc_3x3_meander", function(from) {
	p = new("sfc_3x3_meander")
	p@seq = from@seq
	levels(p@seq) = sfc_universe(SFC_RULES_3x3_MEANDER)
	if(any(is.na(p@seq))) {
		stop_wrap("Base letters should all be in `sfc_universe(SFC_RULES_3x3_MEANDER)`.")
	}
	p@rot = from@rot
	p@universe = sfc_universe(SFC_RULES_3x3_MEANDER)
	p@level = 0L
	p@mode = 3L
	p@rules = SFC_RULES_3x3_MEANDER

	p
})

setAs("sfc_sequence", "sfc_nxn", function(from) {
	p = new("sfc_nxn")
	p@seq = from@seq
	p@rot = from@rot
	p@universe = levels(from@seq)
	p@level = 0L

	p
})


#' @param object The corresponding object.
#' @rdname show
#' @export
setMethod("show",
	signature = "sfc_nxn",
	definition = function(object) {

	cat("An", class(object)[1], "object.\n")
	cat("  Increase mode: ", object@mode, " x ", object@mode, "\n", sep = "")
	cat("  Level: ", object@level, "\n", sep = "")
	cat("  Expansion rule:", object@rules@name, "\n")

	if(length(object) < (object@mode^2)^object@level) {
		cat("  A fragment from the original curve.\n")
	}
	cat("\n")

	callNextMethod(object)
	
	cat("\n")
	cat("Seed: ")
	callNextMethod(object@seed)
})


#' Level-1 unit in the Peano curve and Meander curve
#' @aliases level1_unit_orientation
#' @rdname level1_unit_orientation
#' @param p For `level1_unit_orientation()`, it is an `sfc_3x3_peano` or `sfc_3x3_meander` unit on level 1.
#'       For `change_level1_unit_orientation()`, it is a normal `sfc_3x3_peano` or `sfc_3x3_meander` object.
#' @return `level1_unit_orientation()` returns "verticalhorizontal" (on the `sfc_3x3_peano` object) or "forward/backward" (on the `sfc_3x3_meander` object).
#' @details
#' "vertical" and "horizontal" correspond to the direction of the "long segments" in a Peano curve (see **Examples**).
#' 
#' `level1_unit_orientation()` is normally used inside [`sfc_apply()`]. `change_level1_unit_orientation()`
#' changes all level-1 units of a Peano curve or a Meander curve simultaneously.
#' @export
setMethod("level1_unit_orientation",
	signature = "sfc_3x3_peano",
	definition = function(p) {
	if(length(p) != 9) {
		stop_wrap("`u` should be an sfc_3x3_peano unit with length of 9.")
	}
    loc = sfc_segments(p)
    if(equal_to(loc[1, 1], loc[2, 1]) && equal_to(loc[2, 1], loc[3, 1])) {
        "vertical"
    } else {
        "horizontal"
    }
})
 
#' @aliases change_level1_unit_orientation
#' @param to A string of "vertical/horizontal" (on the `sfc_3x3_peano` object) or "forward/backward" (on the `sfc_3x3_meander` object).
#' @return `change_level1_unit_orientation()` returns an `sfc_3x3_peano` or `sfc_3x3_meander` object.
#' @export
#' @rdname level1_unit_orientation
#' @examples
#' p = sfc_3x3_peano("I", 111)
#' # the first level-1 unit
#' level1_unit_orientation(p[1:9, TRUE])
#' # the fourth level-1 unit
#' level1_unit_orientation(p[1:9 + 27, TRUE])
#' p2 = change_level1_unit_orientation(p, "horizontal")
#' p3 = change_level1_unit_orientation(p, "vertical")
#' draw_multiple_curves(p, p2, p3, 
#'     title = c("original", "all horizontal", "all vertical"), nrow = 1)
setMethod("change_level1_unit_orientation",
	signature = "sfc_3x3_peano",
	definition = function(p, to = c("horizontal", "vertical")) {
	to = match.arg(to)
	sfc_apply(p, log(length(p))/log(9) - 1, function(x) {
	    if(level1_unit_orientation(x) != to) {
	        sfc_flip_unit(x)
	    } else {
	        x
	    }
	})
})


#' @rdname level1_unit_orientation
#' @export
setMethod("level1_unit_orientation",
	signature = "sfc_3x3_meander",
	definition = function(p) {
	if(length(p) != 9) {
		stop_wrap("`u` should be an sfc_3x3_meander unit with length of 9.")
	}
    loc = sfc_segments(p)
    if( (loc[1, 1] == loc[2, 1] && loc[2, 1] == loc[3, 1]) || (loc[1, 2] == loc[2, 2] && loc[2, 2] == loc[3, 2])) {
        "forward"
    } else {
        "backward"
    }
})
 
#' @export
#' @rdname level1_unit_orientation
#' @examples
#' # by default, orientations of all level-1 units in Meander curve are forward
#' p = sfc_3x3_meander("I", 111)
#' level1_unit_orientation(p[1:9, TRUE])
#' p2 = change_level1_unit_orientation(p, "backward")
#' draw_multiple_curves(p, p2,
#'    title = c("all forward", "all backward"), nrow = 1)
setMethod("change_level1_unit_orientation",
	signature = "sfc_3x3_meander",
	definition = function(p, to = c("forward", "backward")) {
	to = match.arg(to)
	sfc_apply(p, log(length(p))/log(9) - 1, function(x) {
	    if(level1_unit_orientation(x) != to) {
	        sfc_flip_unit(x)
	    } else {
	        x
	    }
	})
})

setClass("sfc_4x4_meander",
	slots = c("type" = "integer"),
	contains = "sfc_nxn",
	prototype = list(mode = 4L))

setClass("sfc_4x4_meander_1",
	contains = "sfc_4x4_meander",
	prototype = list(type = 1L))

setClass("sfc_4x4_meander_2",
	contains = "sfc_4x4_meander",
	prototype = list(type = 2L))


#' 4x4 space-filling curves in meander type
#' 
#' @rdname sfc_4x4_meander
#' @param seed The seed sequence. In most cases, the seed sequence is a single base pattern, which can be specified as a single letter, then `rot` controls
#'      the initial rotation of the base pattern. It also supports a sequence with more than one base patterns as the seed sequence. In this case,
#'      it can be specified as a string of more than one base letters, then `rot` can be set to a single rotation scalar which controls the rotation of the
#'      first letter, or a vector with the same length as the number of base letters.
#' @param code A vector of the expansion code. The left side corresponds to the higher levels (more to the top-level) of the curve and the right side corresponds to the lower level (more to the bottom-level) of the curve.
#'      The value can be set as a vector e.g. `c(1, 2, 1)`, or as a string e.g. `"121"`, or as a number e.g. `121`.
#' @param rot Rotation of the seed sequence, measured in the polar coordinate system, in degrees.
#' @param flip The same setting as in [`sfc_3x3_peano()`] or [`sfc_3x3_meander()`].
#' @param type Which type of rules to use? 1 for [`SFC_RULES_4x4_MEANDER_1`] and 2 for [`SFC_RULES_4x4_MEANDER_2`].
#' 
#' @details
#' It is an extension of the 3x3 Meander curves to mode 4. For simplicity, it only supports `I/R/L` base patterns.
#' @export
#' @return
#' `sfc_4x4_meander()` returns an `sfc_4x4_meander` object.
#' @examples
#' draw_multiple_curves(
#'     sfc_4x4_meander("I", "11", type = 1),
#'     sfc_4x4_meander("I", "12", type = 1),
#'     sfc_4x4_meander("I", "11", type = 2),
#'     sfc_4x4_meander("I", "12", type = 2),
#'     nrow = 2
#' )
#' seed = paste(rep(paste0("R", sapply(0:10, function(i) strrep("I", i))), each = 2), collapse="")
#' sfc_4x4_meander(seed, 1) |> plot()
sfc_4x4_meander = function(seed, code = integer(0), rot = 0L, flip = FALSE, type = 1L) {

	code = .parse_code(code, 1:2)

	if(inherits(seed, "character")) {
		seed = sfc_seed(seed, rot = rot, universe = sfc_universe(SFC_RULES_4x4_MEANDER_1))
	} else if(inherits(seed, "sfc_seed")) {
		seed@seq = factor(as.vector(seed@seq), levels = sfc_universe(SFC_RULES_4x4_MEANDER_1))
	} else {
		seed = sfc_seed(seq = seed@seq, rot = seed@rot)
	}

	if(type == 1) {
		p = as(seed, "sfc_4x4_meander_1")
	} else {
		p = as(seed, "sfc_4x4_meander_2")
	}
	p@seed = seed
	p@level = 0L
	p@mode = 4L

	if(is.logical(flip)) {
		if(!(length(flip) == length(seed) || length(flip) == 16 || length(flip) == 1)) {
			stop_wrap("If `flip` is a logical vector, it should have a length the same as `seed` or 16\n")
		}
	}

	if(is.function(flip)) {
		for(i in seq_along(code)) {
			p = sfc_expand(p, code[i], flip = flip(p))
		}
	} else {
		for(i in seq_along(code)) {
		
			if(i == 1) {
				if(length(flip) == length(seed)) {
					p = sfc_expand(p, code[i], flip = flip)
				} else {
					p = sfc_expand(p, code[i], flip = flip[1])
				}
			} else if (i == 2) {
				if(length(flip) == 16) {

				} else {
					flip = rep(flip, each = 16)
				}
				
				p = sfc_expand(p, code[i], flip = flip)
			} else {
				flip = rep(flip, each = 16)
				p = sfc_expand(p, code[i], flip = flip)
			}
		}
	}

	p@expansion = as.integer(code)
	p@universe = sfc_universe(SFC_RULES_4x4_MEANDER_1)
	p
	
}

setAs("sfc_sequence", "sfc_4x4_meander_1", function(from) {
	p = new("sfc_4x4_meander_1")
	p@seq = from@seq
	levels(p@seq) = sfc_universe(SFC_RULES_4x4_MEANDER_1)
	if(any(is.na(p@seq))) {
		stop_wrap("Base letters should all be in `sfc_universe(SFC_RULES_4x4_MEANDER_1)`.")
	}
	p@rot = from@rot
	p@universe = sfc_universe(SFC_RULES_4x4_MEANDER_1)
	p@level = 0L
	p@mode = 4L
	p@rules = SFC_RULES_4x4_MEANDER_1

	p
})

setAs("sfc_sequence", "sfc_4x4_meander_2", function(from) {
	p = new("sfc_4x4_meander_2")
	p@seq = from@seq
	levels(p@seq) = sfc_universe(SFC_RULES_4x4_MEANDER_2)
	if(any(is.na(p@seq))) {
		stop_wrap("Base letters should all be in `sfc_universe(SFC_RULES_4x4_MEANDER_2)`.")
	}
	p@rot = from@rot
	p@universe = sfc_universe(SFC_RULES_4x4_MEANDER_2)
	p@level = 0L
	p@mode = 4L
	p@rules = SFC_RULES_4x4_MEANDER_2

	p
})

#' @rdname sfc_4x4_meander
#' @param p An `sfc_4x4_meander` object.
#' @export
setMethod("sfc_expand",
	signature = "sfc_4x4_meander",
	definition = function(p, code, flip = FALSE) {

	seq = p@seq
	rot = p@rot
	n = length(p@seq)

	rules = p@rules

	tl = integer(n)
	tl[1] = code

	if(n > 1) {
		for(i in 2:n) {
			tl[i] = traverse_type_2x2(tl[i-1], rot[i-1], rot[i])
		}
	}

	sfc_expand_by_rules(rules, p, code = tl, flip = flip)

})

#' @rdname sfc_4x4_meander
#' @export
#' @examples
#' draw_rules_4x4_meander(type = 1)
#' draw_rules_4x4_meander(type = 2)
draw_rules_4x4_meander = function(type = 1, flip = FALSE) {

    p = sfc_4x4_meander("I", type = type, flip = flip)

    grid.newpage()

    if(flip) {
        rules = p@rules@flip
    } else {
        rules = p@rules@rules
    }
    equation_max_width = max(do.call("unit.c", lapply(names(rules), function(nm) {
        do.call("unit.c", lapply(seq_along(rules[[nm]]), function(i) {
            convertWidth(grobWidth(grob_math(tex_pattern(nm, i,  rules[[nm]][[i]]), x = 0, y = 0)), "mm")
        }))
    })))

    gb1 = grob_single_base_rule(p, "I", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size, just = c("left", "top"))
    nc = length(gb1$children)
    gb1$children[[nc]]$width = gb1$children[[nc]]$width 
    grid.draw(gb1)

    gb2 = grob_single_base_rule(p, "R", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size - gb1$vp$height, just = c("left", "top"))
    nc = length(gb2$children)
    gb2$children[[nc]]$width = gb2$children[[nc]]$width 
    grid.draw(gb2)

    gb3 = grob_single_base_rule(p, "L", equation_max_width = equation_max_width, flip = flip, x = size, y = unit(1, "npc") - size - gb1$vp$height - gb2$vp$height, just = c("left", "top"))
    nc = length(gb3$children)
    gb3$children[[nc]]$width = gb3$children[[nc]]$width
    grid.draw(gb3)

}

#' @keywords internal 
#' @details
#' Please go to the package's website: \url{https://jokergoo.github.io/sfcurve/} for more details.
"_PACKAGE"

setClass("sfc_sequence",
	slots = c("seq" = "factor",
		      "rot" = "integer",
		      "universe" = "character"))

setClass("sfc_seed",
	contains = "sfc_sequence")

setClass("sfc_unit",
	slots = list("corner" = "integer"),
	contains = "sfc_sequence")

setClass("sfc_rules",
    slots = c("rules" = "list",
    	      "flip" = "list",
              "universe" = "character",
              "name" = "character",
              "bases" = "list"))

setClass("sfc_nxn",
	slots = c("seed" = "sfc_seed",
	          "rules" = "sfc_rules",
		      "expansion" = "integer",
		      "level" = "integer",
		      "mode" = "integer"),
	contains = "sfc_sequence")

setClass("sfc_2x2",
	contains = "sfc_nxn",
	prototype = list(mode = 2L))

setClass("sfc_3x3_peano",
	contains = "sfc_nxn",
	prototype = list(mode = 3L))

setClass("sfc_3x3_meander",
	contains = "sfc_nxn",
	prototype = list(mode = 3L))


setGeneric('sfc_expand', function(p, ...) standardGeneric('sfc_expand'))
setGeneric('sfc_expand_by_rules', function(p, seq, ...) standardGeneric('sfc_expand_by_rules'))
setGeneric('sfc_grob', function(p, ...) standardGeneric('sfc_grob'))
setGeneric('sfc_rotate', function(p, ...) standardGeneric('sfc_rotate'))
setGeneric('sfc_hflip', function(p, ...) standardGeneric('sfc_hflip'))
setGeneric('sfc_vflip', function(p, ...) standardGeneric('sfc_vflip'))
setGeneric('sfc_dflip', function(p, ...) standardGeneric('sfc_dflip'))
setGeneric('sfc_reverse', function(p, ...) standardGeneric('sfc_reverse'))
setGeneric('sfc_segments', function(p, ...) standardGeneric('sfc_segments'))
setGeneric('sfc_shape', function(p, ...) standardGeneric('sfc_shape'))
setGeneric('sfc_vaidate', function(p, ...) standardGeneric('sfc_vaidate'))
setGeneric('sfc_to_base', function(p, ...) standardGeneric('sfc_to_base'))
setGeneric('sfc_level', function(p, ...) standardGeneric('sfc_level'))
setGeneric('sfc_mode', function(p, ...) standardGeneric('sfc_mode'))
setGeneric('sfc_universe', function(p, ...) standardGeneric('sfc_universe'))
setGeneric('sfc_validate', function(p, ...) standardGeneric('sfc_validate'))
setGeneric('sfc_previous_point', function(p, ...) standardGeneric('sfc_previous_point'))
setGeneric('sfc_next_point', function(p, ...) standardGeneric('sfc_next_point'))
setGeneric('sfc_is_compatible', function(p1, p2, ...) standardGeneric('sfc_is_compatible'))
setGeneric('sfc_index', function(p, ...) standardGeneric('sfc_index'))
setGeneric('sfc_flip_unit', function(p, ...) standardGeneric('sfc_flip_unit'))
setGeneric('sfc_apply', function(p, ...) standardGeneric('sfc_apply'))
setGeneric('sfc_reduce', function(p, ...) standardGeneric('sfc_reduce'))
setGeneric('level1_unit_orientation', function(p, ...) standardGeneric('level1_unit_orientation'))
setGeneric('change_level1_unit_orientation', function(p, ...) standardGeneric('change_level1_unit_orientation'))

#' @aliases sfc_grob
#' @rdname sfc_grob
#' @param bases A list of base patterns, normally [`BASE_LIST`] is used.
#' @param extend Whether to add the entry and exit segments?
#' @param title Whether to add title on the top of the plot? The title is constructed in the form of `initial_seed|expansion_code`, e.g. `I|111`.
#'      The value can also be a string.
#' @param closed Whether the curve is closed? 
#' @param lwd Line width.
#' @param col Color for segments. If the value is `NULL`, it uses the "Spectral" color palettes.
#' @param ... Other arguments passed to [`grid::viewport()`] or `sfc_grob()`.
#' 
#' @details
#' If `p` is an `sfc_sequence` and `p` contains base patterns defined in `"I/J/R/L/U/B/D/P/Q/C"`,
#' the default [`BASE_LIST`] is automatically used for `bases`. If `p` is an `sfc_nxn` object, `bases`
#' is already stored in `p` and it is passed to this function automatically.
#' 
#' @export
#' @import colorRamp2
setMethod("sfc_grob",
	signature = "sfc_sequence",
	definition = function(p, bases = NULL, extend = FALSE, title = FALSE, closed = FALSE, lwd = 4, col = NULL, ...) {

	if(is.null(bases)) {
		bases = BASE_LIST[sfc_universe(p)]
	}

	loc = sfc_segments(p, bases)
	if(closed) {
		loc = rbind(loc, loc[1, ])
	}

	n = nrow(loc)

	rgx = range(loc[, 1])
	rgx[1] = rgx[1] - 1; rgx[2] = rgx[2] + 1
	rgy = range(loc[, 2])
	rgy[1] = rgy[1] - 1; rgy[2] = rgy[2] + 1
	
	r = (diff(rgx) + 1)/(diff(rgy) + 1)

	vp = viewport(name = "vp_sfc_sequence", xscale = rgx, yscale = rgy, width = unit(r, "snpc"), height = unit(1, "snpc"), gp = gpar(lineend = "square", linejoin = "mitre"), ...)

	gbl = list()

	if(n > 1) {
		if(is.null(col)) {
			col_fun = colorRamp2(seq(1, n, length = 11), c("#9E0142", "#D53E4F", "#F46D43", "#FDAE61", "#E0C785", "#E2E2AB", "#CDD993", "#ABDDA4", "#66C2A5", "#3288BD", "#5E4FA2"))
			gbl[[1]] = segmentsGrob(loc[1:(n-1), 1], loc[1:(n-1), 2], loc[2:n, 1], loc[2:n, 2], default.units = "native", gp = gpar(col = col_fun(1:(n-1)), lwd = lwd))
		} else {
			gbl[[1]] = segmentsGrob(loc[1:(n-1), 1], loc[1:(n-1), 2], loc[2:n, 1], loc[2:n, 2], default.units = "native", gp = gpar(col = col, lwd = lwd))
		}
	
	} else {
		gbl[[1]] = pointsGrob(loc[, 1], loc[, 2], pch = 16, size = unit(4, "pt"), gp = gpar(col = "#9E0142"))
	}

	if(extend) {
		seq = as.character(p@seq)
		rot = p@rot

		b = bases[[ seq[1] ]]

		if(is.na(b@in_direction)) {
			gbl[[2]] = pointsGrob(loc[1, 1], loc[1, 2], default.units = "native", pch = 4, size = unit(4, "pt"), gp = gpar(col = "grey"))
		} else {
			prev = sfc_previous_point(b, loc[1, ], rot[1], length = 0.6)
			gbl[[2]] = segmentsGrob(prev[1], prev[2], 0, 0, default.units = "native", gp = gpar(col = "grey", lwd = lwd))
		}

		b = bases[[ seq[n] ]]

		if(is.na(b@out_direction)) {
			gbl[[3]] = pointsGrob(loc[n, 1], loc[n, 2], default.units = "native", pch = 4, size = unit(4, "pt"), gp = gpar(col = "grey"))
		} else {
			last = sfc_next_point(b, loc[n, ], rot[n], length = 0.6)
			gbl[[3]] = segmentsGrob(loc[n, 1], loc[n, 2], last[1], last[2], default.units = "native", gp = gpar(col = "grey", lwd = lwd), arrow = arrow(length = unit(0.2, "native"), angle = 15))
		}

		gbl = gbl[c(2, 3, 1)]
	}

	if(is.logical(title)) {
		if(title) {
			if(inherits(p, "sfc_nxn")) {
				seed = p@seed
				pt = paste0("paste(", paste("italic(", seed@seq[1], ")^", seed@rot[1], sep = "", collapse = ","), ")")
				pt = paste0("paste(", pt, ", symbol('|'), ", paste(p@expansion, collapse = ""), ")")
				gbl[[ length(gbl) + 1 ]] = textGrob(parse(text = pt), x = unit(mean(rgx), "native"), y = unit(rgy[2]-0.25, "native"), gp = gpar(fontsize = 12, fontfamily = "Times"))
			}
		}
	} else {
		gbl[[ length(gbl) + 1 ]] = textGrob(title, x = unit(mean(rgx), "native"), y = unit(rgy[2]-0.25, "native"), gp = gpar(fontsize = 12))
	}

	args = gbl
	args$vp = vp
	args$cl = "grob_sfc_sequence"

	do.call(grobTree, args)
})

#' @param x The corresponding object.
#' @param grid Whether to add grid lines on the plot?
#' @export
#' @rdname sfc_grob
#' @examples
#' plot(sfc_2x2("I", "11"))
#' plot(sfc_2x2("I", "11"), extend = TRUE, title = TRUE, grid = TRUE)
#' plot(sfc_sequence("IIIRRR"))
plot.sfc_sequence = function(x, bases = NULL, grid = FALSE, 
	extend = FALSE, title = FALSE, closed = FALSE, ...) {

	if(is.null(bases)) {
		bases = BASE_LIST[sfc_universe(x)]
	}

	gb = sfc_grob(x, bases, extend = extend, title = title, closed = closed, ...)
	grid.newpage()
	grid.draw(gb)

	if(grid) {
		add_grid_lines()
	}
}

#' @rdname sfc_grob
#' @export
setMethod("sfc_grob",
	signature = "sfc_nxn",
	definition = function(p, bases = p@rules@bases, extend = FALSE, title = FALSE, closed = FALSE, ...) {

	callNextMethod(p, bases, extend = extend, title = title, closed = closed, ...)
})

#' @export
makeContext.grob_sfc_sequence = function(x) {
    vp_width = convertWidth(unit(1, "npc"), "in", valueOnly = TRUE)
    vp_height = convertHeight(unit(1, "npc"), "in", valueOnly = TRUE)

    r = as.numeric(x$vp$width)

    if(length(r)) {
	    if(vp_width > r*vp_height) {
	        x$vp$width = unit(r*vp_height, "in")
	        x$vp$height = unit(vp_height, "in")
	    } else {
	        x$vp$width = unit(vp_width, "in")
	        x$vp$height = unit(vp_width/r, "in")
	    }
	   }
    x
}


#' @export
#' @rdname sfc_grob
plot.sfc_nxn = function(x, grid = FALSE, extend = FALSE, title = FALSE, closed = FALSE, ...) {
	gb = sfc_grob(x, extend = extend, title = title, closed = closed, ...)
	grid.newpage()
	grid.draw(gb)

	if(grid) {
		add_grid_lines()
	}
}


add_grid_lines = function() {
	vp = current.viewport()
	nm = vp$name
	downViewport("vp_sfc_sequence")
	vp = current.viewport()
	xscale = vp$xscale
	yscale = vp$yscale

	nx = xscale[2] - xscale[1] - 1
	ny = yscale[2] - yscale[1] - 1

	grid.segments(seq(xscale[1] + 0.5, xscale[2] - 0.5, by = 1), rep(yscale[1] + 0.5, ny), 
		          seq(xscale[1] + 0.5, xscale[2] - 0.5, by = 1), rep(yscale[2] - 0.5, ny), 
		          default.units = "native", gp = gpar(col = "#CCCCCC", lty = 2))
	grid.segments(rep(xscale[1] + 0.5, ny), seq(yscale[1] + 0.5, yscale[2] - 0.5, by = 1),
		          rep(xscale[2] - 0.5, ny), seq(yscale[1] + 0.5, yscale[2] - 0.5, by = 1), 
		          default.units = "native", gp = gpar(col = "#CCCCCC", lty = 2))
}

#' @rdname sfc_grob
#' @export
setMethod("sfc_grob",
	signature = "matrix",
	definition = function(p, title = NULL, closed = FALSE, lwd = 4, col = NULL, ...) {

	loc = p
	if(closed) {
		loc = rbind(loc, loc[1, ])
	}

	n = nrow(loc)

	rgx = range(loc[, 1])
	rgx[1] = rgx[1] - 1; rgx[2] = rgx[2] + 1
	rgy = range(loc[, 2])
	rgy[1] = rgy[1] - 1; rgy[2] = rgy[2] + 1
	
	r = (diff(rgx) + 1)/(diff(rgy) + 1)

	vp = viewport(xscale = rgx, yscale = rgy, width = unit(r, "snpc"), height = unit(1, "snpc"), gp = gpar(lineend = "square", linejoin = "mitre"), ...)

	gbl = list()

	if(n > 1) {
		if(is.null(col)) {
			col_fun = colorRamp2(seq(1, n, length = 11), c("#9E0142", "#D53E4F", "#F46D43", "#FDAE61", "#E0C785", "#E2E2AB", "#CDD993", "#ABDDA4", "#66C2A5", "#3288BD", "#5E4FA2"))
			gbl[[1]] = segmentsGrob(loc[1:(n-1), 1], loc[1:(n-1), 2], loc[2:n, 1], loc[2:n, 2], default.units = "native", gp = gpar(col = col_fun(1:(n-1)), lwd = lwd))
		} else {
			gbl[[1]] = segmentsGrob(loc[1:(n-1), 1], loc[1:(n-1), 2], loc[2:n, 1], loc[2:n, 2], default.units = "native", gp = gpar(col = col, lwd = lwd))
		}
	} else {
		gbl[[1]] = pointsGrob(loc[, 1], loc[, 2], pch = 16, size = unit(4, "pt"), gp = gpar(col = "#9E0142"))
	}

	if(is.character(title)) {
		gbl[[ length(gbl) + 1 ]] = textGrob(title, x = unit(mean(rgx), "native"), y = unit(rgy[2], "native") - unit(1, "native") + unit(4, "pt"), just = "bottom", gp = gpar(fontsize = 10))
	}

	args = gbl
	args$vp = vp
	args$cl = "grob_sfc_sequence"

	do.call(grobTree, args)
})

setClass("sfc_base",
	slots = c("letter" = "character",
		      "preceeding" = "numeric",
		      "current" = "numeric",
		      "succeeding" = "numeric",
		      "in_direction" = "numeric",
		      "out_direction" = "numeric",
		      "primary" = "logical",
		      "open" = "logical",
		      "grob" = "ANY"),
	prototype = list("current" = c(0, 0),
		             "grob" = gTree()))


#' Constructor of the sfc_base class
#' 
#' @param letter A single letter to represent the base pattern.
#' @param in_direction The direction of the segment that enters the point, measured in the polar coordinate system, in degrees.
#' @param out_direction The direction of the segment that leaves the point, measured in the polar coordinate system, in degrees.
#' @param grob A [`grid::grob()`] object of this base pattern. If it is not set, it is generated according to `in_direction` and `out_direction`.
#' @param primary Currently, going forward, turning left and turning right can be set as primary base patterns because other high-level patterns
#'        can be built from them.
#' @param open Can the base pattern be connected to other base patterns?
#' 
#' @details
#' The "base pattern" is designed not only for single point but also for combination of points that form a "base curve". However,
#' currently, it is fixed to the single point base pattern. 
#' 
#' Currently, this package supports 2x2 and 3x3 space-filling curves that fills grids in 2D space constructed by the Gaussian integers.
#' And when the curve expands, we only allow the segments to go forward, backward, left and right. Thus there are the following base patterns
#' pre-defined in this package:
#' 
#' - [`BASE_I`]/[`BASE_J`]: go forward.
#' - [`BASE_R`]: turn right.
#' - [`BASE_L`]: turn left.
#' - [`BASE_U`]: go backward.
#' - [`BASE_B`]: leave the start point where the start point is closed.
#' - [`BASE_D`]: leave the start point where the start point is closed.
#' - [`BASE_P`]: return to the end point where the end point is closed.
#' - [`BASE_Q`]: return to the end point where the end point is closed.
#' - `BASE_C`: self-closed.
#' 
#' The base pattern determines the final form of the curve. 
#' 
#' @return
#' An `sfc_base` object.
#' 
#' @export
#' @import grid
#' @examples
#' BASE_I
sfc_base = function(letter, in_direction, out_direction, grob = NULL, 
	primary = TRUE, open = TRUE) {

	current = c(0, 0)

	in_direction = in_direction %% 360
	out_direction = out_direction %% 360

	x = current[1] - cos(in_direction/180*pi)
	y = current[2] - sin(in_direction/180*pi)

	preceeding = c(x, y)

	x = current[1] + cos(out_direction/180*pi)
	y = current[2] + sin(out_direction/180*pi)

	succeeding = c(x, y)
	
	b = new("sfc_base")
	b@letter = letter
	b@preceeding = preceeding
	b@current = current
	b@succeeding = succeeding
	b@in_direction = in_direction
	b@out_direction = out_direction
	b@primary = primary
	b@open = open

	if(is.null(grob)) {
		
		gbl = list()
		gbl[[1]] = segmentsGrob(b@preceeding[1], b@preceeding[2], b@current[1], b@current[2], default.units = "native", gp = gpar(col = "grey"))
		gbl[[2]] = segmentsGrob(b@current[1], b@current[2], b@succeeding[1], b@succeeding[2], default.units = "native", gp = gpar(col = "grey"), arrow = arrow(angle = 15, length = unit(0.1, "npc")))
		gbl[[3]] = pointsGrob(b@current[1], b@current[2], default.units = "native", pch = 16, size = unit(4, "pt"))
		gb = grobTree(gbl[[1]], gbl[[2]], gbl[[3]], vp = viewport(xscale = c(-1, 1), yscale = c(-1, 1)))
		b@grob = gb
	} else {
		b@grob = grob
	}

	b
}


#' @export
plot.sfc_base = function(x, ...) {
	
	grid.newpage()
	pushViewport(viewport(width = unit(2, "cm"), height = unit(2, "cm")))
	grid.draw(x@grob)
	popViewport()
	
}

#' Print the object
#' 
#' @param object The corresponding object.
#' @rdname show
#' @export
setMethod("show",
	signature = "sfc_base",
	definition = function(object) {

	if(is.na(object@in_direction) && is.na(object@out_direction)) {
		cat("Base: ", object@letter, ", self-closed.\n", sep = "")
	} else if(is.na(object@in_direction) && !is.na(object@out_direction)) {
		cat("Base: ", object@letter, ", with in-closed and ", object@out_direction, " degrees out.\n", sep = "")
	} else if(!is.na(object@in_direction) && is.na(object@out_direction)) {
		cat("Base: ", object@letter, ", with ", object@in_direction, " degrees in and out-closed.\n", sep = "")
	} else {
		cat("Base: ", object@letter, ", with ", object@in_direction, " degrees in and ", object@out_direction, " degrees out.\n", sep = "")
	}
})

#' The graphics object
#' 
#' @rdname sfc_grob
#' @param p The corresponding object.
#' 
#' @return A [`grid::grob()`] object.
#' @export
#' @examples
#' sfc_grob(BASE_I)
setMethod("sfc_grob",
	signature = "sfc_base", 
	definition = function(p) {
	p@grob
})


#' The previous and the next point
#' @aliases sfc_previous_point
#' @rdname sfc_previous_point
#' @param p An `sfc_base` object.
#' @param x The coordinate of the current point.
#' @param rot Rotation of the current point.
#' @param length Length of the segment between the previous/next point and the current point.
#'
#' @return A vector of length two.
#' @export
#' @examples
#' sfc_previous_point(BASE_R, c(0, 0), 0)
#' sfc_previous_point(BASE_R, c(0, 0), 90)
#' sfc_previous_point(BASE_R, c(0, 0), 180)
#' sfc_previous_point(BASE_R, c(1, 0), 0)
#' sfc_previous_point(BASE_R, c(1, 0), 90)
#' sfc_previous_point(BASE_R, c(1, 0), 180)
setMethod("sfc_previous_point",
	signature = "sfc_base",
	definition = function(p, x, rot, length = 1) {

	preceeding = numeric(2)
	pos = rotate_coord(p@current, -rot)
	preceeding[1] = p@current[1] + length*cos(pi + p@in_direction/180*pi)
	preceeding[2] = p@current[2] + length*sin(pi + p@in_direction/180*pi)
	pos = rotate_coord(preceeding, rot)
	pos = move_coord(pos, x)
	pos

})


#' @aliases sfc_next_point
#' @rdname sfc_previous_point
#' @export
#' @examples
#' sfc_next_point(BASE_R, c(0, 0), 0)
#' sfc_next_point(BASE_R, c(0, 0), 90)
#' sfc_next_point(BASE_R, c(0, 0), 180)
#' sfc_next_point(BASE_R, c(1, 0), 0)
#' sfc_next_point(BASE_R, c(1, 0), 90)
#' sfc_next_point(BASE_R, c(1, 0), 180)
setMethod("sfc_next_point",
	signature = "sfc_base",
	definition = function(p, x, rot, length = 1) {

	succeeding = numeric(2)
	succeeding[1] = p@current[1] + length*cos(p@out_direction/180*pi)
	succeeding[2] = p@current[2] + length*sin(p@out_direction/180*pi)
	pos = rotate_coord(succeeding, rot)
	pos = move_coord(pos, x)
	pos
})


# are two base patterns horizontally or vertically connected?
connected_patterns = function(x1, rot1, pos1, x2, rot2, pos2, bases) {
	p2 = sfc_next_point(bases[[ x1 ]], pos1, rot1)
	p1 = sfc_previous_point(bases[[ x2 ]], pos2, rot2)

	all(equal_to(p1, pos1) & equal_to(p2, pos2))
}



# given a previous point, a next point, a current point, choose the base pattern and rotation
guess_base_pattern = function(bases, in_direction, out_direction) {
	diff = out_direction - in_direction
	diff = diff %% 360L
	for(i in seq_along(bases)) {
		b = bases[[i]]
		diff2 = (b@out_direction - b@in_direction) %% 360L
		if(equal_to(diff, diff2)) {
			return(list(letter = b@letter, rot = (in_direction - b@in_direction) %% 360L))
		}
	}
	return(NULL)
}


base_in_direction = function(b, rot) {
	(b@in_direction + rot) %% 360L
}

base_out_direction = function(b, rot) {
	(b@out_direction + rot) %% 360L
}

reduce_loc_to_level_1 = function(p, n, step = 1) {

	loc = sfc_segments(p)
	while(step && nrow(loc) > n*n) {
		loc2 = matrix(nrow = nrow(loc)/n/n, ncol = 2)
		for(i in seq_len(nrow(loc2))) {
			loc2[i, ] = colMeans(loc[seq((i-1)*n*n + 1, i*n*n), ])
		}
		loc = loc2

		step = step - 1
	}

	loc
}


#' @rdname sfc_flip_unit
#' @details
#' The orientation of a unit is the orientation
#' of the line connected by the entry-corner and exit-corner of that unit.
#' @return
#' `unit_orientation()` returns a string one of "vertical", "horizontal", "diagonal_1" and "diagonal_-1".
#' @export
unit_orientation = function(p, index = "") {
	if(inherits(p, "sfc_unit")) {
		loc = sfc_segments(p)
	} else if(length(index) == 0 || identical(index, "")) {
		unit = p[index]

		loc = sfc_segments(unit)
	} else {
		if(!is.character(index)) {
			stop_wrap("`index` should be a character scalar.")
		}
		
		unit = p[index]
		loc = sfc_segments(unit)
		
	}

	n = nrow(loc)
	min_x = min(loc[, 1])
	max_x = max(loc[, 1])
	min_y = min(loc[, 2])
	max_y = max(loc[, 2])

	if( (loc[1,1] < loc[n, 1] && loc[1, 2] < loc[n, 2]) || (loc[1,1] > loc[n, 1] && loc[1, 2] > loc[n, 2]) ) {
		return("diagonal_1")
	} else if( (loc[1,1] > loc[n, 1] && loc[1, 2] < loc[n, 2]) || (loc[1,1] < loc[n, 1] && loc[1, 2] > loc[n, 2]) ) {
		return("diagonal_-1")
	} else if(equal_to(loc[1, 1], loc[n, 1])) {
		return("vertical")
	} else if(equal_to(loc[1, 2], loc[n, 2])) {
		return("horizontal")
	} else {
		return("unknown")
	}
}

#' Flip units
#' @rdname sfc_flip_unit
#' @aliases sfc_flip_unit
#' @param p The corresponding object.
#' @param index A string of digits representing the path on the hierarchy of the curve. The left 
#'         side corresponds to the top level and the right side corresponds to the bottom level 
#'         on the curve. For the 2x2 curve, the digits can only be 1-4, and for the Peano and 
#'         Meander curves, the digits can be 1-9. The hierarchical index should be specified in a format of `i1:i2:i3:...`
#'         where `:` can be replaced by any non-number character. For 2x2 and 3x3 curves, `:` can be omitted and the
#'         hierarchical index can be specified as `i1i2i3...`. See examples in [`sfc_index()`]. The value can also
#'         be a vector where each flipping is applied in sequence.
#' @param to The orientation to flip to. If the specified unit already has such orientation, the function returns
#'       the original curve.
#' @details
#' A unit in the curve is represented as a square block (e.g. `2^k x 2^k` for the 2x2 curve and `3^k x 3^k` for the Peano and Meander curves, `k` between 1 and the level of the curve).
#' In the 2x2 curve, if an unit can be flipped, it is symmetric, thus flipping in the 2x2 curve does not change its form.
#' The flipping is mainly applied on the Peano curve and the Meander curves. Peano curve only allows flippings by the diagonals and the Meander
#' curve only allows flipping horizontally or vertically. The type of flipping is choosen automatically in the function.
#'
#' Currently, `sfc_flip_unit()` only works on curves with a single base pattern as the seed.
#' 
#' @export
#' @return
#' `sfc_flip_unit` returns an `sfc_nxn` object.
#' @examples
#' p = sfc_3x3_meander("I", 11)
#' draw_multiple_curves(
#'     p, 
#'     sfc_flip_unit(p, "1"), # bottom left
#'     sfc_flip_unit(p, "2"), # bottom middle
#'     sfc_flip_unit(p, "3"), # bottom right
#'     nrow = 2)
#' 
#' p = sfc_3x3_peano("I", level = 3)
#' draw_multiple_curves(
#'     p, 
#'     sfc_flip_unit(p, ""),
#'     sfc_flip_unit(p, "2"),
#'     sfc_flip_unit(p, "2:1"),
#'     nrow = 2)
#' 
#' p = sfc_3x3_peano("I", level = 2)
#' draw_multiple_curves(p, 
#'     sfc_flip_unit(p, c("4", "7")),
#'     sfc_flip_unit(p, c("1", "2", "3", "5", "6", "8", "9")),
#'     nrow = 1)
setMethod("sfc_flip_unit",
	signature = "sfc_nxn",
	definition = function(p, index = "", to = NULL) {

	if(length(p@seed) != 1) {
		stop_wrap("Currently `sfc_flip_unit()` only works on curves with a single base pattern as the seed.")
	}

	if(length(p) < (p@mode^2)^p@level) {
		if(!length(p) %in% (p@mode^2)^(seq_len(p@level))) {
			stop_wrap("Since `p` is only a fragment of the curve, it should be represented as a square with length of ", paste((p@mode^2)^(seq_len(p@level)), collapse = ", "), ".")
		}
	}

	if(!is.character(index)) {
		stop_wrap("`index` should be in character.")
	}

	if(length(index) > 1) {
		for(i in index) {
			p = sfc_flip_unit(p, i)
		}
		return(p)
	}

	unit = p[index]
	orientation = unit_orientation(p, index)

	if(identical(orientation, to)) {
		return(p)
	}

	if(orientation == "vertical") {
		unit = sfc_vflip(unit, fix_ends = TRUE, bases = p@rules@bases)
	} else if(orientation == "horizontal") {
		unit = sfc_hflip(unit, fix_ends = TRUE, bases = p@rules@bases)
	} else if(orientation == "diagonal_1") {
		unit = sfc_dflip(unit, slop = 1, fix_ends = TRUE, bases = p@rules@bases)
	} else if(orientation == "diagonal_-1") {
		unit = sfc_dflip(unit, slop = -1, fix_ends = TRUE, bases = p@rules@bases)
	} else {
		stop_wrap("Cannot identify the orientation of the unit.")
	}

	p[sfc_index(p, index)] = unit
	p

})

#' @rdname sfc_flip_unit
#' @param bases Normally use [`BASE_LIST`].
#' @export
setMethod("sfc_flip_unit",
	signature = "sfc_unit",
	definition = function(p, bases) {

	unit = p
	orientation = unit_orientation(unit, "")

	if(orientation == "vertical") {
		unit = sfc_vflip(unit, fix_ends = TRUE, bases = bases)
	} else if(orientation == "horizontal") {
		unit = sfc_hflip(unit, fix_ends = TRUE, bases = bases)
	} else if(orientation == "diagonal_1") {
		unit = sfc_dflip(unit, slop = 1, fix_ends = TRUE, bases = bases)
	} else if(orientation == "diagonal_-1") {
		unit = sfc_dflip(unit, slop = -1, fix_ends = TRUE, bases = bases)
	} else {
		stop_wrap("Cannot identify the orientation of the unit")
	}

	unit
})


# just assume using R, L and I
# the side of the square is determined by the first L/R
guess_unit_corner = function(seq, which = "first") {
	n = length(seq)
	if(which == "first") {
		for(i in 2:n) {
			if(seq@seq[i] %in% c("R", "L")) {
				break
			}
		}
		fragment = seq[1:(i+2)]
		loc = sfc_segments(fragment)
		min_x = min(loc[, 1])
		max_x = max(loc[, 1])
		min_y = min(loc[, 2])
		max_y = max(loc[, 2])
		if(equal_to(loc[1,1], min_x) && equal_to(loc[1,2], min_y)) {
			"bottomleft"
		} else if(equal_to(loc[1,1], min_x) && equal_to(loc[1,2], max_y)) {
			"topleft"
		} else if(equal_to(loc[1,1], max_x) && equal_to(loc[1,2], min_y)) {
			"bottomright"
		} else if(equal_to(loc[1,1], max_x) && equal_to(loc[1,2], max_y)) {
			"topright"
		} else {
			""
		}
	} else {
		for(i in (n-1):1) {
			if(seq@seq[i] %in% c("R", "L")) {
				break
			}
		}

		fragment = seq[(i-2):n]
		loc = sfc_segments(fragment)
		k = nrow(loc)
		min_x = min(loc[, 1])
		max_x = max(loc[, 1])
		min_y = min(loc[, 2])
		max_y = max(loc[, 2])
		if(equal_to(loc[k,1], min_x) && equal_to(loc[k,2], min_y)) {
			"bottomleft"
		} else if(equal_to(loc[k,1], min_x) && equal_to(loc[k,2], max_y)) {
			"topleft"
		} else if(equal_to(loc[k,1], max_x) && equal_to(loc[k,2], min_y)) {
			"bottomright"
		} else if(equal_to(loc[k,1], max_x) && equal_to(loc[k,2], max_y)) {
			"topright"
		} else {
			""
		}
	}
}

---
title: "The sfcurve package"
author: "Zuguang Gu (guzuguang@suat-sz.edu.cn)"
date: '`r Sys.Date()`'
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{The sfcurve package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

Please go to the package's website: https://jokergoo.github.io/sfcurve/.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_base.R
\name{sfc_previous_point,sfc_base-method}
\alias{sfc_previous_point,sfc_base-method}
\alias{sfc_previous_point}
\alias{sfc_next_point,sfc_base-method}
\alias{sfc_next_point}
\title{The previous and the next point}
\usage{
\S4method{sfc_previous_point}{sfc_base}(p, x, rot, length = 1)

\S4method{sfc_next_point}{sfc_base}(p, x, rot, length = 1)
}
\arguments{
\item{p}{An `sfc_base` object.}

\item{x}{The coordinate of the current point.}

\item{rot}{Rotation of the current point.}

\item{length}{Length of the segment between the previous/next point and the current point.}
}
\value{
A vector of length two.
}
\description{
The previous and the next point
}
\examples{
sfc_previous_point(BASE_R, c(0, 0), 0)
sfc_previous_point(BASE_R, c(0, 0), 90)
sfc_previous_point(BASE_R, c(0, 0), 180)
sfc_previous_point(BASE_R, c(1, 0), 0)
sfc_previous_point(BASE_R, c(1, 0), 90)
sfc_previous_point(BASE_R, c(1, 0), 180)
sfc_next_point(BASE_R, c(0, 0), 0)
sfc_next_point(BASE_R, c(0, 0), 90)
sfc_next_point(BASE_R, c(0, 0), 180)
sfc_next_point(BASE_R, c(1, 0), 0)
sfc_next_point(BASE_R, c(1, 0), 90)
sfc_next_point(BASE_R, c(1, 0), 180)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_apply.R
\name{sfc_apply,sfc_nxn-method}
\alias{sfc_apply,sfc_nxn-method}
\alias{sfc_apply}
\title{Apply to every unit in the sfc_nxn curve}
\usage{
\S4method{sfc_apply}{sfc_nxn}(p, depth = 1, fun = function(x) x)
}
\arguments{
\item{p}{An `sfc_nxn` object.}

\item{depth}{An integer between 0 and `level-1` of the curve.}

\item{fun}{A function of which the argument `x` is a subunit in the curve. The subunit is an `sfc_nxn` object but only contains the current sub-sequence.
The function should return an `sfc_seuqence` object with the same length as of `x`. The function can take an optional second argument which
the index of the current subunit in the curve.}
}
\value{
An `sfc_nxn` object.
}
\description{
Apply to every unit in the sfc_nxn curve
}
\details{
This function is mainly used to flip subunits on various levels on the curve, thus mainly on the Peano curve and the Meander curve.
A depth of 0 corresponds to the complete curve. A depth of 1 corresponds to the nine first-level units, et al.

Currently, `sfc_apply()` only works on curves with a single base pattern as the seed.
}
\examples{
p = sfc_3x3_peano("I", level = 3)
# flip the global curve
draw_multiple_curves(
    p, 
    sfc_apply(p, 0, sfc_flip_unit),
    nrow = 1
)

# flip all the subunits on depth = 1
draw_multiple_curves(
    p, 
    sfc_apply(p, 1, sfc_flip_unit),
    nrow = 1
)

# flip all the subunits on depth = 2
draw_multiple_curves(
    p, 
    sfc_apply(p, 2, sfc_flip_unit),
    nrow = 1
)

# flip all level-1 patterns on the Peano curve to horizontal
# only works on the lowest subunit, 
p2 = sfc_apply(p, 2, function(x) {
    if(level1_unit_orientation(x) == "vertical") {
        sfc_flip_unit(x)
    } else {
        x
    }
})
# then on depth=1, only flip the unit with odd index
p3 = sfc_apply(p2, 1, function(x, i) {
    if(i \%\% 2 == 1) {
        sfc_flip_unit(x)
    } else {
        x
    }
})
draw_multiple_curves(p2, p3, nrow = 1)

# flip all level-1 patterns to vertical
p3 = sfc_apply(p, 2, function(x) {
    if(level1_unit_orientation(x) == "horizontal") {
        sfc_flip_unit(x)
    } else {
        x
    }
})
draw_multiple_curves(p, p3, nrow = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_flip_unit.R
\name{unit_orientation}
\alias{unit_orientation}
\alias{sfc_flip_unit,sfc_nxn-method}
\alias{sfc_flip_unit}
\alias{sfc_flip_unit,sfc_unit-method}
\title{Flip units}
\usage{
unit_orientation(p, index = "")

\S4method{sfc_flip_unit}{sfc_nxn}(p, index = "", to = NULL)

\S4method{sfc_flip_unit}{sfc_unit}(p, bases)
}
\arguments{
\item{p}{The corresponding object.}

\item{index}{A string of digits representing the path on the hierarchy of the curve. The left 
side corresponds to the top level and the right side corresponds to the bottom level 
on the curve. For the 2x2 curve, the digits can only be 1-4, and for the Peano and 
Meander curves, the digits can be 1-9. The hierarchical index should be specified in a format of `i1:i2:i3:...`
where `:` can be replaced by any non-number character. For 2x2 and 3x3 curves, `:` can be omitted and the
hierarchical index can be specified as `i1i2i3...`. See examples in [`sfc_index()`]. The value can also
be a vector where each flipping is applied in sequence.}

\item{to}{The orientation to flip to. If the specified unit already has such orientation, the function returns
the original curve.}

\item{bases}{Normally use [`BASE_LIST`].}
}
\value{
`unit_orientation()` returns a string one of "vertical", "horizontal", "diagonal_1" and "diagonal_-1".

`sfc_flip_unit` returns an `sfc_nxn` object.
}
\description{
Flip units
}
\details{
The orientation of a unit is the orientation
of the line connected by the entry-corner and exit-corner of that unit.

A unit in the curve is represented as a square block (e.g. `2^k x 2^k` for the 2x2 curve and `3^k x 3^k` for the Peano and Meander curves, `k` between 1 and the level of the curve).
In the 2x2 curve, if an unit can be flipped, it is symmetric, thus flipping in the 2x2 curve does not change its form.
The flipping is mainly applied on the Peano curve and the Meander curves. Peano curve only allows flippings by the diagonals and the Meander
curve only allows flipping horizontally or vertically. The type of flipping is choosen automatically in the function.

Currently, `sfc_flip_unit()` only works on curves with a single base pattern as the seed.
}
\examples{
p = sfc_3x3_meander("I", 11)
draw_multiple_curves(
    p, 
    sfc_flip_unit(p, "1"), # bottom left
    sfc_flip_unit(p, "2"), # bottom middle
    sfc_flip_unit(p, "3"), # bottom right
    nrow = 2)

p = sfc_3x3_peano("I", level = 3)
draw_multiple_curves(
    p, 
    sfc_flip_unit(p, ""),
    sfc_flip_unit(p, "2"),
    sfc_flip_unit(p, "2:1"),
    nrow = 2)

p = sfc_3x3_peano("I", level = 2)
draw_multiple_curves(p, 
    sfc_flip_unit(p, c("4", "7")),
    sfc_flip_unit(p, c("1", "2", "3", "5", "6", "8", "9")),
    nrow = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{draw_multiple_curves}
\alias{draw_multiple_curves}
\title{Draw multiple curves}
\usage{
draw_multiple_curves(
  ...,
  list = NULL,
  nrow = NULL,
  ncol = NULL,
  extend = TRUE,
  title = FALSE,
  closed = FALSE,
  padding = unit(0, "pt"),
  lwd = 4,
  col = NULL
)
}
\arguments{
\item{...}{A list of `sfc_sequence` objects or objects in its child classes, a list of [`grid::grob`] objects 
or a list of two-column coordinate matrices, i.e., all the forms that can represent curves in this package.}

\item{list}{The list of curve object can also be directly specified as a "list" object.}

\item{nrow}{Number of rows in the layout.}

\item{ncol}{Number of columns in the layout.}

\item{extend}{Whether to draw the entry and exit segments? It is only used when input is a list of `sfc_sequence` objects.}

\item{title}{Whether to add titles on each panel? The title is constructed in the form of `initial_seed|expansion_codes`, e.g. `I|111`.
The value can be a vector of user-defined strings.}

\item{closed}{Whether the curves are closed? The value should be a logical vector. If it is `TRUE`, the last point
is connected to the first point in the curve to form a closed curve. Length of `closed` can be 1 or the number of curves.}

\item{padding}{Space around each curve. The value should be a [`grid::unit`] object with length 1.}

\item{lwd}{Line width with length 1.}

\item{col}{Color for the segments with length 1. If the value is `NULL`, it uses the "Spectral" color palettes from the **RColorBrewer** package.}
}
\value{
No value is returned.
}
\description{
Draw multiple curves
}
\details{
This function is used for quick comparison on curves.
}
\examples{
# for all forms of curves initialized by base pattern 'R', with rotation 0, and on level 3
draw_multiple_curves(
    sfc_2x2("R", "111"),
    sfc_2x2("R", "112"),
    sfc_2x2("R", "121"),
    sfc_2x2("R", "122"),
    sfc_2x2("R", "211"),
    sfc_2x2("R", "212"),
    sfc_2x2("R", "221"),
    sfc_2x2("R", "222"),
    nrow = 2, title = TRUE)

# simply a list of sequences
# note they only contain I/R/L, so the base patterns I/R/L are internally used
draw_multiple_curves(
    sfc_sequence("IIII"),
    sfc_sequence("RRRR"),
    sfc_sequence("RRLL"),
    nrow = 1
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_validate.R
\name{sfc_validate,sfc_sequence-method}
\alias{sfc_validate,sfc_sequence-method}
\alias{sfc_validate}
\alias{sfc_validate,character-method}
\title{Validate the sequence}
\usage{
\S4method{sfc_validate}{sfc_sequence}(p, by = "sfc_2x2")

\S4method{sfc_validate}{character}(p, by = "sfc_2x2")
}
\arguments{
\item{p}{An `sfc_sequence` object or a character string. If it is a character string, rotation of zero is assigned to the first letter.}

\item{by}{One of `sfc_2x2`, `sfc_3x3_peano` and `sfc_3x3_meander`.}
}
\value{
A logical scalar.
}
\description{
Validate the sequence
}
\details{
It is mainly used to validate a seed sequence whether they follow the forward-left-right rule.
}
\examples{
try(sfc_validate("LLLLL"))
try(sfc_validate(sfc_sequence("IIIII", rot = c(0, 90, 180, 270, 0), 
        universe = sfc_universe(SFC_RULES_2x2))))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_rules.R
\name{draw_rules_2x2}
\alias{draw_rules_2x2}
\alias{draw_rules_3x3_peano}
\alias{draw_rules_3x3_meander}
\title{Draw the expansion rules}
\usage{
draw_rules_2x2()

draw_rules_3x3_peano(flip = FALSE)

draw_rules_3x3_meander(flip = FALSE)
}
\arguments{
\item{flip}{Whether to use the "flipped" rules? For the Peano curve and the Meander curve, there is also a "fliiped" version 
of the curve expansion rules. See the vignettes for details.}
}
\value{
No value is returned.
}
\description{
Draw the expansion rules
}
\details{
The expansion rules define how the curve is expanded from level-0 to level-1.
}
\examples{
draw_rules_2x2()
# the units in the main rules of the Peano curve are vertical
draw_rules_3x3_peano()
# the units in the flipped rules of the Peano curve are horizontal
draw_rules_3x3_peano(flip = TRUE)
# the units in the main rules of the Meander curve are "forward"
# i.e. the direction of the "wave" is the same as the direction of the curve
draw_rules_3x3_meander()
# the units in the flipped rules of the Meander curve are "backward"
draw_rules_3x3_meander(flip = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/traverse_path.R
\name{all_traverse_paths}
\alias{all_traverse_paths}
\alias{get_one_traverse_path}
\alias{plot_traverse_paths}
\title{All traverse paths of a sequence}
\usage{
all_traverse_paths(rules, p)

get_one_traverse_path(rules, p)

plot_traverse_paths(rules, p, type = c("all", "11|22", "12|21"))
}
\arguments{
\item{rules}{An `sfc_rules` object.}

\item{p}{An `sfc_sequence` sequence. `p` and `rules` should have the same universe base set.
Please provide `p` as a small sequence because the total number of all traverse paths might be very huge.}

\item{type}{If the value is `"11|22"`, it highlights the paths only via 1-1/2-2 corners. If the value is `"12|21"`, it highlights the paths
only via 1-2/2-1 corners.}
}
\value{
`all_traverse_paths()` returns a list of integer vectors.

`get_one_traverse_path()` returns an integer vector.
}
\description{
All traverse paths of a sequence
}
\details{
Given an input sequence with rotations, `all_traverse_paths()` lists all combinations of expansion 
codes from the first letter to the last letter in `p` (i.e. all possible traverse paths).

`get_one_traverse_path()` returns one random traverse path.
}
\examples{
# expansion rules for the general 3x3 curves
p = SFC_RULES_3x3_COMBINED@rules$I[[3]]
get_one_traverse_path(SFC_RULES_3x3_COMBINED, p)
get_one_traverse_path(SFC_RULES_3x3_COMBINED, p)
get_one_traverse_path(SFC_RULES_3x3_COMBINED, p)
get_one_traverse_path(SFC_RULES_3x3_COMBINED, p)
# 
p = SFC_RULES_3x3_COMBINED@rules$I[[3]]
plot_traverse_paths(SFC_RULES_3x3_COMBINED, p)
plot_traverse_paths(SFC_RULES_3x3_COMBINED, p, type = "11|22")
plot_traverse_paths(SFC_RULES_3x3_COMBINED, p, type = "12|21")

# 2x2 curve
p = sfc_2x2("I", 11)
plot_traverse_paths(SFC_RULES_2x2, p)

# Peano curve
p = sfc_3x3_peano("I", 1)
plot_traverse_paths(SFC_RULES_3x3_PEANO, p)

# Meander curve
p = sfc_3x3_meander("I", 1)
plot_traverse_paths(SFC_RULES_3x3_MEANDER, p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_base.R, R/sfc_grob.R
\name{sfc_grob,sfc_base-method}
\alias{sfc_grob,sfc_base-method}
\alias{sfc_grob,sfc_sequence-method}
\alias{sfc_grob}
\alias{plot.sfc_sequence}
\alias{sfc_grob,sfc_nxn-method}
\alias{plot.sfc_nxn}
\alias{sfc_grob,matrix-method}
\title{The graphics object}
\usage{
\S4method{sfc_grob}{sfc_base}(p)

\S4method{sfc_grob}{sfc_sequence}(
  p,
  bases = NULL,
  extend = FALSE,
  title = FALSE,
  closed = FALSE,
  lwd = 4,
  col = NULL,
  ...
)

\method{plot}{sfc_sequence}(
  x,
  bases = NULL,
  grid = FALSE,
  extend = FALSE,
  title = FALSE,
  closed = FALSE,
  ...
)

\S4method{sfc_grob}{sfc_nxn}(
  p,
  bases = p@rules@bases,
  extend = FALSE,
  title = FALSE,
  closed = FALSE,
  ...
)

\method{plot}{sfc_nxn}(x, grid = FALSE, extend = FALSE, title = FALSE, closed = FALSE, ...)

\S4method{sfc_grob}{matrix}(p, title = NULL, closed = FALSE, lwd = 4, col = NULL, ...)
}
\arguments{
\item{p}{The corresponding object.}

\item{bases}{A list of base patterns, normally [`BASE_LIST`] is used.}

\item{extend}{Whether to add the entry and exit segments?}

\item{title}{Whether to add title on the top of the plot? The title is constructed in the form of `initial_seed|expansion_code`, e.g. `I|111`.
The value can also be a string.}

\item{closed}{Whether the curve is closed?}

\item{lwd}{Line width.}

\item{col}{Color for segments. If the value is `NULL`, it uses the "Spectral" color palettes.}

\item{...}{Other arguments passed to [`grid::viewport()`] or `sfc_grob()`.}

\item{x}{The corresponding object.}

\item{grid}{Whether to add grid lines on the plot?}
}
\value{
A [`grid::grob()`] object.
}
\description{
The graphics object
}
\details{
If `p` is an `sfc_sequence` and `p` contains base patterns defined in `"I/J/R/L/U/B/D/P/Q/C"`,
the default [`BASE_LIST`] is automatically used for `bases`. If `p` is an `sfc_nxn` object, `bases`
is already stored in `p` and it is passed to this function automatically.
}
\examples{
sfc_grob(BASE_I)
plot(sfc_2x2("I", "11"))
plot(sfc_2x2("I", "11"), extend = TRUE, title = TRUE, grid = TRUE)
plot(sfc_sequence("IIIRRR"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_global.R
\docType{data}
\name{BASE_I}
\alias{BASE_I}
\alias{BASE_J}
\alias{BASE_R}
\alias{BASE_L}
\alias{BASE_U}
\alias{BASE_B}
\alias{BASE_D}
\alias{BASE_P}
\alias{BASE_Q}
\alias{BASE_C}
\alias{BASE_LIST}
\title{Base patterns}
\format{
An object of class \code{sfc_base} of length 1.

An object of class \code{sfc_base} of length 1.

An object of class \code{sfc_base} of length 1.

An object of class \code{sfc_base} of length 1.

An object of class \code{sfc_base} of length 1.

An object of class \code{sfc_base} of length 1.

An object of class \code{sfc_base} of length 1.

An object of class \code{sfc_base} of length 1.

An object of class \code{sfc_base} of length 1.

An object of class \code{sfc_base} of length 1.

An object of class \code{list} of length 10.
}
\usage{
BASE_I

BASE_J

BASE_R

BASE_L

BASE_U

BASE_B

BASE_D

BASE_P

BASE_Q

BASE_C

BASE_LIST
}
\value{
`sfc_base` objects.
}
\description{
A list of pre-defined base patterns. See the **Examples** section.
}
\details{
`BASE_I` and `BASE_J` are identical. They are only used to distinguish the two
"going forward" patterns for the level-1 units with 11/22 corner values, i.e. bottom-left to top-right, and bottom-right
to top-left.
}
\examples{
BASE_I
BASE_J
BASE_R
BASE_L
BASE_U
BASE_B
BASE_D
BASE_P
BASE_Q
BASE_C
draw_multiple_curves(
    BASE_I, BASE_J, BASE_R, BASE_L, BASE_U,
    BASE_B, BASE_D, BASE_P, BASE_Q, BASE_C,
    nrow = 2
)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_segments.R
\name{sfc_segments,sfc_nxn-method}
\alias{sfc_segments,sfc_nxn-method}
\alias{sfc_segments}
\alias{sfc_segments,sfc_sequence-method}
\title{Coordinates of the points on the curve}
\usage{
\S4method{sfc_segments}{sfc_nxn}(p, bases = p@rules@bases, start = c(0, 0), ...)

\S4method{sfc_segments}{sfc_sequence}(p, bases = NULL, start = c(0, 0), by = "Cpp")
}
\arguments{
\item{p}{An `sfc_nxn` or `sfc_sequence` object.}

\item{bases}{A list of base patterns, consider to use [`BASE_LIST`].}

\item{start}{Coordinate of the start point.}

\item{...}{Other argument.}

\item{by}{Which implementation? Only for the testing purpose.}
}
\value{
A two-column matrix of coordinates of points on the curve.
}
\description{
Coordinates of the points on the curve
}
\details{
For the `sfc_segments()` on the `sfc_sequence` object, if `bases` is not set,
it uses [`BASE_LIST`] internally. Make sure the sequence only contains the pre-defined base patterns.
}
\examples{
p = sfc_2x2("I", "11")
loc = sfc_segments(p)
plot(loc, type = "l", asp = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_curve.R
\name{sfc_2x2}
\alias{sfc_2x2}
\alias{sfc_3x3_peano}
\alias{sfc_3x3_meander}
\title{Create space-filling curves}
\usage{
sfc_2x2(seed, code = integer(0), rot = 0L)

sfc_3x3_peano(seed, code = integer(0), rot = 0L, level = NULL, flip = FALSE)

sfc_3x3_meander(seed, code = integer(0), rot = 0L, flip = FALSE)
}
\arguments{
\item{seed}{The seed sequence. In most cases, the seed sequence is a single base pattern, which can be specified as a single letter, then `rot` controls
the initial rotation of the base pattern. It also supports a sequence with more than one base patterns as the seed sequence. In this case,
it can be specified as a string of more than one base letters, then `rot` can be set to a single rotation scalar which controls the rotation of the
first letter, or a vector with the same length as the number of base letters.}

\item{code}{A vector of the expansion code. The left side corresponds to the top levels of the curve and the right side corresponds to the bottom level of the curve.
The value can be set as a vector e.g. `c(1, 2, 1)`, or as a string e.g. `"121"`, or as a number e.g. `121`.}

\item{rot}{Rotation of the seed sequence, measured in the polar coordinate system, in degrees.}

\item{level}{Specifically for `sfc_3x3_peano()`, since there is only one expansion code 1, it can also be generated by `rep(1, level)`.}

\item{flip}{Whether to use the "flipped" rules? For the Peano curve and the Meander curve, there is also a "fliiped" version 
of curve expansion rules. On each level expansion in the Peano curve and the Meander curve, a point expands to nine points in 
3x3 grids. Thus the value of `flip` can be set as a logical vector of length of nine that controls whether to use the flipped expansion
for the corresponding unit. Besides such "1-to-9" mode, `flip` can also be set as a function which acccepts the number of current points in the curve and return
a logical vector with the same length, i.e. the "all-to-all*9" mode.}
}
\value{
- `sfc_hilbert()` returns an `sfc_2x2` object.
- `sfc_peano()` returns an `sfc_3x3_peano` object.
- `sfc_meander()` returns an `sfc_3x3_meander` object.

These three classes are child classes of `sfc_nxn`.
}
\description{
Create space-filling curves
}
\details{
- `sfc_2x2()` generates the Hilbert curve from the seed sequence.
- `sfc_3x3_peano()` generates the Peano curve from the seed sequence.
- `sfc_3x3_meander()` generates the Meander curve from the seed sequence.
}
\examples{
sfc_2x2("I", "111") |> plot()
sfc_2x2("I", "111", rot = 90) |> plot()
sfc_2x2("IR", "111", rot = 90) |> plot()
sfc_3x3_peano("I", "111") |> plot()
sfc_3x3_peano("I", "111", 
    flip = c(FALSE, FALSE, FALSE, TRUE, FALSE, FALSE, TRUE, FALSE, FALSE)) |> plot()
sfc_3x3_peano("IJ", "111") |> plot()

sfc_3x3_peano("I", level = 4, flip = function(p) {
    p@rot \%in\% c(90, 270)
}) |> plot(lwd = 1)

level = 4
sfc_3x3_peano("I", level = level, flip = function(p) {
     if(length(p) == 9^(level-1)) {
         l = rep(FALSE, length(p))
         ind = 1:9^2 + 9^2*rep(c(0, 2, 4, 6, 8), each = 9^2)
         l[ind] = p@rot[ind] \%in\% c(90, 270)

         ind = 1:9^2 + 9^2*rep(c(1, 3, 5, 7), each = 9^2)
         l[ind] = p@rot[ind] \%in\% c(0, 180)

         l
    } else {
         rep(FALSE, length(p))
    }
}) |> plot(lwd = 1)

sfc_3x3_meander("I", "111") |> plot()
sfc_3x3_meander("I", "111", 
    flip = c(TRUE, FALSE, TRUE, FALSE, FALSE, TRUE, FALSE, FALSE, TRUE)) |> plot()
sfc_3x3_meander("IR", "111") |> plot()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h_curve.R
\name{sfc_h}
\alias{sfc_h}
\alias{expand_h}
\title{H-curve}
\usage{
sfc_h(h, iteration = 1, connect = c("h", "v"), random = FALSE)

expand_h(h1, h2 = h1, h3 = h1, h4 = h1, connect = "hhhh")
}
\arguments{
\item{h}{The seed of the H-curve. The value should be one of [`H0`], [`H1`] or [`H2`].}

\item{iteration}{Number of iterations.}

\item{connect}{How the four subunits are connected to form the H-curve on the next level. See **Details**.}

\item{random}{Whether to generate subunits randomly on each iteration.}

\item{h1}{The first subunit on the bottom left.}

\item{h2}{The second subunit on the top left.}

\item{h3}{The third subunit on the top right.}

\item{h4}{The fourth subunit on the bottom right.}
}
\value{
A two-column matrix of coordinates of points on the curve.
}
\description{
H-curve
}
\details{
An H-curve on level k is composed with four subunits on level k-1. If we number the four subunits in the following order:

```
2  3
1  4
```

where subunit 1 connects to subunit 2, subunit 2 connects to subunit 3, et al., and e.g. subunit 1 connects to subunit 2 
via its toprigth corner. Since H-curve can be thought of as a closed curve, to, e.g. let subunit 1 to connect to  
subunit 2, its topright corner needs to be opened. There are two segments on subunit 1 that can be removed/opened: the horizontal
segment and the vertical segment on the topright corner in subunit 1.

In this way, in `sfc_h()`, the argument `connect` only accepts a single value of `"h"` or `"v"` where the types of segments for 
all the four subunits are the same, i.e. whether all the horizontal corner segments are opened or whether all the vertical corner
segments are opened. In `expand_h()`, the argument `connect` can be set to a string with four letters or a vector of length four where the type of segments of the
four subunits can be set separately.

In the random mode, each subunit is generated randomly, the type of the open segment is choosen randomly, also each subunit has a probability of 0.5
to rotate by 90 degrees.
}
\examples{
draw_multiple_curves(
    sfc_h(H0, iteration = 2),
    sfc_h(H2, iteration = 2),
    closed = TRUE, nrow =1
)
draw_multiple_curves(
    sfc_h(H1, iteration = 3, random = TRUE),
    sfc_h(H1, iteration = 3, random = TRUE),
    closed = TRUE, nrow = 1
)
draw_multiple_curves(
    expand_h(H0, connect = "hvvh"),
    expand_h(H1, connect = "vvhh"),
    closed = TRUE, nrow = 1
)

# set the four subunits separately
h1 = expand_h(H0, connect = "hhhh")
h2 = expand_h(H0, connect = "vvvv")
h3 = expand_h(H0, connect = "hvhv")
h4 = expand_h(H0, connect = "hvvh")
expand_h(h1, h2, h3, h4, connect = "vhvh") |> 
    plot_segments(closed = TRUE)

fun = function(h, iteration) {
    for(i in 1:iteration) h = expand_h(h, connect = "vhvh")
    h
}
fun(H0, 4) |> plot_segments(closed = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_curve.R
\name{sfc_level,sfc_nxn-method}
\alias{sfc_level,sfc_nxn-method}
\alias{sfc_level}
\title{The level of the curve}
\usage{
\S4method{sfc_level}{sfc_nxn}(p)
}
\arguments{
\item{p}{An `sfc_nxn` object.}
}
\value{
An integer.
}
\description{
The level of the curve
}
\examples{
p = sfc_2x2("I", "11")
sfc_level(p)

p = sfc_2x2("I", "1111")
sfc_level(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_curve.R, R/sfc_rules.R
\name{sfc_mode,sfc_nxn-method}
\alias{sfc_mode,sfc_nxn-method}
\alias{sfc_mode,sfc_rules-method}
\alias{sfc_mode}
\title{The mode of the curve}
\usage{
\S4method{sfc_mode}{sfc_nxn}(p)

\S4method{sfc_mode}{sfc_rules}(p)
}
\arguments{
\item{p}{The corresponding object.}
}
\value{
An integer.
}
\description{
The mode of the curve
}
\examples{
p = sfc_2x2("I", "1")
sfc_mode(p)
p = sfc_3x3_peano("I", "1")
sfc_mode(p)
sfc_mode(SFC_RULES_2x2)
sfc_mode(SFC_RULES_3x3_PEANO)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_generator.R
\name{sfc_generator}
\alias{sfc_generator}
\title{Generate a nxn curve based on expansion rules}
\usage{
sfc_generator(
  rules,
  name,
  envir = topenv(parent.frame()),
  flippable = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{rules}{An `sfc_rules` object.}

\item{name}{Name of the curve. The name will be used for the functions that will be generated.}

\item{envir}{Environment where the functions are exported.}

\item{flippable}{Whether `rules` can have flipped version? If it is `TRUE`, the generated function also accepts the `flip` argument.}

\item{verbose}{Whether to print messages?}
}
\value{
No value is returned.
}
\description{
Generate a nxn curve based on expansion rules
}
\details{
Two functions will be exported:

- `sfc_\{name\}()`
- `draw_rules_\{name\}()`

For the simplicity, flipping is not supported yet.
}
\examples{
UNIVERSE_4x4_PEANO = c("I", "R", "L")

RULES_4x4_PEANO = list()
RULES_4x4_PEANO[["I"]][[1]] = sfc_unit("RIILLIIRRIILLIIR", rot = 0, universe = UNIVERSE_4x4_PEANO)
RULES_4x4_PEANO[["I"]][[2]] = sfc_hflip(RULES_4x4_PEANO[["I"]][[1]])
RULES_4x4_PEANO[["R"]][[1]] = sfc_unit("IIIRRIILLIIRRIIL", rot = 0, universe = UNIVERSE_4x4_PEANO)      
RULES_4x4_PEANO[["R"]][[2]] = sfc_rotate(sfc_unit("LIIRRIILLIIRRIII", 
                                rot = 270, universe = UNIVERSE_4x4_PEANO), 90)
RULES_4x4_PEANO[["L"]][[1]] = sfc_hflip(RULES_4x4_PEANO[["R"]][[2]])
RULES_4x4_PEANO[["L"]][[2]] = sfc_hflip(RULES_4x4_PEANO[["R"]][[1]])

SFC_RULES_4x4_PEANO = sfc_rules(rules = RULES_4x4_PEANO,
        name = "Peano 4x4",
        bases = BASE_LIST[UNIVERSE_4x4_PEANO])

sfc_generator(SFC_RULES_4x4_PEANO, "4x4_peano")
draw_rules_4x4_peano()
sfc_4x4_peano("I", 111) |> plot()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_base.R, R/sfc_curve.R, R/sfc_rules.R,
%   R/sfc_sequence.R
\name{show,sfc_base-method}
\alias{show,sfc_base-method}
\alias{show,sfc_nxn-method}
\alias{show,sfc_rules-method}
\alias{show,sfc_sequence-method}
\title{Print the object}
\usage{
\S4method{show}{sfc_base}(object)

\S4method{show}{sfc_nxn}(object)

\S4method{show}{sfc_rules}(object)

\S4method{show}{sfc_sequence}(object)
}
\arguments{
\item{object}{The corresponding object.}
}
\value{
No value is returned.
}
\description{
Print the object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_rules.R
\name{sfc_expand_by_rules,sfc_rules,sfc_nxn-method}
\alias{sfc_expand_by_rules,sfc_rules,sfc_nxn-method}
\alias{sfc_expand_by_rules}
\alias{sfc_expand_by_rules,sfc_rules,factor-method}
\alias{sfc_expand_by_rules,sfc_rules,character-method}
\title{Expand a sequence}
\usage{
\S4method{sfc_expand_by_rules}{sfc_rules,sfc_nxn}(p, seq, code = 1L, flip = FALSE, by = "Cpp")

\S4method{sfc_expand_by_rules}{sfc_rules,factor}(p, seq, code = 1L, flip = FALSE, by = "Cpp")

\S4method{sfc_expand_by_rules}{sfc_rules,character}(p, seq, code = 1L, flip = FALSE, by = "Cpp")
}
\arguments{
\item{p}{An `sfc_rules` object.}

\item{seq}{An `sfc_nxn` object or other objects.}

\item{code}{The expansion code.}

\item{flip}{For the Peano curve and the Meander curves, each unit can be flipped without affecting other parts in the curve. This argument
controls whether to flip the unit. Since currently it only works on the Peano curve and the Meander curve, `flip` should be a logical
vector of length one or with the same length as `seq`. Whether it flips horizontally, vertically or against the diagonal line is automatically choosen.
The value of `flip` can also be a function which takes the current curve as the only argument.}

\item{by}{Which implementation? Only for the testing purpose.}
}
\value{
If `seq` is an `sfc_nxn` object, the function also returns an "expanded" `sfc_nxn` object. Or else it returns an `sfc_sequence` object.
}
\description{
Expand a sequence
}
\examples{
sfc_expand_by_rules(SFC_RULES_2x2, sfc_2x2("I"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_rules.R
\name{sfc_rules}
\alias{sfc_rules}
\title{Constructor of the sfc_rules class}
\usage{
sfc_rules(rules, bases, flip = list(), name = "sfc_rules")
}
\arguments{
\item{rules}{A list of rules.}

\item{bases}{A list of base patterns.}

\item{flip}{A list of rules. They are "flipped" version of `rules`. The value can also simply be `TRUE`, then
the flipped version is automatically generated from `rules`.}

\item{name}{A self-defined string.}
}
\value{
An `sfc_rules` object.
}
\description{
Constructor of the sfc_rules class
}
\details{
It is mainly used internally.

`rules` is a two-level list. It is in a format of `rules[[ base ]][[ expansion_code ]] = sfc_unit()`.
In the following example where we define the expansion rules for the 2x2 curve:

```
UNIVERSE_2x2 = c("I", "R", "L", "U", "B", "D", "P", "Q", "C")
RULES_2x2 = list()
RULES_2x2[["I"]][[1]] = sfc_unit(c("R", "L", "L", "R"), rot = 0, universe = UNIVERSE_2x2)
```

`I` is the level-0 base pattern, `[[1]]` corresponds to the first form of expansion to level-1, and the value
assigned is a [`sfc_unit()`] object which is basically a list of base patterns.

Then we also need to provide the base patterns which define how to extend the curve. The list of base patterns
is assigned to the `bases` argument. In the same example, we set `bases` as:

```
list("I" = BASE_I, "R" = BASE_R, "L" = BASE_L, "U" = BASE_U, ...)
```

where e.g. [`BASE_I`] is a pre-defined base pattern in the [`sfc_base`] class.

There are the following pre-defined rules:

- [`SFC_RULES_2x2`]
- [`SFC_RULES_3x3_PEANO`]
- [`SFC_RULES_3x3_MEANDER`]
- [`SFC_RULES_3x3_COMBINED`]
- [`SFC_RULES_4x4_MEANDER_1`]
- [`SFC_RULES_4x4_MEANDER_2`]

Check \url{https://github.com/jokergoo/sfcurve/blob/master/R/zz_global.R} to see how these pre-defined rules are constructed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_base.R
\name{sfc_base}
\alias{sfc_base}
\title{Constructor of the sfc_base class}
\usage{
sfc_base(
  letter,
  in_direction,
  out_direction,
  grob = NULL,
  primary = TRUE,
  open = TRUE
)
}
\arguments{
\item{letter}{A single letter to represent the base pattern.}

\item{in_direction}{The direction of the segment that enters the point, measured in the polar coordinate system, in degrees.}

\item{out_direction}{The direction of the segment that leaves the point, measured in the polar coordinate system, in degrees.}

\item{grob}{A [`grid::grob()`] object of this base pattern. If it is not set, it is generated according to `in_direction` and `out_direction`.}

\item{primary}{Currently, going forward, turning left and turning right can be set as primary base patterns because other high-level patterns
can be built from them.}

\item{open}{Can the base pattern be connected to other base patterns?}
}
\value{
An `sfc_base` object.
}
\description{
Constructor of the sfc_base class
}
\details{
The "base pattern" is designed not only for single point but also for combination of points that form a "base curve". However,
currently, it is fixed to the single point base pattern. 

Currently, this package supports 2x2 and 3x3 space-filling curves that fills grids in 2D space constructed by the Gaussian integers.
And when the curve expands, we only allow the segments to go forward, backward, left and right. Thus there are the following base patterns
pre-defined in this package:

- [`BASE_I`]/[`BASE_J`]: go forward.
- [`BASE_R`]: turn right.
- [`BASE_L`]: turn left.
- [`BASE_U`]: go backward.
- [`BASE_B`]: leave the start point where the start point is closed.
- [`BASE_D`]: leave the start point where the start point is closed.
- [`BASE_P`]: return to the end point where the end point is closed.
- [`BASE_Q`]: return to the end point where the end point is closed.
- `BASE_C`: self-closed.

The base pattern determines the final form of the curve.
}
\examples{
BASE_I
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h_curve.R
\docType{data}
\name{H0}
\alias{H0}
\alias{H1}
\alias{H2}
\title{Seed sequences of the H-curve}
\format{
An object of class \code{matrix} (inherits from \code{array}) with 4 rows and 2 columns.

An object of class \code{matrix} (inherits from \code{array}) with 16 rows and 2 columns.

An object of class \code{matrix} (inherits from \code{array}) with 16 rows and 2 columns.
}
\usage{
H0

H1

H2
}
\value{
Two-column matrices.
}
\description{
Seed sequences of the H-curve
}
\details{
The three objects simply contain coordinates of points on the three base H-curves.
}
\examples{
H0
draw_multiple_curves(H0, H1, H2, nrow = 1, closed = TRUE)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_global.R
\docType{data}
\name{SFC_RULES_2x2}
\alias{SFC_RULES_2x2}
\alias{SFC_RULES_3x3_PEANO}
\alias{SFC_RULES_3x3_MEANDER}
\alias{SFC_RULES_3x3_COMBINED}
\alias{SFC_RULES_4x4_MEANDER_1}
\alias{SFC_RULES_4x4_MEANDER_2}
\title{Rules}
\format{
An object of class \code{sfc_rules} of length 1.

An object of class \code{sfc_rules} of length 1.

An object of class \code{sfc_rules} of length 1.

An object of class \code{sfc_rules} of length 1.

An object of class \code{sfc_rules} of length 1.

An object of class \code{sfc_rules} of length 1.
}
\usage{
SFC_RULES_2x2

SFC_RULES_3x3_PEANO

SFC_RULES_3x3_MEANDER

SFC_RULES_3x3_COMBINED

SFC_RULES_4x4_MEANDER_1

SFC_RULES_4x4_MEANDER_2
}
\value{
`sfc_rules` objects.
}
\description{
A list of pre-defined expansion rules for different curves.
}
\details{
`SFC_RULES_3x3_PEANO`, `SFC_RULES_3x3_MEANDER` and `SFC_RULES_3x3_COMBINED`, `SFC_RULES_MEANDER_4x4_1`, `SFC_RULES_MEANDER_4x4_2` also contain the "flipped" expansion rules.

`SFC_RULES_3x3_COMBINED` is a combination of `SFC_RULES_3x3_PEANO` and `SFC_RULES_3x3_MEANDER` where
in `SFC_RULES_3x3_PEANO`, `J` is replaced by its original pattern `I`.

`SFC_RULES_4x4_MEANDER_1` and `SFC_RULES_4x4_MEANDER_2` are extension rules of Meander curves (3x3)
on the 4x4 mode It is only for the demonstration purpose, thus only `I/R/L` are supported.
}
\examples{
SFC_RULES_2x2
SFC_RULES_3x3_PEANO
SFC_RULES_3x3_MEANDER
SFC_RULES_3x3_COMBINED
SFC_RULES_4x4_MEANDER_1
SFC_RULES_4x4_MEANDER_2
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_index.R
\name{[.sfc_nxn}
\alias{[.sfc_nxn}
\alias{sfc_index,sfc_nxn-method}
\alias{sfc_index}
\alias{test_sfc_index}
\title{Subunit in the curve}
\usage{
\method{[}{sfc_nxn}(x, i, j, ..., drop = TRUE)

\S4method{sfc_index}{sfc_nxn}(p, index = "")

test_sfc_index(p, index)
}
\arguments{
\item{x}{An `sfc_nxn` object.}

\item{i, index}{A string of digits representing the path on the hierarchy of the curve. The left side
corresponds to the top level and the right side corresponds to the bottom level on the curve. For the
2x2 curve, the digits can only be 1-4, and for the Peano and Meander curves, the digites can be 1-9.
The hierarchical index should be specified in a format of `i1:i2:i3:...`
  where `:` can be replaced by any non-number character. For 2x2 and 3x3 curves, `:` can be omitted and the
  hierarchical index can be specified as `i1i2i3...`. See the **Examples** section.}

\item{j}{A value of `TRUE` or `FALSE` that controls whether to keep the `sfc_nxn` class or degenerate to the `sfc_sequence` class.}

\item{...}{Ignore.}

\item{drop}{A value of `TRUE` or `FALSE` that controls whether to keep the `sfc_nxn` class or degenerate to the `sfc_sequence` class.}

\item{p}{An `sfc_nxn` object.}
}
\value{
`sfc_index()` returns an integer vector.
}
\description{
Subunit in the curve
}
\details{
`sfc_index()` only works on square curves (i.e. a curve with a single base letter as seed.)

`test_sfc_index()` is a helper function for demonstrating `sfc_index()`.
}
\examples{
p = sfc_2x2("I", "11111")
p["3:2:1"]
# for 2x2 and 3x3 curves, ":" can be omitted
p["321"]
p["3:2:1", TRUE] # or p["3:2:1", drop = FALSE]
# only for testing
p = sfc_2x2("I", "11111")
om = par(no.readonly = TRUE)
par(mfrow = c(2, 2))
test_sfc_index(p, "3")
test_sfc_index(p, "3:2")
test_sfc_index(p, "3:2:1")
test_sfc_index(p, "3:2:1:1")
par(om)

p = sfc_3x3_meander("I", "11111")
om = par(no.readonly = TRUE)
par(mfrow = c(2, 2))
test_sfc_index(p, "7")
test_sfc_index(p, "7:5")
test_sfc_index(p, "7:5:9")
test_sfc_index(p, "7:5:9:2")
par(om)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_4x4_meander.R
\name{sfc_4x4_meander}
\alias{sfc_4x4_meander}
\alias{sfc_expand,sfc_4x4_meander-method}
\alias{draw_rules_4x4_meander}
\title{4x4 space-filling curves in meander type}
\usage{
sfc_4x4_meander(seed, code = integer(0), rot = 0L, flip = FALSE, type = 1L)

\S4method{sfc_expand}{sfc_4x4_meander}(p, code, flip = FALSE)

draw_rules_4x4_meander(type = 1, flip = FALSE)
}
\arguments{
\item{seed}{The seed sequence. In most cases, the seed sequence is a single base pattern, which can be specified as a single letter, then `rot` controls
the initial rotation of the base pattern. It also supports a sequence with more than one base patterns as the seed sequence. In this case,
it can be specified as a string of more than one base letters, then `rot` can be set to a single rotation scalar which controls the rotation of the
first letter, or a vector with the same length as the number of base letters.}

\item{code}{A vector of the expansion code. The left side corresponds to the higher levels (more to the top-level) of the curve and the right side corresponds to the lower level (more to the bottom-level) of the curve.
The value can be set as a vector e.g. `c(1, 2, 1)`, or as a string e.g. `"121"`, or as a number e.g. `121`.}

\item{rot}{Rotation of the seed sequence, measured in the polar coordinate system, in degrees.}

\item{flip}{The same setting as in [`sfc_3x3_peano()`] or [`sfc_3x3_meander()`].}

\item{type}{Which type of rules to use? 1 for [`SFC_RULES_4x4_MEANDER_1`] and 2 for [`SFC_RULES_4x4_MEANDER_2`].}

\item{p}{An `sfc_4x4_meander` object.}
}
\value{
`sfc_4x4_meander()` returns an `sfc_4x4_meander` object.
}
\description{
4x4 space-filling curves in meander type
}
\details{
It is an extension of the 3x3 Meander curves to mode 4. For simplicity, it only supports `I/R/L` base patterns.
}
\examples{
draw_multiple_curves(
    sfc_4x4_meander("I", "11", type = 1),
    sfc_4x4_meander("I", "12", type = 1),
    sfc_4x4_meander("I", "11", type = 2),
    sfc_4x4_meander("I", "12", type = 2),
    nrow = 2
)
seed = paste(rep(paste0("R", sapply(0:10, function(i) strrep("I", i))), each = 2), collapse="")
sfc_4x4_meander(seed, 1) |> plot()
draw_rules_4x4_meander(type = 1)
draw_rules_4x4_meander(type = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_sequence.R
\name{[.sfc_sequence}
\alias{[.sfc_sequence}
\alias{[<-.sfc_sequence}
\alias{length.sfc_sequence}
\alias{c.sfc_sequence}
\title{Utility functions}
\usage{
\method{[}{sfc_sequence}(x, i)

\method{[}{sfc_sequence}(x, i) <- value

\method{length}{sfc_sequence}(x)

\method{c}{sfc_sequence}(...)
}
\arguments{
\item{x}{An `sfc_sequence` object.}

\item{i}{Numeric index.}

\item{value}{An `sfc_sequence` object.}

\item{...}{A list of `sfc_sequence` objects or other arguments.}
}
\value{
`length.sfc_sequence()` returns an integer scalar.

`c.sfc_sequence()` returns an `sfc_sequence` object.
}
\description{
Utility functions
}
\details{
For efficiency, `c.sfc_sequence()` does not check whether the input `sfc_sequence` objects are compatible.
}
\examples{
p = sfc_sequence("ABCDEFGH")
p
p[1:4]
p[1:4] = p[8:5]
p
length(p)
c(p, p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_expand.R
\name{sfc_expand,sfc_2x2-method}
\alias{sfc_expand,sfc_2x2-method}
\alias{sfc_expand}
\alias{sfc_expand,sfc_3x3_peano-method}
\alias{sfc_expand,sfc_3x3_meander-method}
\title{Expand the curve to the next level}
\usage{
\S4method{sfc_expand}{sfc_2x2}(p, code, flip = FALSE)

\S4method{sfc_expand}{sfc_3x3_peano}(p, code = 1, flip = FALSE)

\S4method{sfc_expand}{sfc_3x3_meander}(p, code, flip = FALSE)
}
\arguments{
\item{p}{An `sfc_2x2` object or other related objects.}

\item{code}{Expansion code, a single integer.}

\item{flip}{Whethe to flip level-1 units? The value should be a logical vector of length one or the same as the length of `p`.}
}
\value{
An object in the same class as the input.
}
\description{
Expand the curve to the next level
}
\details{
For the Hilbert curve and Meander curve, as long as the expansion code of the first base pattern in the sequence is determinted, 
the expansion codes for other base patterns in the sequence are all determined. For the Peano curve, since
there is only one traverse path on any level, `code` is ignored.

These functions are mainly used internally.
}
\examples{
p = sfc_2x2("I", 11)
sfc_expand(p, 2) # I|211
p = sfc_3x3_peano("I", 11)
sfc_expand(p, 2) # I|211
p = sfc_3x3_meander("I", 11)
sfc_expand(p, 2) # I|211
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_rules.R, R/sfc_sequence.R
\name{sfc_universe,sfc_rules-method}
\alias{sfc_universe,sfc_rules-method}
\alias{sfc_universe}
\alias{sfc_universe,sfc_sequence-method}
\title{The universe base pattern set}
\usage{
\S4method{sfc_universe}{sfc_rules}(p)

\S4method{sfc_universe}{sfc_sequence}(p)
}
\arguments{
\item{p}{The corresponding object.}
}
\value{
A vector of base patterns.
}
\description{
The universe base pattern set
}
\examples{
sfc_universe(SFC_RULES_2x2)
sfc_universe(SFC_RULES_3x3_PEANO)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/00_class.R
\docType{package}
\name{sfcurve-package}
\alias{sfcurve}
\alias{sfcurve-package}
\title{sfcurve: 2x2, 3x3 and Nxn Space-Filling Curves}
\description{
Implementation of all possible forms of 2x2 and 3x3 space-filling curves, i.e., the generalized forms of the Hilbert curve \url{https://en.wikipedia.org/wiki/Hilbert_curve}, the Peano curve \url{https://en.wikipedia.org/wiki/Peano_curve} and the Peano curve in the meander type (Figure 5 in \url{https://eudml.org/doc/141086}). It can generates nxn curves expanded from any specific level-1 units. It also implements the H-curve and the three-dimensional Hilbert curve. See \doi{10.48550/arXiv.2412.16962} for more details.
}
\details{
Please go to the package's website: \url{https://jokergoo.github.io/sfcurve/} for more details.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/jokergoo/sfcurve}
  \item \url{https://jokergoo.github.io/sfcurve/}
}

}
\author{
\strong{Maintainer}: Zuguang Gu \email{guzuguang@suat-sz.edu.cn} (\href{https://orcid.org/0000-0002-7395-8709}{ORCID})

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/standard_curve.R
\name{plot_segments}
\alias{plot_segments}
\title{Plot segments}
\usage{
plot_segments(x, grid = FALSE, title = FALSE, closed = FALSE, ...)
}
\arguments{
\item{x}{A two-column matrix of coordinates of points.}

\item{grid}{Whether to add grid lines on the plot?}

\item{title}{The value should be `FALSE` or a string.}

\item{closed}{Whether the curve is closed?}

\item{...}{Other arguments passed to [`sfc_grob()`].}
}
\value{
No value is returned.
}
\description{
Plot segments
}
\details{
This function is only for a quick demonstration of curves represented as two-column coordinate matrices.
}
\examples{
pos = cbind(c(0, 0, 1, 1, 2, 2, 3, 3, 2, 2, 1, 1),
            c(1, 2, 2, 3, 3, 2, 2, 1, 1, 0, 0, 1))
plot_segments(pos)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_reduce.R
\name{sfc_reduce,sfc_nxn-method}
\alias{sfc_reduce,sfc_nxn-method}
\alias{sfc_reduce}
\alias{sfc_reduce,matrix-method}
\alias{add_base_structure}
\title{Reduce a curve}
\usage{
\S4method{sfc_reduce}{sfc_nxn}(p, to = sfc_level(p) - 1)

\S4method{sfc_reduce}{matrix}(p, to = level - 1, mode = NULL)

add_base_structure(gb, level = 1)
}
\arguments{
\item{p}{An `sfc_nxn` object.}

\item{to}{Which level to reduce to? Value should be between 1 and `sfc_level(p) - 1`.}

\item{mode}{Mode of the curve.}

\item{gb}{A `grob` object returned by [`sfc_grob()`] or a `sfc_nxn` object then [`sfc_grob()`] is internally applied.}

\item{level}{The level of the unit.}
}
\value{
A two-column matrix of coordinates of the reduced curve.
}
\description{
Reduce a curve
}
\details{
The reduction is applied on the coordinates of points.

`add_base_structure()` adds a base structure on a certain level to the curve.
}
\examples{
p = sfc_3x3_peano("I", level = 3)
draw_multiple_curves(
    p, 
    sfc_reduce(p, 2), 
    sfc_reduce(p, 1), 
    nrow = 1)
p = hilbert_curve(level = 4)
draw_multiple_curves(
    p, 
    sfc_reduce(p, 3), 
    sfc_reduce(p, 2), 
    nrow = 1)
p = hilbert_curve(3)
draw_multiple_curves(
    add_base_structure(p, level = 1),
    add_base_structure(p, level = 2),
    nrow = 1
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_curve.R
\name{level1_unit_orientation,sfc_3x3_peano-method}
\alias{level1_unit_orientation,sfc_3x3_peano-method}
\alias{level1_unit_orientation}
\alias{change_level1_unit_orientation,sfc_3x3_peano-method}
\alias{change_level1_unit_orientation}
\alias{level1_unit_orientation,sfc_3x3_meander-method}
\alias{change_level1_unit_orientation,sfc_3x3_meander-method}
\title{Level-1 unit in the Peano curve and Meander curve}
\usage{
\S4method{level1_unit_orientation}{sfc_3x3_peano}(p)

\S4method{change_level1_unit_orientation}{sfc_3x3_peano}(p, to = c("horizontal", "vertical"))

\S4method{level1_unit_orientation}{sfc_3x3_meander}(p)

\S4method{change_level1_unit_orientation}{sfc_3x3_meander}(p, to = c("forward", "backward"))
}
\arguments{
\item{p}{For `level1_unit_orientation()`, it is an `sfc_3x3_peano` or `sfc_3x3_meander` unit on level 1.
For `change_level1_unit_orientation()`, it is a normal `sfc_3x3_peano` or `sfc_3x3_meander` object.}

\item{to}{A string of "vertical/horizontal" (on the `sfc_3x3_peano` object) or "forward/backward" (on the `sfc_3x3_meander` object).}
}
\value{
`level1_unit_orientation()` returns "verticalhorizontal" (on the `sfc_3x3_peano` object) or "forward/backward" (on the `sfc_3x3_meander` object).

`change_level1_unit_orientation()` returns an `sfc_3x3_peano` or `sfc_3x3_meander` object.
}
\description{
Level-1 unit in the Peano curve and Meander curve
}
\details{
"vertical" and "horizontal" correspond to the direction of the "long segments" in a Peano curve (see **Examples**).

`level1_unit_orientation()` is normally used inside [`sfc_apply()`]. `change_level1_unit_orientation()`
changes all level-1 units of a Peano curve or a Meander curve simultaneously.
}
\examples{
p = sfc_3x3_peano("I", 111)
# the first level-1 unit
level1_unit_orientation(p[1:9, TRUE])
# the fourth level-1 unit
level1_unit_orientation(p[1:9 + 27, TRUE])
p2 = change_level1_unit_orientation(p, "horizontal")
p3 = change_level1_unit_orientation(p, "vertical")
draw_multiple_curves(p, p2, p3, 
    title = c("original", "all horizontal", "all vertical"), nrow = 1)
# by default, orientations of all level-1 units in Meander curve are forward
p = sfc_3x3_meander("I", 111)
level1_unit_orientation(p[1:9, TRUE])
p2 = change_level1_unit_orientation(p, "backward")
draw_multiple_curves(p, p2,
   title = c("all forward", "all backward"), nrow = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_sequence.R
\name{sfc_sequence}
\alias{sfc_sequence}
\alias{sfc_seed}
\alias{sfc_unit}
\title{Constructor of the sfc_sequence class}
\usage{
sfc_sequence(seq, rot = 0L, universe = NULL)

sfc_seed(seq, rot = 0L, universe = NULL)

sfc_unit(seq, rot = 0L, universe = NULL)
}
\arguments{
\item{seq}{A sequence of base patterns. The value can be a vector of letters or a single string.}

\item{rot}{The corresponding rotations of base patterns. If it has length one and the sequence contains R/L/I (right/left/forward),
`rot` controls the rotation of the first base pattern and the rotations for remaining base patterns in the sequence are automatically
calculated.}

\item{universe}{The universe of base patterns. A vector of letters.}
}
\value{
An `sfc_sequence` object.
}
\description{
Constructor of the sfc_sequence class
}
\details{
This funtion is very low-level. Normally, users do not need to directly use this constructor.

`sfc_seed` class is the same as the `sfc_sequence` class. It is used specifically as the "seed sequence"
when generating the curves.

`sfc_unit` class also inherits the `sfc_sequence` class but has one additionally slot: `corner`. 
It is used specifically when defining the expansion rules.
}
\examples{
sfc_sequence("ABCD", rot = c(0, 90, 180, 270), universe = c("A", "B", "C", "D"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_transform.R
\name{sfc_rotate,sfc_sequence-method}
\alias{sfc_rotate,sfc_sequence-method}
\alias{sfc_rotate}
\alias{^.sfc_sequence}
\alias{sfc_hflip,sfc_sequence-method}
\alias{sfc_hflip}
\alias{sfc_vflip,sfc_sequence-method}
\alias{sfc_vflip}
\alias{sfc_dflip,sfc_sequence-method}
\alias{sfc_dflip}
\alias{sfc_reverse,sfc_sequence-method}
\alias{sfc_reverse}
\alias{rev.sfc_sequence}
\title{Transformations of a sequence}
\usage{
\S4method{sfc_rotate}{sfc_sequence}(p, rot)

\method{^}{sfc_sequence}(e1, e2)

\S4method{sfc_hflip}{sfc_sequence}(p, fix_ends = FALSE, bases = NULL)

\S4method{sfc_vflip}{sfc_sequence}(p, fix_ends = FALSE, bases = NULL)

\S4method{sfc_dflip}{sfc_sequence}(p, slop = 1L, fix_ends = FALSE, bases = NULL)

\S4method{sfc_reverse}{sfc_sequence}(p)

\method{rev}{sfc_sequence}(x)
}
\arguments{
\item{p, e1}{An `sfc_sequence` object.}

\item{rot, e2}{Rotation measured in the polar coordinate system, in degrees.}

\item{fix_ends}{By default, the curve is flipped as a complete whole, which means, the associated entry and exit directions of the curve
is also adjusted accordingly. When flipping subunits in a curve, e.g. level-1 subunits in a Peano curve, we want the 
entry and exit direction of the subunit not changed so that the subunits are still connected in the curve after the flipping. In this 
case, `fix_ends` can be set to `TRUE`, then only the subunits are flipped while the connections to neighbouring subunits are not affected. 
See the **Examples** section.}

\item{bases}{A list of base patterns, consider to use [`BASE_LIST`]. It is only used when `fix_ends = TRUE`.}

\item{slop}{Slop of the diagonal. Value can only be 1 or -1.}

\item{x}{An `sfc_sequence` object.}
}
\value{
An `sfc_sequence` object.
}
\description{
Transformations of a sequence
}
\details{
- `sfc_rotate()` and `^()` rotate each base pattern.
- `sfc_hflip()` flips a sequence horizontally.
- `sfc_vflip()` flips a sequence vertically.
- `sfc_dflip()` flips a sequence against a diagonal line (with slop 1 or -1).
}
\examples{
p = sfc_3x3_meander("R", 2, rot = -90)
draw_multiple_curves(
    p, 
    sfc_hflip(p), 
    sfc_hflip(p, fix_ends = TRUE), 
    nrow = 1)
p = sfc_3x3_meander("L", 2, rot = -90)
draw_multiple_curves(
    p, 
    sfc_vflip(p), 
    sfc_vflip(p, fix_ends = TRUE), 
    nrow = 1)
p = sfc_3x3_peano("I", 2)
draw_multiple_curves(
    p, 
    sfc_dflip(p, 1), 
    sfc_dflip(p, 1, fix_ends = TRUE), 
    nrow = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_3x3_combined.R
\name{sfc_3x3_combined}
\alias{sfc_3x3_combined}
\alias{sfc_expand,sfc_3x3_combined-method}
\alias{draw_rules_3x3_combined}
\title{General 3x3 space-filling curves}
\usage{
sfc_3x3_combined(seed, level = 0, rot = 0L, flip = FALSE)

\S4method{sfc_expand}{sfc_3x3_combined}(p, code = NULL, flip = FALSE)

draw_rules_3x3_combined(flip = FALSE)
}
\arguments{
\item{seed}{The seed sequence. In most cases, the seed sequence is a single base pattern, which can be specified as a single letter, then `rot` controls
the initial rotation of the base pattern. It also supports a sequence with more than one base patterns as the seed sequence. In this case,
it can be specified as a string of more than one base letters, then `rot` can be set to a single rotation scalar which controls the rotation of the
first letter, or a vector with the same length as the number of base letters.}

\item{level}{Level of the curve. Currently it is restricted to an integer no bigger than 5.}

\item{rot}{Rotation of the seed sequence, measured in the polar coordinate system, in degrees.}

\item{flip}{The same setting as in [`sfc_3x3_peano()`] or [`sfc_3x3_meander()`].}

\item{p}{An `sfc_3x3_combined` object.}

\item{code}{Ignore. The traverse codes are selected randomly.}
}
\value{
`sfc_3x3_combined()` returns an `sfc_3x3_combined` object.
}
\description{
General 3x3 space-filling curves
}
\details{
This type of 3x3 curve uses the combintation of base patterns from both the Peano curve and the Meander curve.
On each level, the traverse path is randomly selected.
}
\examples{
draw_multiple_curves(
    sfc_3x3_combined("I", level = 3),
    sfc_3x3_combined("I", level = 3),
    sfc_3x3_combined("I", level = 3),
    nrow = 1
)
draw_rules_3x3_combined()
draw_rules_3x3_combined(flip = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_shape.R
\name{sfc_shape,sfc_2x2-method}
\alias{sfc_shape,sfc_2x2-method}
\alias{sfc_shape}
\alias{all_2x2_shapes}
\alias{sfc_shape,sfc_3x3_peano-method}
\alias{all_3x3_peano_shapes}
\alias{sfc_shape,sfc_3x3_meander-method}
\alias{all_3x3_meander_shapes}
\title{Shape of the curve}
\usage{
\S4method{sfc_shape}{sfc_2x2}(p)

all_2x2_shapes(level = 2)

\S4method{sfc_shape}{sfc_3x3_peano}(p)

all_3x3_peano_shapes(level = 2)

\S4method{sfc_shape}{sfc_3x3_meander}(p)

all_3x3_meander_shapes(level = 2)
}
\arguments{
\item{p}{An `sfc_2x2` object.}

\item{level}{Level of the 2x2 curve.}
}
\value{
`sfc_shape()` returns a two-column data frame of the xy-coordinates of the shape curve.

`all_2x2_shapes()` returns a list of `n` two-column data frames where each data frame corresponds to
the xy-coordnates of the corresponding shape curve.
}
\description{
Shape of the curve
}
\details{
The shape of the curve is defined as a form of the curve without considering entry/exit directions, rotation, 
flipping (reflection) nor reversing.

## 2x2 curve

The process of selecting the shape segment of the curve denoted as `P` is:

1. The entry-point should locate in the bottom left subunit and the exit-point should 
   locate in the bottom right subunit. We try the four rotations (0, 90, 180, 270), and 
   the four rotations on the horizontally flipped curve. Once we find the transformed curve
   that satisfies this criterion, we name it as `P2`.
2. We also generate `P3` which is a horizontally flipped version of `rev(P2)`.
3. We compare the first point `p` of `P2` and `P3`, and select the one whose `p` has the smaller
   x-coordinate (i.e. more to the left of the curve). If the x-coordinates of `p` are the same in
   `P2` and `P3`, we select the one whose `p` has the smaller y-coordinate.

## 3x3 Peano curve

The process of selecting the shape segment of the curve denoted as `P` is:

1. The entry-point should locate in the bottom left subunit and the exit-point should 
   locate in the top right subunit. We try the four rotations (0, 90, 180, 270). Once we find the transformed curve
   that satisfies this criterion, we name it as `P2`.
2. We also generate `P3` which is a 180 degrees rotation on the reversed `P2`, `P4` which is a diagonal flip with slop of 1 on `P2`
   and `P5` which is a diagonal flip with slop of 1 on `P3`.
3. We calculate the "UID" of `P(2-5)` and pick the one with the smallest UID as the final curve.

The UID of a 3x3 Peano curve is based on the hierarchical indices of the units on it. The hierarchy of the
Peano curve is traversed in a depth-first manner. On each node, the orientation of the corresponding unit is
calculated where vertical is 1 and horizontal is 2. The digits are concatenated into a long string.

`all_3x3_peano_shapes()` only calculates all shapes for Peano curve on level 2.

`all_3x3_meander_shapes()` only considers the Meander curve with all subunits on all levels in forward orientation.
}
\examples{
p1 = sfc_2x2("I", 11)
p2 = sfc_2x2("R", 22)
draw_multiple_curves(
    p1, p2, 
    sfc_shape(p1), sfc_shape(p2), 
    col = "black")
sl = all_2x2_shapes(2)
draw_multiple_curves(list = sl, lwd = 2, col = "black")
sl = all_2x2_shapes(3)
draw_multiple_curves(list = sl, lwd = 2, col = "black")
p = sfc_3x3_peano("I", 11)
draw_multiple_curves(
    p, sfc_dflip(p),
    sfc_shape(p), sfc_shape(sfc_dflip(p)),
    col = "black")
\donttest{
sl = all_3x3_peano_shapes()
length(sl)
# the first 8 shapes
draw_multiple_curves(sl[1:8], col = "black")
}
sl = all_3x3_meander_shapes(2)
draw_multiple_curves(list = sl, lwd = 2, col = "black")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/standard_curve.R
\name{hilbert_curve}
\alias{hilbert_curve}
\alias{moore_curve}
\alias{beta_omega_curve}
\alias{peano_curve}
\alias{meander_curve}
\alias{h_curve}
\title{Various curves in their standard forms}
\usage{
hilbert_curve(level = 2L, by = "Cpp")

moore_curve(level = 2L)

beta_omega_curve(level = 2L)

peano_curve(level = 2L, pattern = "vvvvvvvvv", by = "Cpp")

meander_curve(level = 2L, pattern = "fffffffff", code = rep(1, level))

h_curve(iteration = 2L)
}
\arguments{
\item{level}{Level of the curve.}

\item{by}{Which implementation? Only for the testing purpose.}

\item{pattern}{The orientation of units on level-2, i.e. the orientation of the 9 3x3 units. The 
value should be a string with 9 letters of "v"/"h" (vertical or horizontal) for the Peano curve,
and "f"/"b" (forward or backward) for the Meander curve. The length of the string should be maximal 9.
If the length is smaller than 9, the stringis automatically recycled.}

\item{code}{Internally used.}

\item{iteration}{Number of iterations.}
}
\value{
A two-column matrix of coordinates of points on the curve.
}
\description{
Various curves in their standard forms
}
\details{
These are just special forms of [`sfc_2x2()`], [`sfc_3x3_peano()`], [`sfc_3x3_meander()`] and [`sfc_h()`].
}
\examples{
hilbert_curve(2)
draw_multiple_curves(
    hilbert_curve(3),
    hilbert_curve(4),
    nrow = 1
)
draw_multiple_curves(
    moore_curve(3),
    moore_curve(4),
    nrow = 1
)
draw_multiple_curves(
    beta_omega_curve(3),
    beta_omega_curve(4),
    nrow = 1
)
draw_multiple_curves(
    peano_curve(2),
    peano_curve(3),
    nrow = 1
)
draw_multiple_curves(
    peano_curve(3, pattern = "vh"),
    peano_curve(3, pattern = "vvvhhhvvv"),
    nrow = 1
)
draw_multiple_curves(
    meander_curve(2),
    meander_curve(3),
    nrow = 1
)
draw_multiple_curves(
    meander_curve(3, pattern = "fbfbfbfbf"),
    meander_curve(3, pattern = "bbbbbffff"),
    nrow = 1
)
draw_multiple_curves(
    h_curve(1),
    h_curve(2),
    nrow = 1, closed = TRUE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hilbert_3d.R
\name{hilbert_3d}
\alias{hilbert_3d}
\title{Three dimensional Hilbert curve}
\usage{
hilbert_3d(level = 2L)
}
\arguments{
\item{level}{The level of the curve.}
}
\value{
A three-column matrix of coordinates of points on the 3D Hilbert curve.
}
\description{
Three dimensional Hilbert curve
}
\details{
There are many forms of 3D Hilbert curve. Here we only implement one specific form.
}
\examples{
pos = hilbert_3d(2)
\dontrun{
if(require(rgl) && interactive()) {
    plot3d(pos, type = "l", lwd = 4, col = 2)
}
}
}
\seealso{
Michael Bader. Space-Filling Curves: An Introduction with Applications in Scientific Computing, Springer Science & Business Media, 2012. \doi{10.1007/978-3-642-31046-1}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sfc_sequence.R
\name{sfc_is_compatible,sfc_sequence,sfc_sequence-method}
\alias{sfc_is_compatible,sfc_sequence,sfc_sequence-method}
\alias{sfc_is_compatible}
\alias{sfc_is_compatible,sfc_sequence,sfc_rules-method}
\alias{sfc_is_compatible,sfc_rules,sfc_sequence-method}
\title{Whether two sfc_sequence objects are compatible}
\usage{
\S4method{sfc_is_compatible}{sfc_sequence,sfc_sequence}(p1, p2, strict = TRUE)

\S4method{sfc_is_compatible}{sfc_sequence,sfc_rules}(p1, p2)

\S4method{sfc_is_compatible}{sfc_rules,sfc_sequence}(p1, p2)
}
\arguments{
\item{p1}{An `sfc_sequence` object.}

\item{p2}{An `sfc_sequence` object.}

\item{strict}{`TRUE` or `FALSE`, see **Details**.}
}
\value{
A logical scalar.
}
\description{
Whether two sfc_sequence objects are compatible
}
\details{
The function compares whether the two universe base pattern sets are identical.
If `strict` is `TRUE`, the order of the two universe sets should also be the same.
If `strict` is `FALSE`, the universe set of `p2` can be a subset of the universe set of `p1`.
}
\examples{
p1 = sfc_2x2("I")
p2 = sfc_2x2("R")
sfc_is_compatible(p1, p2)

p1 = sfc_2x2("I")
p2 = sfc_sequence("R")
sfc_is_compatible(p1, p2)
sfc_is_compatible(p1, p2, strict = FALSE)

p1 = sfc_sequence("ABC")
p2 = sfc_sequence("DEF")
sfc_is_compatible(p1, p2)
}
